/*
 * action.h: 
 * declare action functions (eg. for copy, move)
 */

#ifndef ACTION_H
#define ACTION_H

#include "global.h"

/* versions that keep the question form on screen, until the action 
 * is done. They include a progress indicator. It is possible to set 
 * the default action (return button) to the cancel button. 
 */
int ask_action_work_question(
    const char *question, 
    const char *source_dir, 
    const char *target_dir,
    int n_files,                // 1 (good defaults)
    const char **help_text,     // NULL
    const char *help_title,     // NULL
    int default_go_flag,  	// TRUE		     
    int &new_target_dir_flag,   // important
    int no_multi_work_flag,     // FALSE
    int danger_flag	        // FALSE
);

int ask_action_work_question(
    const char *question, 
    const char *source_dir, 
    int n_files,                // 1
    const char **help_text,     // NULL,
    const char *help_title,     // NULL,
    int default_go_flag,        // TRUE			     
    int no_multi_work_flag,     // FALSE,
    int danger_flag	        // FALSE
);

/* If the form was never shown in the first place, don't fail the
 * checks. Use the following function for proper initialization.
 */
void action_work_noform(void);

/* crucial for interaction between tree-change-target when the action
 * work form is shown.
 */
void action_work_form_target_sync(const char *new_target_dir);

void action_end(void);
void action_msg(char *buffer);
int action_step(int check_forms_flag = TRUE);
int action_mini_step(int check_forms_flag = TRUE);

/* unspecific action. No reference to browsers or anything. 
 * the action_fp should return the results from do_fork_cmd.
 */
typedef int (*action_fpt)(void);

int fork_action(
    action_fpt action_fp, 
    const char *question,
    const char **help_text = NULL,
    const char *help_title = NULL,
    int   danger_flag = FALSE
);
 

// do action on _fname_ in _directory_.
typedef int (*source_dir_action_single_fpt)(const char *, const char *);
typedef int (*source_dir_action_many_fpt)  (const char *);

/* action that takes place on the source browsers directory only.
 * the flag can be used/usec to ensure that even for single file op's,
 * there'll always be a dialog box. Also, a indicator box can be 
 * switched off.
 */
int source_dir_action(
    source_dir_action_single_fpt source_dir_action_single,
    source_dir_action_many_fpt   source_dir_action_many,
    const char                   *action_string,
    int                          action_do_ask_one_file_flag = FALSE,
    int                          action_default_go_flag = TRUE,
    const char                   **help_text = NULL,
    int                          action_no_multi_work_flag  = FALSE,
    int                          action_no_file_string_flag = FALSE,
    int                          danger_flag = FALSE
);

/* version of source_dir_action that works with source and target
 * browsers. source and target browsers cannot be the same.
 */

// do action on _fname_ from _directory_ to _directory_.
typedef int (*source_and_target_dir_action_single_fpt)(const char *, const char *, const char *);
typedef int (*source_and_target_dir_action_many_fpt)  (const char *, const char *);

int source_and_target_dir_action(
    source_and_target_dir_action_single_fpt source_and_target_dir_action_single,
    source_and_target_dir_action_many_fpt   source_and_target_dir_action_many,
    const char                   *action_string,
    int                          action_do_ask_one_file_flag = FALSE,
    const char                   **help_text = NULL
);

/* useful help function HARDWIRED with action functions that allows
 * to maintain a sensible position for the current line _after_ the 
 * action. Not perfect, we also need an 'changed' option, e.g. for 
 * rename. adjacent == 1 step below is good for move or remove, also
 * for copy.
 * call get_old before doing the actual file operation, and call
 * current only if the file operation succeeded. It is kludgy, but
 * in the nature of the operation (?).
 */
//int adjacent_get_oldline(void);
//void adjacent_current_line(const char *fname, int oldline);

/* best versions, set names in both source and target browsers .
 * also, have the versions which restore and reset them. 
 */

void remember_source_current(const char *fname);
void remember_target_current(const char *fname);

void restore_source_current(void);
void restore_target_current(void);

void reset_source_current(void);
void reset_target_current(void);

/* extra: get a new target dir via the tree, if necessary */
int fully_validate_source_target_dirs(const char *source_dir, char *target_dir, const char *action_string);

void action_sync_defaults(void);

/* allows to bypass the mechanism that binds the target dir to
 * the target browser. If you set the override, make sure it
 * gets queried unequivocally to reset the internal flags!
 */
int target_dir_get_override(char target_dir[]);
void target_dir_set_override(const char *override_dir);

#ifdef OLDCODE
/* two versions, one that does work only on one source directory,
 * one that uses both a source and a target directory.
 */
int ask_action_question(const char *question, const char *source_dir);
int ask_action_question(const char *question, const char *source_dir, const char *target_dir);

#endif

#endif	// ACTION_H

