
#ifndef BROWSER_H
#define BROWSER_H

#include <limits.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fstream.h>
#include <unistd.h>

#include "forms.h"
#include "global.h"

/* mini-kludge to allow default arguments for get_first/next_selection.
 */
static int dummy_zero = 0;

/* necessary for toolbox, so that it can do fkeys etc.
 */
int browser_handle_free(FL_OBJECT *obj, int event, FL_Coord, FL_Coord, int key, void *xev);

/*
 * A convenient way to deal with selections is to pass an
 * action function to the browser. It will call the function
 * on all selected file names. Note that if the function returns
 * false, the complete action loop is terminated. So, an action
 * function can include a `cancel all' feature _and_ a `cancel file'
 * feature.
 * Also note that if nothing is selected, the current line is used.
 *
 * action functions take file names and return `cancel all' status.
 * they also take an int counter, which is incremented by 1 if the
 * action was performed on the current file. This allows detailed
 * tracking of what happened during the loop.
 */

typedef int (*action_func_pt)(const char *, int &);

/*
 * note that the browser_c class contains functionality to interface
 * with the file access monitor daemon.
 *
 * default directory on startup is '.'
 * call show() instead of fudging a fl_show_form(form). 
 */

class browser_c {
public:
    static int
	auto_raise_flag,
	next_ID,			// handle with care.
	directory_update_count		// number of directory checks per 10 seconds. 
					// Default to 40, equiv. 4 per second.
    ;                  

    FL_FORM *form;
    FL_OBJECT
	*free_obj, *source_box, *target_box, *menu_text,
        *text, *text_box, *search_text, *underscores,
	*panel, *menu_hidden_button, *column_hidden_button
    ;

    browser_c(void);
   ~browser_c();
    
    /*
     * make another browser, insert it into the browser ring.
     * return success.
     */
    int clone(void);

    /* raise the window */
    void win_raise(void);

    /* freezing, unfreezing & redrawing the browser form */
    void 
        freeze(void), 
        unfreeze(int full_redraw = FALSE),
	redraw(void)
    ;

    /* time stamp the browser dir */
    void time_stamp(void);
    int  time_stamp_check(int update_flag = TRUE);
    struct stat stamp_stat;

    /*
     * callback look-alike functions named after suggested
     * keybindings.
     * the insert key should be bound to toggle_select_current();
     * the enter/return key is too complex to be included into
     * the class protocol.
     */

    void
        home(void),
	end(void),
	down_arrow(int n_lines = 1),
	up_arrow  (int n_lines = 1),
	page_down(void),
	page_up(void)
    ;

    /*
     * this function reliably cuts out the last token from
     * the desired line in the browser. This happens to be
     * the file's name.
     * This function is clever enough to supply the _link name_
     * in case the file is a symbolic link, notwithstanding the
     * fact that the display uses the a -> b notation (i.e., we
     * extract `a'). How can we be sure we got the a properly 
     * terminated to the left (possible pathologic case) ?
     */

    const char *extract_filename(int line);
    llong_t extract_size(int line);
    char extract_type(int line);
    char *extract_item(int line, sorttype_et which);

    /*
     * key job of the browser is to maintain the selection
     * state of the files it contains. The external program
     * can access them as follows:
     * call get_first_selection for the first one, then 
     * continue calling get_next_selection until it returns
     * NULL. This is like good ole strtok(3C).
     * If you need the actual line numbers, supply an argument
     * for the reference (data out) parameter.
     * 
     * The `check_' variant doesn't doesn't have any internal state,
     * it just returns the first selection without interference. 
     * 
     * call get_current_fname for the file name of the current line.
     * call set_current_fname to specify a preferred location for
     * the current bar. It fails gracefully if fname cannot
     * be found, and returns success. It also has a flag which
     * allows to autocenter the new line if it isn't visible. 
     * If it is already visible, it will just be set.
     * use the set_preferred_fname version _before_ an actual
     * redraw is being made. This avoids one redraw.
     *
     * use exact if you only want an exact match.
     *
     * (de) select_file works on that file name only. It has an
     * internal access cache that makes it work very efficiently
     * in conjunction with the get_first and get_next functions.
     */
    
    int   set_current_fname        (const char *fname, int auto_center_flag = FALSE);
    int   set_current_exact_fname  (const char *fname, int auto_center_flag = FALSE);
    int   set_current_below_match  (const char *pattern, int below_line);
    int   set_current_first_match  (const char *pattern);
    void  set_preferred_fname      (const char *fname);    
    int   get_current_line         (void);    
    const char *get_current_fname  (int &current_line        = dummy_zero);
    const char *get_first_selection(int &first_selected_line = dummy_zero);
    const char *get_next_selection (int &next_selected_line  = dummy_zero);

    const char *check_first_selection(int &first_selected_line = dummy_zero);

    int selection_cmd(const char *file, const char *arg0, const char *arg1, const char *arg2 = NULL);

    void deselect_file(const char *fname);
    void mirror_selection(void);


    /*
     * the select_all/deselect_all functions work on all
     * files in the browser, including directories, etc.
     * exception: select_all DOES NOT  select the two
     * directories '.' and '..'. In the future, there
     * behaviour will be further limited by extra flags,
     * e.g. match only files (not directories). However,
     * as an extra precaution, the dot dirs are also
     * deselected in deselect_all.
     * It is also possible to select on the full line.
     * still, it is tried to prevent . and .. from being selected. 
     */

    void deselect_all_line_matches(const char *pattern);
    void deselect_all_name_matches(const char *pattern);
    int deselect_first_name_match(const char *pattern);   // return true if successful deselection.
    void select_all_line_matches(const char *pattern);
    void select_all_name_matches(const char *pattern);

    void deselect_all_item_matches(const char *pattern, sorttype_et which);
    void select_all_item_matches(const char *pattern, sorttype_et which);

    /* execute action on all selected filenames. See the
     * notes on the definition of the action function type
     * for more information.
     *
     * we return the number of times `action' was executed,
     * so you can avoid an update if e.g. op was cancelled,
     * or has no effect on the current dir. 
     */
    int selection_action_loop(action_func_pt action);
    
    /* toggle_select_current toggles the select status
     * of the current line. Kind of handy for the insert key
     * callback.
     *
     * also handy from time to time to get the number of
     * selected entrys.
     */
    void toggle_select_current(void);
    int get_n_selected(void);
    llong_t get_selection_byte_size(int use_current_flag = FALSE);
    llong_t get_total_byte_size(void);
  
    /*
     * the following make the browser source/passive. It is
     * visible by toggling a little light from green (source)
     * to grey (passive) and vice versa.
     *
     * we can also set_target({TRUE, FALSE}). The target
     * browser gets a yellow light. 
     * for speed, we have a function which sets both in one go. 
     */

    void set_source(int boolval, int update_title_flag = TRUE);
    void set_target(int boolval, int update_title_flag = TRUE);
    void set_source_target(int source_val, int target_val);
    
    /* set the complete window title, including the
     * correct current working directory, and the
     * source/passive status of the browser.
     * assumes that the cwd_buf is correctly set.
     */
    void set_complete_wintitle(const char *override_str = NULL);

    /* set the filter pattern used for display. Setting it to
     * the string "*" switches it off.
     * If the filter_full_line flag is set, full lines, not only
     * file names, are used for filtering. 
     */
    void set_filter_pattern_and_update(const char *new_pattern);
    inline const char *get_filter_pattern(void){
	return filter_pattern;
    }
    int filter_full_line_flag;

    /* save and retrieve your default settings. Note that it is
     * assumed that the file is not open (!). That's a bit of
     * one-time resource wasting, but makes things more independent.
     */

    int defaults_save(ofstream &ofile);
    int defaults_load(ifstream &ifile);
    int defaults_load(const char *fname);
    
    /* this is an error checking version of the fl internal
     * function. Better catch the bugs early. Also, have a
     * function to just center (ctrl-l) the current line.
     * On autocenter, center the current line if it wasn't
     * already visible. 
     */
    void set_current(int line, int auto_center_flag = FALSE);
    void center_current(void);

    /*
     * show is a convenient shortcut to fl_show_form.
     * additionally, it remembers the window ID. Usually
     * _don't_ supply an argument to show - it's internal.
     */

    void show(int defaults_checked_flag = FALSE);

    /*
     * the update function reloads the browser from its
     * current directory. This function is clever enough
     * to change directory, reset the window title, etc.
     * if necessary.
     * It also assembles the necessary command from the
     * browser's internal flags.
     *
     * it takes a flag as argument indicating wether the
     * list of selections should be maintained over the
     * directory re-read. It will also try to maintain 
     * the current bar on the file it was before the 
     * update, and the scrolling position (topline).
     */

    void update(int keep_selection_current_flag = FALSE);

    /* 
     * this object contains lots of state, both about what
     * to display and how to sort it. 
     * note that we can sort on multiple fields.
     *
     * show options:
	 * There is no separate show option for extensions.
     */
    int list_hidden_flag,
	show_inode_flag,
	show_filetype_flag,
	show_perms_flag,
	show_nlinks_flag,
	show_owner_flag,
	show_symbolic_flag,
	show_group_flag,
	show_size_flag,
	show_modific_flag,
	show_access_flag,
	show_time_seconds_flag,
	show_time_years_flag,
	show_type_char_flag
    ;

    int
	show_inode_len,
	show_filetype_len,
	show_perms_len,
	show_nlinks_len,
	show_owner_len,
	show_symbolic_len,
	show_group_len,
	show_size_len,
	show_modific_len,
	show_access_len,
	show_time_seconds_len,
	show_time_years_len,
	show_type_char_len
    ;

    /* treated among the show options:
     */
    int follow_links_flag, long_wintitle_flag;

    /* 
     * sort options and convenience functions.
     */
    sorttype_et
        fst_sort,
	snd_sort
    ;
    int
	sort_fst_reverse_flag,
	sort_snd_reverse_flag
    ;

    void new_fst_sort(sorttype_et sorttype, int reverse_flag = FALSE);
    void new_snd_sort(sorttype_et sorttype, int reverse_flag = FALSE);

    /*
     * let's also have a pointer to the next/prevs browser, so all
     * browsers in one app can form a ring. Default initialize
     * this thing to point to itself.
     */

    browser_c *next_browser, *prev_browser;

    /*
     * functions to access/set the browser's internal ID.
     */
    int get_ID(void){return my_ID;}
    void set_ID(int new_ID){
	if(new_ID >= 0){
	    my_ID = new_ID;
	    set_complete_wintitle();
	}
    }
	 
    /*
     * standard colour manipulation:
     */
    void set_dark_yellow_bright(void);
    void set_dark_yellow(void);
    void set_dark_red(void);
    void set_dark_green(void);
    void set_blueish(void);
    void set_reddish(void);
    void set_greenish(void);
    void set_yellowish(void);
    void set_greyish(void);
    void set_whitish(void);

    void sync_source_target_colour(void);
    void sync_colour_opts (browser_c *browser);
    void sync_display_opts(browser_c *browser);
    void sync_sort_opts   (browser_c *browser);

    /* font size & style.
     * suggested are styles 3 (courier normal) and 4 (courier bold)
     * and sizes around 11.0.
     */

    void set_fontsize(int newsize);
    void set_fontstyle(int newstyle);    

    /* 
     * setting display, font and sort options defaults.
     */

    void set_font_defaults(void);
    void set_show_defaults(void);
    void set_sort_defaults(void);

    /*
     * ringbuffer functions.
     * find the last browser in the ringbuffer.
     * insert a new browser into the ringbuffer after us.
     */

    browser_c* rb_get_last_browser(void);
    void rb_insert_after(browser_c* new_browser);

    /* wrapper for key speed search processing, and ... */
    int key_event_callback(char key, KeySym keysym);

    /* process the ctrl-s requests. */
    void search_event_cb(void);
    int search_again_event_cb(void);
    int search_last_line;

    /* return true if the key event was eaten by the search mode */    
    int search_key_event_callback(char key, KeySym keysym);

    /* access function */
    int get_selection_workline(void);

    /*
     * linecheck is a use-always procedure that barfs if a
     * browser function is called with an invalid line.
     */
    int linecheck(int line, int quiet_flag = FALSE);

    /*
     * The margin function is used internally by up and down arrow.
     * It maintains a +-2 top or bottom margin where possible.
     * handle with care, it is new.
     */
    void set_current_with_margin(int newline);

    /* for insert key optimization, remember the direction of the last cur. mov. 
     * default resets to false.
     */
    int cur_up_flag;

    /* before executing a callback, do this (hides the search text).
     */
    void default_actions(void);

    /* trigger the menus at the top of the browser 
     */
    void column_hidden_button_cb(FL_OBJECT *, long);
    FL_Coord compute_column_x_offset(sorttype_et which);

    int line_is_dir(int line, const char *fname);

    const char *get_cwd(void);
    void set_cwd_buf(const char *buf);

    void rescale(int force_flag = FALSE);

    /* window parameters, used for saving and restore
     */
    int win_x, win_y, win_w, win_h;

    void forward_selection(void);

    /* for external tracking. Assumes event is within browser form */
    const char *xdnd_track(FL_Coord mx, FL_Coord my);

    /* for raw callback. */
    int handle_keypress(XKeyEvent *xkev);
    int handle_keyrelease(XKeyEvent *xkev);
private:
    char *last_wintitle;   // optimize wintitle setting.
    int old_n_selected;
    void selection_remember(void);
    void selection_change_msg(void);

    /* for the goddamn automounters in this world. */
    FILE *amd_dir_kludge;

    /*
     * let's have a buffer for the current directory string.
     * be a bit careful setting this to invalid things.
     */
   
    static int cwd_bufsize;
    char *cwd_buf;

    const char * adjust_length(const char *status_buf);
    /*
     * internal function that adds a date format string
     * to the cmd being assembled.
     */
    
    void add_date_string(char *buf);
    
    /* used for extraction. */
    int calc_fieldnum(sorttype_et which);

    /*
     * this creation function is deliberately kept separate
     * from the constructor, so that form design updates are
     * easier to incorporate.
     */

    void create(void);

    /*
     * headline is used to generate the proper column headers
     * for the text object.
     */

    void headline(void);

    /* have menus at correct position when called via shortcuts */
    void move_focus_to_column(sorttype_et which);

    /*
     * The following two functions are used to generate a 
     * properly formatted line for the ls-module. 
     * note that one should call print_cmd after sort_cmd.
     */

    void print_cmd(int &argc, char *argv[]);
    void sort_cmd(int &argc, char *argv[]);
    void sort_process_option(
	char *buf,
	sorttype_et sorttype,
	int reverse_flag
    );

    /*
     * the following kludge is used for communication
     * between get_first_selection and get_next_selection.
     * To justify its existence a bit more, it is used for
     * a consistency check (can't call get_next without a
     * preceding get_first). It is also useful as a cache
     * for fast access.
     * also for actions, we have a public access function for it. 
     */
    int selection_workline;

    /*
     * store wether we are source or passive.
     * default on startup is passive.
     * also store wether we're target. Default
     * on startup is no target.
     */

    int source_flag;
    int target_flag;


    /*
     * we don't want lunatic settings, so we check.
     */

    void window_props_update(void);
    void window_no_lunatic(void);

    /*
     * browser colour parameters are intrinsic in the 
     * browser data structure and not replicated here.
     */

    /*
     * store a unique ID for each instance.
     * useful for loading/saving defaults.
     * change this only in exceptional circumstances,
     * via the get/set_ID functions.
     */
    
    int my_ID;

    /* to indicate wether we are currently reading a dir. */
    int reading_flag;

    /* file filter set-up */
    char *filter_pattern;

    /* remember your search position, indicate wether we can wrap */
    int search_again_line,
	search_wrap_possible_flag
    ;
    char search_last_fname_buf[NAME_MAX+1];
};

#endif  // browser.h

