//#include <iostream.h>
#include <stdio.h>
#include <string.h>
#include <sys/errno.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <unistd.h>

#include "browser.h"
#include "cb_duplicate.h"
#include "error_notifier_forms_aux.h"
#include "help_forms_aux.h"
#include "file_utility.h"
#include "global.h"
#include "help_text.h"
#include "pathnames.h"
#include "utility.h"

#include "action_forms.h"
#include "action_forms_aux.h"
#include "options_forms.h"
#include "toolbox_forms_aux.h"
#include "tree_forms_aux.h"
#include "fork_pipe.h"

/* all functions/callbacks associated with the duplicate op.
 *
 * This version uses the most advanced general concept.
 * it uses source_dir_action for the loop through 1/many setup
 *         one unified testing/access function, switched by loop_flag
 *         selection_action_loop from browser_c::
 *
 * now extended to full copy/rename.
 */

static int
    copy_flag                  = FALSE,   // allow for full copy renames
    loop_flag                  = FALSE,   // are we in a loop ?
    loop_ask_directory_flag    = TRUE,
    loop_ask_overwrite_flag    = TRUE,
    loop_ask_writeprotect_flag = TRUE
;

static const char
    *loop_target_dir = NULL,
    *loop_source_dir = NULL
;

static int copy_rename_single_file(const char *fname, const char *source_dir, const char *target_dir);
static int copy_rename_many_files (const char *source_dir, const char *target_dir);
static int duplicate_single_file(const char *fname, const char *source_dir);
static int duplicate_many_files (const char *source_dir);
int do_fork_duplicate   (const char *full_source_fname, const char *full_target_fname);

void DuplicateButtonCB(FL_OBJECT *, long){
	if (toolbox_tree_flag) {
		tree_duplicate();
		return;
	}
    copy_flag = FALSE;
    source_dir_action(duplicate_single_file, duplicate_many_files, "Duplicate", TRUE, TRUE, duplicate_general_help_text);
}

void CopyAndRenameCB(FL_OBJECT *, long){
    copy_flag = TRUE;
    source_and_target_dir_action(copy_rename_single_file, copy_rename_many_files, "Copy and rename", TRUE, copy_rename_help_text);
}

static int duplicate_do_each(const char *fname, int &counter);

static int duplicate_many_files(const char *source_dir){
    if(!source_dir || !source_browser) return 0;
    loop_flag                  = TRUE;
    loop_ask_directory_flag    = TRUE;
    loop_ask_overwrite_flag    = TRUE;       // always, for the time being
    loop_ask_writeprotect_flag = TRUE;       // always
    loop_source_dir            = source_dir;
    int retval = source_browser->selection_action_loop(duplicate_do_each);
    OKCancelClose();
    return retval;
}

static int copy_rename_many_files(const char *source_dir, const char *target_dir){
    if(!source_dir || !source_browser || !target_dir || !target_browser) return 0;
    loop_flag                  = TRUE;
    loop_ask_directory_flag    = TRUE;
    loop_ask_overwrite_flag    = TRUE;       // always, for the time being
    loop_ask_writeprotect_flag = TRUE;       // always
    loop_source_dir            = source_dir;
    loop_target_dir            = target_dir;
    int retval = source_browser->selection_action_loop(duplicate_do_each);
    OKCancelClose();
    return retval;
}

static int duplicate_do_each(const char *fname, int &counter){
    if(!fname){
	return FALSE;
    } else {
	//cerr << "duplicate do each works on " << fname << endl;
    }

    char buffer[1024], full_source_fname[1024], full_target_fname[1024];
    struct stat source_file_statbuf, target_file_statbuf;

    // can't duplicate '.' or '..' :
    if(!dot_dotdot_test(fname)){
	timed_error_notifier("Can't duplicate `.' or `..'.");
	return FALSE;
    }
	
    // construct full name
    strcpy(full_source_fname, loop_source_dir);  
    strcat(full_source_fname, "/");
    strcat(full_source_fname, fname);

    /* does the source file exist ?
     * if it doesn't, it is too important an error to
     * continue blindly, without user interaction.
     */
    
    if(lstat(full_source_fname, &source_file_statbuf) != 0){
	sprintf(buffer, "Couldn't get lstat info for file `%s'.", fname);
	timed_strerror_notifier(buffer, notifier_secs);
	source_browser->update();
	return FALSE;
    }
    
    if(copy_flag){
        strcpy(full_target_fname, loop_target_dir);  // construct full target
        strcat(full_target_fname, "/");
        strcat(full_target_fname, fname);
    }

    /* get a new name */
    const char *result;
    if(copy_flag){
	sprintf(buffer, "Copying and renaming `%s' to:", full_source_fname);
	int begin_range, end_range;
	begin_range = strlen(loop_target_dir) + 1;
	end_range = begin_range + strlen(fname);
	result = OKCancel(buffer, full_target_fname, copy_rename_help_text, "Copy & Rename", loop_flag, TRUE, begin_range, end_range);
    } else {
	sprintf(buffer, "Duplicating `%s' to:", fname);
	result = OKCancel(buffer, fname, duplicate_help_text, "Duplicate", loop_flag);
    }
    // empty ?
    if(!result){
	return FALSE;
    }

    if(copy_flag){
	strcpy(full_target_fname, result);
    } else {
	sprintf(full_target_fname, "%s/%s", loop_source_dir, result);
    }

    // same name on simple rename ? (doesn't hurt on move);
    if(!copy_flag && strcmp(result, fname) == 0) return TRUE;

    /* does the target exist ? Are source and target the same ?
     * Is the target a directory ? Should we overwrite ?
     * Is it writeprotected, and should we override ?
     * we then ask for confirmation if the respective flag is set.
     */
    if(lstat(full_target_fname, &target_file_statbuf) == 0){
	if(source_file_statbuf.st_dev == target_file_statbuf.st_dev &&
	   source_file_statbuf.st_ino == target_file_statbuf.st_ino){
	    sprintf(buffer, "Source and target files are identical - not duplicated.");
	    timed_error_notifier(buffer, notifier_secs);
	    return TRUE;
	} else if(S_ISDIR(target_file_statbuf.st_mode)){
	    sprintf(buffer, "Target file is a directory, can't remove it - not duplicated.");
	    timed_error_notifier(buffer, notifier_secs);
	    return TRUE;
	} else if(loop_ask_overwrite_flag){
	    sprintf(buffer, "Target file `%s' exists. Try to overwrite ?", get_basename(full_target_fname));
	    if(loop_flag){
		switch(ask_multi_question(TRUE, 1, buffer, 
		    "Cancel All", "Cc", NoSymbol, "Cancel File", "Ff", NoSymbol, "Do Overwrite", "Dd", NoSymbol, "Overwrite ALL", "Aa"
	        )){
		case -1:
		case 0:
		    //cerr << "overwrite cancel all." << endl;
		    return FALSE;
		case 1:
		    //cerr << "overwrite cancel this file." << endl;
		    return TRUE;
		case 2:
		    //cerr << "overwrite do file" << endl;
		    break;
		case 3:
		    //cerr << "overwrite do all." << endl;
		    loop_ask_overwrite_flag = FALSE;
		    break;
		default:
		    program_unstable("Strange return from switch(action_ctrl)");
		    return FALSE;
		}
	    } else {
		if(!ask_question(TRUE, buffer)){
		    //cerr << "don't overwrite" << endl;
		    return TRUE;
		} else {
		    //cerr << "overwrite" << endl;
		}
	    }
	}
	if(loop_ask_writeprotect_flag){
	    if(!try_access(&target_file_statbuf, W_OK) && 
	       !S_ISLNK(target_file_statbuf.st_mode)){
		sprintf(buffer, "Target file `%s' is write protected. Force overwrite ?", fname);
		if(loop_flag){
		    switch(ask_multi_question(TRUE, 1, buffer,
		        "Cancel All", "Cc", NoSymbol, "Cancel File", "Ff", NoSymbol, "Do Force", "Dd", NoSymbol, "Force ALL", "Aa"
		    )){
		    case -1:
		    case 0:
			//cerr << "force overwrite cancel all." << endl;
			return FALSE;
		    case 1:
			//cerr << "force overwrite cancel this file." << endl;
			return TRUE;
		    case 2:
			//cerr << "force overwrite do file" << endl;
			break;
		    case 3:
			//cerr << "force overwrite do all." << endl;
			loop_ask_writeprotect_flag = FALSE;
			break;
		    default:
			program_unstable("Strange return from switch(action_ctrl)");
			return FALSE;
		    }
		} else {
		    if(!ask_question(TRUE, buffer)){
			//cerr << "don't force overwrite" << endl;
			return TRUE;
		    } else {
			//cerr << "force overwrite" << endl;
		    }
		}		    
	    }
	}
	// next comes the else-clause for the target lstat call.
	// we double check that it was the desired error (file didn't exist).
    } else if(errno != ENOENT){
	// lstat failed _not_ because target didn't exist
	sprintf(buffer, "lstat(%s) syscall failed.", full_target_fname);
	timed_strerror_notifier(buffer, notifier_secs);
	return TRUE;
    }

    // check wether we're about to duplicate a complete directory tree
    int n_files_in_dir = 0;
    if(loop_ask_directory_flag){
	if(S_ISDIR(source_file_statbuf.st_mode)){
	    n_files_in_dir = files_in_dir(fname);
	    if(n_files_in_dir != 2){
		sprintf(buffer, "File `%s' is a directory (%d files). Duplicate complete tree ?", fname, n_files_in_dir);
		if(loop_flag){
		    switch(ask_multi_question(FALSE, 2, buffer, 
                        "Cancel All", "Cc", NoSymbol, "Cancel File", "Ff", NoSymbol, "Do Duplicate Tree", "Dd", NoSymbol, "D. ALL Trees", "Aa"
                    )){
		    case -1:
		    case 0:
			//cerr << "ask directory cancel all." << endl;
			return FALSE;
		    case 1:
			//cerr << "ask directory cancel this file." << endl;
			return TRUE;   
		    case 2:
			//cerr << "ask directory do file" << endl;
			break;
		    case 3:
			//cerr << "ask directory do all." << endl;
			loop_ask_directory_flag = FALSE;
			break;
		    default:
			program_unstable("Strange return from switch(action_ctrl)");
			return FALSE;
		    }
		} else {
		    if(!ask_question(FALSE, buffer)){
			//cerr << "don't duplicate tree" << endl;
			return TRUE;
		    } else {
			//cerr << "do duplicate tree" << endl;
		    }
		}
	    }
	}
    }
    //cerr << "deselect .. duplicating file " << fname << endl;

    int retval = do_fork_duplicate(full_source_fname, full_target_fname);

    if(retval){
	if(loop_flag){
	    source_browser->deselect_file(fname);
	}
	counter++;
	if(copy_flag){
	    remember_source_current(get_basename(full_source_fname));
	    remember_target_current(get_basename(full_target_fname));
	} else {
	    remember_source_current(get_basename(full_target_fname));
	}
	action_step();
    } else {
	//cerr << "do_fork_duplicate failed." << endl;
    }    
    return retval;
}  

static int duplicate_single_file(const char *fname, const char *source_dir){
    if(!fname || !source_dir) return FALSE;
    int counter = 0;          
    loop_flag = FALSE;
    loop_ask_directory_flag    = TRUE;
    loop_ask_overwrite_flag    = TRUE;       // always, for the time being
    loop_ask_writeprotect_flag = TRUE;       // always
    loop_source_dir            = source_dir;
    action_work_noform();
    duplicate_do_each(fname, counter);
    OKCancelClose();
    return counter;
}

static int copy_rename_single_file(const char *fname, const char *source_dir, const char *target_dir){
    if(!fname || !source_dir || !target_dir) return FALSE;
    int counter = 0;          
    loop_flag = FALSE;
    loop_ask_directory_flag    = TRUE;
    loop_ask_overwrite_flag    = TRUE;       // always, for the time being
    loop_ask_writeprotect_flag = TRUE;       // always
    loop_source_dir            = source_dir;
    loop_target_dir            = target_dir;
    action_work_noform();
    duplicate_do_each(fname, counter);
    OKCancelClose();
    return counter;
}

int do_fork_duplicate(const char *full_source_fname, const char *full_target_fname){
    if(!full_source_fname ||  !full_target_fname){
	program_unstable("UC internal: do_fork_duplicate null ptrs.");
	return FALSE;
    }

    // we can do -r: on files, it doesn't matter; on dirs, we asked.
    static const char *dup_argv[5] = {"cp", "-r", NULL, NULL, NULL};
    dup_argv[1] = tool_cp_minusp_flag ? "-rp" : "-r";
    dup_argv[2] = full_source_fname;
    dup_argv[3] = full_target_fname;
    return do_fork_cmd(cp_path, dup_argv);
}

