//#include <iostream.h>
#include <stdio.h>
#include <string.h>
#include <sys/errno.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <unistd.h>

#include "browser.h"
#include "error_notifier_forms_aux.h"
#include "help_forms_aux.h"
#include "file_utility.h"
#include "fork_pipe.h"
#include "global.h"
#include "help_text.h"
#include "pathnames.h"
#include "tree_forms_aux.h"
#include "utility.h"

#include "action_forms.h"
#include "action_forms_aux.h"
#include "options_forms.h"
#include "toolbox_forms_aux.h"

/*
 * all functions/callbacks associated with the rename op.
 *
 * This version uses the most advanced general concept.
 * it uses source_dir_action for the loop through 1/many setup
 *         one unified testing/access function, switched by loop_flag
 *         selection_action_loop from browser_c::
 *
 * note that we use lstat, not stat, for rename.
 * note as well that we only have one source directory,
 * so this is not a binary source/target operation like move.
 * now, in the newest version, we also have full move and rename in a
 * unified function, which is binary.
 */

static int move_rename_single_file(const char *fname, const char *source_dir, const char *target_dir);
static int move_rename_many_files (const char *source_dir, const char *target_dir);
static int rename_single_file(const char *fname, const char *source_dir);
static int rename_many_files (const char *source_dir);
static int do_rename         (const char *full_source_fname, const char *full_target_fname);

static int
    move_flag           = FALSE,  // allow for full rename_moves.
    rename_verbose_flag = TRUE
;

static int
    loop_flag                  = FALSE,   // are we in a loop ?
    loop_ask_overwrite_flag    = TRUE,
    loop_ask_writeprotect_flag = TRUE
;

static const char
    *loop_source_dir = NULL,
    *loop_target_dir = NULL
;


void RenameButtonCB(FL_OBJECT *, long){
	if (toolbox_tree_flag) {
		tree_rename();
		return;
	}
    move_flag = FALSE;
    source_dir_action(rename_single_file, rename_many_files, "Rename", FALSE, TRUE, rename_general_help_text);
}

void MoveAndRenameCB(FL_OBJECT *, long){
    move_flag = TRUE;
    source_and_target_dir_action(move_rename_single_file, move_rename_many_files, "Move and rename", FALSE, move_rename_help_text);
}

static int rename_do_each(const char *fname, int &counter);

static int move_rename_many_files(const char *source_dir, const char *target_dir){
    if(!source_dir || !source_browser || !target_dir) return 0;
    loop_flag                  = TRUE;
    loop_ask_overwrite_flag    = TRUE;       // always, for the time being
    loop_ask_writeprotect_flag = TRUE;       // always
    loop_source_dir            = source_dir;
    loop_target_dir            = target_dir;
    int retval = source_browser->selection_action_loop(rename_do_each);
    OKCancelClose();
    return retval;
}

static int rename_many_files(const char *source_dir){
    if(!source_dir || !source_browser) return 0;
    loop_flag                  = TRUE;
    loop_ask_overwrite_flag    = TRUE;       // always, for the time being
    loop_ask_writeprotect_flag = TRUE;       // always
    loop_source_dir            = source_dir;
    int retval = source_browser->selection_action_loop(rename_do_each);
    OKCancelClose();
    return retval;
}

static int move_rename_single_file(const char *fname, const char *source_dir, const char *target_dir){
    if(!fname || !source_dir || !target_dir) return FALSE;
    int counter = 0;          
    loop_flag = FALSE;
    loop_ask_overwrite_flag    = TRUE;       // always, for the time being
    loop_ask_writeprotect_flag = TRUE;       // always
    loop_source_dir            = source_dir;
    loop_target_dir            = target_dir;
    action_work_noform();
    rename_do_each(fname, counter);
    OKCancelClose();
    return counter;
}

static int rename_do_each(const char *fname, int &counter){
    if(!fname){
	return FALSE;
    } else {
	//cerr << "rename do each works on " << fname << endl;
    }

    char buffer[1024], full_source_fname[1024], full_target_fname[1024];
    struct stat source_file_statbuf, target_file_statbuf;
    // can't rename '.' or '..' :
    if(!dot_dotdot_test(fname)){
	timed_error_notifier("Can't rename `.' or `..'.");
	return FALSE;
    }
	
    /* just go, until the first difficult one is hit. Then, show
     * the aux form with extra buttons. Cancel All stops the complete
     * operation, cancel file this file. Go confirms this file, Go all
     * confirms this particular operation/issue (e.g overwrite, directory)
     * for all items. Thus, Go all can be called for different ones, and
     * it's label needs to be set accordingly.
     */

    strcpy(full_source_fname, loop_source_dir);  // construct full name
    strcat(full_source_fname, "/");
    strcat(full_source_fname, fname);

    if(move_flag){
        strcpy(full_target_fname, loop_target_dir);  // construct full targt
        strcat(full_target_fname, "/");
        strcat(full_target_fname, fname);
    }

    /* does the source file exist ?
     * if it doesn't, it is too important an error to
     * continue blindly, without user interaction.
     */
    
    if(lstat(full_source_fname, &source_file_statbuf) != 0){
	sprintf(buffer, "Couldn't get lstat info for file `%s'.", fname);
	timed_strerror_notifier(buffer, notifier_secs);
	source_browser->update();
	return FALSE;
    }
    
    /* get a new name */
    const char *result;
    if(move_flag){
	sprintf(buffer, "Moving and renaming `%s' to:", full_source_fname);
	int begin_range, end_range;
	begin_range = strlen(loop_target_dir) + 1;
	end_range = begin_range + strlen(fname);
	result = OKCancel(buffer, full_target_fname, move_rename_help_text, "Move & Rename", TRUE, TRUE, begin_range, end_range);
    } else {
	sprintf(buffer, "Renaming `%s' to:", fname);
	result = OKCancel(buffer, fname, rename_help_text, "Rename", TRUE);
    }
    // empty ?
    if(!result){
	return FALSE;
    }

    if(move_flag){
	strcpy(full_target_fname, result);
    } else {
	sprintf(full_target_fname, "%s/%s", loop_source_dir, result);
    }

    // same name on simple rename ?
    if(!move_flag && strcmp(result, fname) == 0) return TRUE;
    
    /* does the target exist ? Are source and target the same ?
     * Is the target a directory ? Should we overwrite ?
     * Is it writeprotected, and should we override ?
     * we then ask for confirmation if the respective flag is set.
     */
    if(lstat(full_target_fname, &target_file_statbuf) == 0){
	if(source_file_statbuf.st_dev == target_file_statbuf.st_dev &&
	   source_file_statbuf.st_ino == target_file_statbuf.st_ino){
	    sprintf(buffer, "Source and target files are identical - not renamed.");
	    timed_error_notifier(buffer, notifier_secs);
	    return TRUE;
	} else if(S_ISDIR(target_file_statbuf.st_mode)){
	    sprintf(buffer, "Target file is a directory, can't remove it - not renamed.");
	    timed_error_notifier(buffer, notifier_secs);
	    return TRUE;
	} else if(loop_ask_overwrite_flag){
	    sprintf(buffer, "Target file `%s' exists. Try to overwrite ?", get_basename(full_target_fname));
	    if(loop_flag){
		switch(ask_multi_question(TRUE,1,buffer,
		    "Cancel All", "Cc", NoSymbol, "Cancel File", "Ff", NoSymbol, "Do Overwrite", "Dd", NoSymbol, "Overwrite ALL", "Aa"
	        )){
		case -1:
		case 0:
		    //cerr << "overwrite cancel all." << endl;
		    return FALSE;
		case 1:
		    //cerr << "overwrite cancel this file." << endl;
		    return TRUE;
		case 2:
		    //cerr << "overwrite do file" << endl;
		    break;
		case 3:
		    //cerr << "overwrite do all." << endl;
		    loop_ask_overwrite_flag = FALSE;
		    break;
		default:
		    program_unstable("Strange return from switch(multi)");
		    return FALSE;
		}
	    } else {
		if(!ask_question(TRUE, buffer)){
		    //cerr << "don't overwrite" << endl;
		    return TRUE;
		} else {
		    //cerr << "overwrite" << endl;
		}
	    }
	}
	if(loop_ask_writeprotect_flag){
	    if(!try_access(&target_file_statbuf, W_OK) && 
	       !S_ISLNK(target_file_statbuf.st_mode)){
		sprintf(buffer, "Target file `%s' is write protected. Force overwrite ?", fname);
		if(loop_flag){
		    switch(ask_multi_question(TRUE, 1, buffer, 
		        "Cancel All", "Cc", NoSymbol, "Cancel File", "Ff", NoSymbol, "Do Force", "Dd", NoSymbol, "Force ALL", "Aa"
		    )){
		    case -1:
		    case 0:
			return FALSE;
		    case 1:
			return TRUE;
		    case 2:
			//cerr << "force overwrite do file" << endl;
			break;
		    case 3:
			//cerr << "force overwrite do all." << endl;
			loop_ask_writeprotect_flag = FALSE;
			break;
		    default:
			program_unstable("Strange return from switch(action_ctrl)");
			return FALSE;
		    }
		} else {
		    if(!ask_question(TRUE, buffer)){
			//cerr << "don't force overwrite" << endl;
			return TRUE;
		    } else {
			//cerr << "force overwrite" << endl;
		    }
		}		    
	    }
	}
	// next comes the else-clause for the target lstat call.
	// we double check that it was the desired error (file didn't exist).
    } else if(errno != ENOENT){
	// lstat failed _not_ because target didn't exist
	sprintf(buffer, "lstat(`%s') syscall failed.", full_target_fname);
	timed_strerror_notifier(buffer, notifier_secs);
	return FALSE;
    }
	
    //cerr << "deselect .. renaming file " << fname << endl;
    if(do_rename(full_source_fname, full_target_fname)){
	if(S_ISDIR(source_file_statbuf.st_mode)){
	    tree_rename_path(full_source_fname, full_target_fname);
	}
	if(rename_verbose_flag){
	    //cerr << "Rename of file " << fname << " succeeded." << endl;
	}
	if(loop_flag){
	    source_browser->deselect_file(fname);
	}
	counter++;
	if(move_flag){
	    remember_target_current(get_basename(full_target_fname));
	} else {
	    remember_source_current(get_basename(full_target_fname));
	}
	action_step();
    } else {
	//cerr << "do_rename failed." << endl;
	return FALSE;
    }
    return TRUE;
}  

// does renameing + extra interaction.
static int rename_single_file(const char *fname, const char *source_dir){
    if(!fname || !source_dir) return FALSE;
    int counter = 0;          
    loop_flag = FALSE;
    loop_ask_overwrite_flag    = TRUE;       // always, for the time being
    loop_ask_writeprotect_flag = TRUE;       // always
    loop_source_dir            = source_dir;
    action_work_noform();
    rename_do_each(fname, counter);
    OKCancelClose();
    return counter;
}

static int do_rename(const char *full_source_fname, const char *full_target_fname){
    char buffer[2048];
    if(rename(full_source_fname, full_target_fname) == 0){
	return TRUE;
    }
    if(errno == EXDEV){
	// this takes long anyway, because it involves a copy. 
	static const char *mv_argv[4] = {"mv", NULL, NULL, NULL};
	mv_argv[0] = "mv";
        mv_argv[1] = full_source_fname;
        mv_argv[2] = full_target_fname;
	mv_argv[3] = NULL;
	if(do_fork_cmd(mv_path, mv_argv)){
	    return TRUE;
	} else {
	    timed_strerror_notifier("cross-filesystem rename failed.", notifier_secs);
	    return FALSE;
	}
    } else {
	sprintf(buffer, "Cannot rename `%s' to `%s'.", get_basename(full_source_fname), full_target_fname);
	timed_strerror_notifier(buffer, notifier_secs);
	return FALSE;
    } 
}








