
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <unistd.h>

#include "browser.h"
#include "error_notifier_forms_aux.h"
#include "file_utility.h"
#include "global.h"
#include "help_forms_aux.h"
#include "fork_pipe.h"
#include "status_forms_aux.h"
#include "tree_forms_aux.h"

int files_in_dir(const char *dir){
    if(!dir) return -1;
    DIR *dir_p = opendir(dir);
    if(!dir_p) return -1;
    int count = 0;
    while(readdir(dir_p) != NULL){
	count++;
    }
    closedir(dir_p);
    return count;
}

const char *dir_info_string(const char *dir){
    if(!dir) return "<no directory>";
    static char buffer[MAXPATHLEN+64];
    return buffer;
}

int try_access(const struct stat *statbuf_p, mode_t mode){
    uid_t euid;
    gid_t egid;
    
    unsigned long access_ok;
    // Clear bogobits. 
    mode &= (X_OK | W_OK | R_OK);	

    if (mode == F_OK){  		// The file exists.
	return TRUE;
    }
    
    euid = geteuid ();
    egid = getegid ();

    
    /* The super-user can read and write any file, and execute any file
     * that anyone can execute.
     */
    if(euid == 0 && ((mode & X_OK) == 0 || (statbuf_p->st_mode & 0111))){
	return TRUE;
    }
    if (euid == statbuf_p->st_uid){
	access_ok = (statbuf_p->st_mode & (mode << 6)) >> 6;
    } else if (egid == statbuf_p->st_gid){
	access_ok = (statbuf_p->st_mode & (mode << 3)) >> 3;
    } else {
	access_ok = (statbuf_p->st_mode & mode);
    }
    if(access_ok == mode){
	return TRUE;
    }
    errno = EACCES;
    return FALSE;
}

dir_validate_et validate_source_target_dirs(
    const char *source_dir, 
    const char *target_dir
){
    static struct stat target_dir_statbuf, source_dir_statbuf;
    if(!source_dir || !target_dir){
	program_unstable("UC internal: null ptrs in validate_source_target.");
	return error_e;
    }
    /* double-check the data we get. Of course, the directories are the
     * ones from living browsers, so should be readable, but this is 
     * quality software :-)
     * Hey Henrik, thanks for writing quality software, because I now use the
     * software differently and need these checks !
     */

    if(stat(source_dir, &source_dir_statbuf) != 0){
	timed_strerror_notifier("Couldn't get statinfo for source directory.", notifier_secs);
	return error_e;
    }
    if(!S_ISDIR(source_dir_statbuf.st_mode)){
	timed_error_notifier("specified source is not a directory.", source_dir, notifier_secs);
	return error_e;
    }
    if(stat(target_dir, &target_dir_statbuf) != 0){
	timed_strerror_notifier("Couldn't get statinfo for target directory.", notifier_secs);
	return error_e;
    }
    if(!S_ISDIR(target_dir_statbuf.st_mode)){
	timed_error_notifier("specified target is not a directory.", target_dir, notifier_secs);
	return error_e;
    }
    if(access(target_dir, W_OK | X_OK) != 0){
        timed_strerror_notifier("write/search permission denied for target directory.");		
	return error_e;
    }

    /* check device and inode */
    if(source_dir_statbuf.st_dev == target_dir_statbuf.st_dev &&
       source_dir_statbuf.st_ino == target_dir_statbuf.st_ino){
	return identical_e;
    }
    return ok_e;
}

int validate_source_dir(const char *source_dir){
    static struct stat source_dir_statbuf;
    if(!source_dir){
	program_unstable("UC internal: null ptrs in validate_source_dir.");
	return FALSE;
    }
#ifdef SUN
    if(stat((char *)source_dir, &source_dir_statbuf) != 0){
#else
    if(stat(source_dir, &source_dir_statbuf) != 0){
#endif
	timed_strerror_notifier("Couldn't get statinfo for source directory.", notifier_secs);
	return FALSE;
    }
    if(!S_ISDIR(source_dir_statbuf.st_mode)){
	timed_error_notifier("source directory string seems not to be directory.", notifier_secs);
	return FALSE;
    }
    return TRUE;
}

int validate_dir(const char *dir){
    static struct stat dir_statbuf;
    if(!dir){
	program_unstable("UC internal: null ptrs in validate_dir.");
	return FALSE;
    }
#ifdef SUN
    if(lstat((char *)dir, &dir_statbuf) != 0){
#else
    if(lstat(dir, &dir_statbuf) != 0){
#endif
	return FALSE;
    }
    if(!S_ISDIR(dir_statbuf.st_mode)){
	return FALSE;
    }
    return TRUE;
}

int dot_dotdot_test(const char *fname, const char *msg){
    char buffer[256];
    if(fname[0] == '.' && (fname[1] == '\0' || (fname[1] == '.' && fname[2] == '\0'))){
	sprintf(buffer, "%s: invalid argument: '%s'", msg, fname);
	timed_error_notifier(buffer, notifier_secs);
	return FALSE;
    } else {
	return TRUE;
    }
}

int dot_dotdot_check(const char *fname){
    if(fname[0] == '.' && (fname[1] == '\0' || (fname[1] == '.' && fname[2] == '\0'))){
	return FALSE;
    } else {
	return TRUE;
    }
}

const char *get_basename(const char *path){
    if(!path) return NULL;
    char *basename;
    if(strcmp(path, "/") == 0) return path;
    basename = strrchr(path, '/');
    return basename ? basename + 1 : path;
}

const char *get_dirname(const char *path){
    if(!path || path[0] != '/') return ".";
    static char dirpath[MAXPATHLEN];
    strcpy(dirpath, path);
    int len = strlen(dirpath);
    if(len <= 1) return dirpath;

    if(dirpath[len-1] == '/') return dirpath;
    char *last_slash = strrchr(dirpath, '/');
    *last_slash = '\0';
    return dirpath;
}

/* rename & rmdir raise EINVAL for paths ending in '/' */
const char *strip_trailing_slashes(const char *path){
    static char newpath[MAXPATHLEN];
    if(!path) return "/";
    strcpy(newpath, path);
    int last = strlen(newpath)-1;
    while(last>0 && newpath[last] == '/'){
	newpath[last] = '\0';
	last--;
    }
    return newpath;
}
    
int safe_chdir(const char *path){
    if(!path){
	errno = EFAULT;
	return -1;
    }
    if(access(path, R_OK) < 0){
	return -1;
    }
    return chdir(path);
}

/* char *real_path(const char *path, char resolved_path[MAXPATHLEN]);
 *
 * Find the real name of path, by removing all ".", ".." and symlink
 * components.  Returns (resolved) on success, or (NULL) on failure,
 * in which case the path which caused trouble is left in (resolved).
 * Cool: Always return absolute pathname !
 */

char *real_path(const char *path, char *resolved){
    struct stat sb;
    int fd, n, rootd, serrno;
    char *p, *q, wbuf[MAXPATHLEN], start_buf[MAXPATHLEN];

    /* Save the starting point. */
    if ((fd = open(".", O_RDONLY)) < 0) {
	strcpy(resolved, ".");
	return (NULL);
    }
    if(!getcwd(start_buf,MAXPATHLEN)){
      return NULL;
    }

    /*
     * Find the dirname and basename from the path to be resolved.
     * Change directory to the dirname component.
     * lstat the basename part.
     *     if it is a symlink, read in the value and loop.
     *     if it is a directory, then change to that directory.
     * get the current directory name and append the basename.
     */
    strncpy(resolved, path, MAXPATHLEN - 1);
    resolved[MAXPATHLEN - 1] = '\0';
 loop:
    q = strrchr(resolved, '/');
    if (q != NULL) {
	p = q + 1;
	if (q == resolved)
	    q = "/";
	else {
	    do {
		--q;
	    } while (q > resolved && *q == '/');
	    q[1] = '\0';
	    q = resolved;
	}
	if (chdir(q) < 0)
	    goto err1;
    } else
	p = resolved;

    /* Deal with the last component. */
    if (*p != '\0' && lstat(p, &sb) == 0) {
	if (S_ISLNK(sb.st_mode)) {
	    n = readlink(p, resolved, MAXPATHLEN);
	    if (n < 0)
		goto err1;
	    resolved[n] = '\0';
	    goto loop;
	}
	if (S_ISDIR(sb.st_mode)) {
	    if (chdir(p) < 0)
		goto err1;
	    p = "";
	}
    }

    /*
     * Save the last component name and get the full pathname of
     * the current directory.
     */
    strcpy(wbuf, p);
    if (getcwd(resolved, MAXPATHLEN) == 0)
	goto err1;

    /*
     * Join the two strings together, ensuring that the right thing
     * happens if the last component is empty, or the dirname is root.
     */
    if (resolved[0] == '/' && resolved[1] == '\0')
	rootd = 1;
    else
	rootd = 0;

    if (*wbuf) {
	if (strlen(resolved) + strlen(wbuf) + rootd + 1 > MAXPATHLEN) {
	    errno = ENAMETOOLONG;
	    goto err1;
	}
	if (rootd == 0)
	    strcat(resolved, "/");
	strcat(resolved, wbuf);
    }

    /* Go back to where we came from. */
    /*
    if (fchdir(fd) < 0) {
	serrno = errno;
	goto err2;
    }*/
    if(chdir(start_buf) < 0){
	serrno = errno;
	goto err2;
    }
    

    /* It's okay if the close fails, what's an fd more or less? */
    close(fd);
    return (resolved);

 err1:	serrno = errno;
//    fchdir(fd);
    chdir(start_buf);
 err2:	close(fd);
    errno = serrno;
    return (NULL);
}

int dir_is_child(const char *child, const char *ancestor){
    if(!child || !ancestor) return FALSE;
    char abuf[MAXPATHLEN], cbuf[MAXPATHLEN];
    if(real_path(child, cbuf) && real_path(ancestor, abuf)){
	if(strstr(cbuf, abuf)) return TRUE;
    }
    return FALSE;
}
