/***********************************************************************
 *
 * Module: fl_edit.c
 * Author: Marc van Kempen (marc@bowtie.nl)
 * Desc:   implementation of edit object 
 *
 * Copyright (c) 1996, Marc van Kempen
 *
 * All rights reserved.
 *
 * This software may be used, modified, copied, distributed, and
 * sold, in both source and binary form provided that the above
 * copyright and these terms are retained, verbatim, as the first
 * lines of this file.  Under no circumstances is the author
 * responsible for the proper functioning of this software, nor does
 * the author assume any responsibility for damages incurred with
 * its use.
 *
 ***********************************************************************/

/*       $Id: fl_edit.C,v 1.14 1997/09/07 17:56:36 henrik Exp henrik $
 */
 
#include <iostream.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <sys/param.h>
#include <sys/time.h>
#include <stdlib.h>
#include <unistd.h>
#include <forms.h>
#include "system.h"
#include "fl_edit.h"
#include "textbuf.h"
#include "fl_editscrollbar.h"
#include "error_notifier_forms_aux.h"
#include "gl_utility.h"
#include "help_forms_aux.h"

/***********************************************************************
 *
 * Some definitions 
 *
 ***********************************************************************/

/* a little macro for control characters */
#define ctrl(a) (a-'A'+1)

/***********************************************************************
 *
 * Textarea and coordinates
 * ------------------------
 *                             ob->x + ob->w - 
 *                          2 * ob->bw - sb_width
 *                                           |    +-> ob->x + ob->w
 *  ob->x                                    v    |
 *  +----------------------------------------+----+
 *  | x --+                                  |^^^^|
 *  |     |                                  |  | |
 *  .     v                                  .  +---> TEXT_SB_WIDTH
 *  .  (TEXT_X_OFFSET, TEXT_Y_OFFSET)        .    .
 *
 *  .                                        .    .
 *  |                                        |    |
 *  +----------------------------------------+----+
 *
 * Editscrollbar     
 * ---------
 *             +-----> TEXT_SB_WIDTH = ob->bw + sb_box_width + ob->bw
 *             |
 *       <--------->
 *  ..--+-----------+
 *      |           |
 *      |           |
 *      .           .
 *      .           .
 *
 ***********************************************************************/

/* define the offset of the text from the coordinates of its box */
#define TEXT_X_OFFSET		2
#define TEXT_Y_OFFSET		0
#define TEXT_SB_WIDTH		21

/* this does not seem to be defined in forms.h */
void fl_draw_object_label();

/* local prototypes */
static void fl_textedit_delete_word(FL_OBJECT *ob);
static void fl_textedit_pageup(FL_OBJECT *ob, int bExtendSelection = FALSE);
static void fl_textedit_pagedown(FL_OBJECT *ob, int bExtendSelection = FALSE);
static void fl_textedit_wordup(FL_OBJECT *ob, int bExtendSelection = FALSE);
static void fl_textedit_worddown(FL_OBJECT *ob, int bExtendSelection = FALSE);
static void fl_textedit_end_file(FL_OBJECT *ob, int bExtendSelection = FALSE);
static int  fl_textedit_set_cursor(FL_OBJECT *ob, int mx, int my, int draw);
static int  fl_textedit_selected(FL_OBJECT *ob);
static int  fl_textedit_movecursor(FL_OBJECT *ob, int r, int c);
static void fl_textedit_inssel(FL_OBJECT *ob, char *buf);
//static void fl_textedit_set_heditscrollbar(FL_OBJECT *ob);

static void good_redraw(FL_OBJECT *ob);
static void remove_other_selections(FL_OBJECT *ob);
static void textedit_eol(FL_OBJECT *ob, int bExtendSelection = FALSE);

// handle all editscrollbar setting. Mung topline if necessary.
// set editscrollbar parameters and max. Only force updates if 
// necessary. Meant to replace all other editscrollbar functions.
static void update_vsb(FL_OBJECT *ob);

// Mung leftcol if necessary. Return if redraw has happened.
static int update_hsb(
    FL_OBJECT *ob, 
    int bForceCursorVisible = FALSE,
    int bNoRedraw = FALSE
);

// makes a conditional update of the editscrollbar, but doesn't check
// for a new window configuration.

/* start a new selection or extend an existing one */
static void set_or_extend_selection(
    FL_OBJECT *ob, int old_r, int old_c, int bNoRedraw = FALSE
);

/* assume an existing selection */
static void fl_textedit_extend_selection(
    FL_OBJECT *ob, int old_r, int old_c, int bNoRedraw = FALSE
);

static void Pad5KeyFunc(FL_OBJECT *ob);

static void
delete_selected_block(FL_OBJECT *ob);

typedef enum {NoneE, DownE, UpE} SelectionDirectionType;
static SelectionDirectionType SelectionDirection = NoneE;

// big fonts get clipped without the extra scaling:
static double dblFontScale = 1.2;	

static int
fl_textedit_get_wsize(FL_OBJECT *ob, int *pLH = NULL)
/*
 * Desc: return the textwindow size (in lines)
 */
{
    int nlines, asc, desc, lh;
    SPEC *spec = (SPEC *)ob->spec;

    /* get the lineheight - could be more efficient... */
    lh = fl_get_char_height(spec->text_style, spec->text_size,
	&asc, &desc);
    
    lh = (int) ((double) lh * dblFontScale);
    nlines = (ob->h - 2*abs(ob->bw)) / lh;

    if (pLH) *pLH = lh;
    return(nlines);
} /* fl_textedit_get_wsize() */

static int
fl_textedit_get_csize(FL_OBJECT *ob)
/*
 * Desc: returns screen width in characters
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    int nchars = 0, cw;

    /* get character width */
    cw = fl_get_char_width(spec->text_style, spec->text_size);
    nchars = (ob->w - 2*abs(ob->bw) - 2*TEXT_X_OFFSET) / cw;
    if (nchars <= 0) {
	cerr << "Warning: Editor window has only " << nchars << "characters.\n";
	nchars = 1;
    }
    return(nchars);
}

/***********************************************************************
 *
 * Text rendering routines
 * 
 ***********************************************************************/

static void
fl_textedit_get_textbb(FL_OBJECT *ob, int *x, int *y, int *w, int *h)
/*
 * Desc: return the coordinates for the textarea
 */
{
    /* try to correct for single pix errors */
    *x = ob->x + abs(ob->bw) + TEXT_X_OFFSET + 1;
    *y = ob->y + abs(ob->bw) + TEXT_Y_OFFSET + 1;
    *w = ob->w - 2*abs(ob->bw) - TEXT_X_OFFSET;
    *h = ob->h - 2*abs(ob->bw) - TEXT_Y_OFFSET;
} /* fl_textedit_get_textbb() */

/* return a valid ptr to the line with all tabs expanded.
 * adjust nCur by the number of spaces it is offset.
 * nLeftCol allows working on relative nCurPos'itions. Use 0
 * as default.
 */
static char *expand_tabs(const char *pLine, int &nCurPos, int nLeftCol){
    static char *pLineBuf = NULL;
    static int nBufLen  = 0;
    const int 
	nBufStep = 256, 
	nLenExtra = 256, 
	nTabLen = 8
    ;
    int
	i,
	nMoved = 0,
	nPos,
	nSpaces
    ;

    int nLen = strlen(pLine);
    if (nLen > 16384) {
	fprintf(stderr, "UCEdit: Line too long.\n");
	return NULL;
    }

    int nTargetLen = nLen + nLenExtra + 1;

    if (nTargetLen > nBufLen) {
	if (pLineBuf) fl_free(pLineBuf);
	nBufLen = 0;
	pLineBuf = (char *) fl_malloc(nTargetLen);
	if (!pLineBuf) {
	    fprintf(stderr, "UCEdit: Malloc failed.\n");
	    return NULL;
	}
	nBufLen = nTargetLen;
    }
    strcpy(pLineBuf, pLine);

    char *p, *p1;
    p = pLineBuf;

    nCurPos += nLeftCol;	// effective position
    
    while ((p1 = strchr(p, '\t')) != NULL) {
	/* calculate TAB position */
	nPos = p1 - pLineBuf;
	/* calculate number of SPACEs to insert */
	if (nTabLen > nPos) {
	    nSpaces = nTabLen - nPos;
	} else {
	    nSpaces = nTabLen - (nPos % nTabLen);
	}
	nMoved += nSpaces; /* record how much spaces we added */
	nSpaces--;
	/* if only one SPACE then replace TAB with it and go on */
	if (nSpaces == 0) {
	    *p1 = ' ';
	    p1++;
	    p = p1;
	    continue;	
	}
	/* check if the spaces can be fitted in the buffer */
	if (nLen+nSpaces >= nBufLen) {
	    /* reallocate the buffer space and add a fixed amount of space 
	       to it */
	    char *pNew = (char *) realloc(pLineBuf, nBufLen+nBufStep);
	    if (!pNew) {
		fl_error_msg("expand_tabs(): Could not realloc, spaces not inserted");
		pLineBuf = NULL;
		nBufLen  = 0;
		return NULL;
	    }
	    pLineBuf = pNew;
	    nBufLen = nBufLen + nBufStep;
	}
	/* nLen+nSpaces < nBufLen !!! */

	/* first move the characters after pos spaces position */
	for (i=nLen; i>nPos; i--) {
	    pLineBuf[i+nSpaces] = pLineBuf[i];
	}
	/* insert spaces */
	for (i=nPos; i<=nPos+nSpaces;i++) {
	    pLineBuf[i] = ' ';
	}
	/* adjust strlen */
	nLen += nSpaces;

	/* continue the search */
	p = pLineBuf + nPos + 1;
	//cerr << "nPos: " << nPos << ", nCurPos: " << nCurPos
	//    << ", nSpaces: " << nSpaces << endl;
	if (nCurPos > nPos) {
	    // adjust the cursor position as long as it is affected. 
	    // imagine tab at nPos == 0, nCurPosOrg == 1;
	    // cerr << "adding: " << nSpaces << endl;
	    nCurPos += nSpaces;	// the way the loop works
	}
    }
    //cerr << "line: `" << pLine << "'\n";
    //cerr << "expn: `" << pLineBuf << "'\n";
    nCurPos -= nLeftCol; // correct
    return pLineBuf;
}

// translate an offset in the expanded string into an offset into
// the non-expanded (tabbed) string.
int calculate_real_cpos (char *pLine, int nExpandedPos) {
    int nTestPos;
    int nLen = strlen(pLine);
    if (!pLine) return 0;
    if (nExpandedPos <= 0) return nExpandedPos;
    if (!strchr(pLine, '\t')) return nExpandedPos;	// quick check. 

    //cerr << "Expanded: " << nExpandedPos << endl;
    for (int i=1; i < nLen; i++) {
	nTestPos = i;
	expand_tabs(pLine, nTestPos, 0);
	//cerr << "i, nTestPos: " << i << ", " << nTestPos << endl;
	if (nExpandedPos <= nTestPos) return i;
    }
    return nLen;
}


static void
fl_textedit_draw_textline(int align, FL_Coord x, FL_Coord y, FL_Coord w, FL_Coord h,
    int fgcol, int bgcol, int size, int style, char *line, 
    int ccol, int cpos, int sels, int sele, int leftcol)
/*
 * Desc: Draw the line at the specified position with the alignment <align>
 *	 etc.
 *       Draw the selection if needed
 */
{
    int swidth1, swidth2, slinelen, slen, seltype;
    char *sline;

    int nSelEnd   = sele,
	nSelStart = sels
    ;
    //cerr << "Old s,e: " << nSelStart << ", " << nSelEnd;
    if (sels >= 0) {
	expand_tabs(line, nSelStart, 0);
	sels = nSelStart;
	//cerr << "NEW s: " << nSelStart << endl;
    }
    if (sele >= 0) {
	expand_tabs(line, nSelEnd, 0);
	////cerr << "NEW e: " << nSelEnd << endl;
	sele = nSelEnd;
    }

    int nCurPos = cpos;
    //cerr << "before: " << nCurPos << endl;
    char *pDrawLine = expand_tabs(line, nCurPos, 0);
    if (!pDrawLine) return;
    //cerr << "after:  " << nCurPos << endl;
    cpos = nCurPos;
    // bug: selection end, start and leftcol need to be 
    // adjusted, too. But only if end and start are >= 0;

    if (!pDrawLine) return;

    slen = strlen(pDrawLine);

    /* check how the selection should be drawn */
    if ((sels == 0) && ((sele == -1) || (sele > slen))) {
	seltype = 1;  /* the whole line is selected */
    } else {
	if ((sels == -1) || ((sele >= 0) && (sele <= sels))) {
	    seltype = -1; /* line is not selected */
	} else {
	    seltype = 0;  /* line is partially selected */
	    // bug: it still can be selected until the end of the line.
	}
    }
    //cerr << ", seltype: " << seltype << endl;

    /* line is invisible */
    if (leftcol >= slen) {
	if (seltype == 1) {
	    fl_rectf(x - TEXT_X_OFFSET, y, w + TEXT_X_OFFSET, h, fgcol);
	} else {
	    fl_rectf(x - TEXT_X_OFFSET, y, w + TEXT_X_OFFSET, h, bgcol);
	}
        /* only the cursor is visible */
        if (leftcol == cpos) {
	    //cerr << "cursor position only." << endl;
	    //cerr << "Done\n\n";
	    fl_drw_text_cursor(align, x, y, w, h, fgcol, style, size, "", ccol, 0);
	}
	return;
    }

    pDrawLine += leftcol;
    slen = strlen(pDrawLine);
    cpos -= leftcol;
    if (cpos < 0)
	cpos = -1;
    if (cpos > slen)
	cpos = slen;

    /* no selection */
    if (seltype == -1) {
	//cerr << "Whole line normal: '" << pDrawLine << "'" << endl;
	//cerr << "Done\n\n";
	fl_rectf(x - TEXT_X_OFFSET, y, w + TEXT_X_OFFSET, h, bgcol);
	fl_drw_text_cursor(align, x, y, w, h, fgcol, style, size, pDrawLine, ccol, cpos);
	return;	    					
    }

    /* the whole line is selected */
    if (seltype == 1)	{
	//cerr << "Whole line selected: '" << pDrawLine << "'" << endl;
	//cerr << "Done\n\n";
	fl_rectf(x - TEXT_X_OFFSET, y, w + TEXT_X_OFFSET, h, fgcol);
	fl_drw_text_cursor(align, x, y, w, h, bgcol, style, size, pDrawLine, ccol, cpos);
	return;	    
    } else {
	//cerr << "Partial line selected: " << endl;
    }


    if (sels >= 0) {
	sels -= leftcol;
	if (sels < 0)
	    sels = 0;
    }

    if (sele >= 0) {
	sele -= leftcol;
	if (sele < 0)
	    sele = sels;
	if (sele > slen)
	    sele = slen;
    } else {
	sele = slen;
    }

    /* draw the whole non-selected line */
    fl_rectf(x - TEXT_X_OFFSET, y, w + TEXT_X_OFFSET, h, bgcol);
    //cerr << "drawing whole non-selected part: '" << pDrawLine << "'" << endl;
    fl_drw_text_cursor(align, x, y, w, h, fgcol, style, size, pDrawLine, ccol, cpos);

    /* draw the selected part */
    if (sele <= sels) {
	if (nSelEnd < 0) {
	    //cerr << "include end of line.\n";
	    swidth1 = fl_get_string_width(style, size, pDrawLine, sels);
	    swidth2 = w - swidth1;
	    if (swidth2 > 0) {
		fl_rectf(x - TEXT_X_OFFSET + swidth1, y, swidth2, h, fgcol);
	    }
	}
	return;
    }

    slinelen = sele - sels;
    sline = (char *)malloc(slinelen + 1);
    strncpy(sline, pDrawLine + sels, slinelen);
    sline[slinelen] = '\0';

    //cerr << "drawing selected part: '" << sline << "'" << endl;

    swidth1 = fl_get_string_width(style, size, pDrawLine, sels);
    if (swidth1 > w) {
	free(sline);
	return;
    }

    swidth2 = fl_get_string_width(style, size, sline, slinelen);
    if ((swidth1 + swidth2) > w) {
	swidth2 = w - swidth1;
    }

    // special case: line is partially selected, but up to the end.
    // this is indicated by sele == -1
    int swidth3 = swidth2;
    if (nSelEnd < 0) {
	//cerr << "Special case, include end of line.\n";
	swidth3 = w - swidth1;
    }

    if (sels) {
	fl_rectf(x + swidth1, y, swidth3, h, fgcol);
    } else {
	fl_rectf(x + swidth1 - TEXT_X_OFFSET, y, swidth3, h, fgcol);
    }

    fl_drw_text_cursor(align, 
	x + swidth1, y, swidth2, h, 
	bgcol, style, size, sline, ccol,
	((cpos >= sels) && (cpos <= (sels + slinelen))) ? (cpos - sels) : -1
    );

    XFlush(fl_display);
    //sleep(3);
    //cerr << "Done\n\n";
    free(sline);
} /* fl_textedit_draw_textline() */

static void
fl_textedit_draw_screen(FL_OBJECT *ob)
/*
 * Desc: refresh the screen
 */
{
    int i, w, h, eof, pos, screenlines, lh, sels, sele, linenum;
    int tx, ty, tw, th;		/* textarea coordinates */
    char *line;
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    TextLine *tl;

    if (ob->form->frozen)
	return;

    Window wn = fl_winget();
    fl_winset(ob->form->window);

    /* get the lineheight */
    
    screenlines = fl_textedit_get_wsize(ob, &lh);
    h = lh;

    w = fl_get_char_width(spec->text_style, spec->text_size);

    fl_textedit_get_textbb(ob, &tx, &ty, &tw, &th);
    /*
    fprintf(stderr, "ob->h: %d, tx %d, ty %d, tw %d, th: %d\n", ob->h, 
	tx, ty, tw, th
    );*/

    /* draw the box on the screen */
//    fl_set_text_clipping(tx, ty, tw, th); 

/* original drawing
    fl_drw_frame(ob->boxtype, ob->x + abs(ob->bw), ob->y + abs(ob->bw), tw + TEXT_X_OFFSET, 
	ob->h - 2*abs(ob->bw), ob->col2, abs(ob->bw));
	*/
    /* try to correct for single pix error */

    fl_drw_frame(ob->boxtype, 
	ob->x + abs(ob->bw) + 1, 
	ob->y + abs(ob->bw) + 1, 
	tw + TEXT_X_OFFSET - 2, 
	ob->h - 2*abs(ob->bw) - 2, 
	ob->col2, abs(ob->bw)
    );
    tw -= abs(ob->bw); // mung
    fl_set_text_clipping(tx, ty, tw, th); 
    
    /* draw the text */
    i = 0;
    tb_set_current_line(tb, spec->topline);
    eof = (tb->currentline == NULL);
    if (eof && ob->focus) {
	/* draw an empty line, so the cursor shows */
	fl_textedit_draw_textline(FL_ALIGN_LEFT, tx, ty, tw, h, 
	    ob->col1, ob->col2,
	    spec->text_size, spec->text_style, "", 
	    spec->ccol, 0, -1, -1, spec->leftcol);
	i++;
    }
    while (!eof && (i<screenlines)) {
	linenum = i + spec->topline;
	tb_get_line(tb, &line);
	fl_get_string_dimension(spec->text_style, spec->text_size,
	    line, strlen(line), &w, &h);
	h = (int) ((double) h * dblFontScale);
	if ((linenum == spec->r) && ob->focus) {
	    /* draw the cursor at the right position */
	    pos = spec->c;
	} else {
	    /* draw no cursor */
	    pos = -1;
	}

	sels = -1;
	sele = -1;
	/* calculate selection */
	if ((linenum >= spec->sselr) &&
	    (linenum <= spec->eselr))	{
	    //cerr << "line " << linenum << " may be selected: ";
	    sels = 0;
	    if (linenum == spec->sselr)
		sels = spec->sselc;

	    if (linenum == spec->eselr)
		sele = spec->eselc;
	    //cerr << sels << " -> " << sele << endl;
	}

	fl_textedit_draw_textline(FL_ALIGN_LEFT, tx, ty + i*h, tw, h, 
	    tb_get_linefgcolor(tb),
	    tb_get_linebgcolor(tb),
	    spec->text_size, spec->text_style, line, 
	    spec->ccol, pos, sels, sele, spec->leftcol);
	eof = !tb_set_next_line(tb);
	i++;
    }
    fl_unset_text_clipping(); 
/*
    fprintf(stderr, "i=%d, screenlines=%d, col2: %d\n", 
	i, screenlines, ob->col2
    );
    */
    lh = i*h;

    // this is a hack. 2*abs would be correct, but it doesn't draw
    // right. so mung it. Fill in the little remaining bar under the
    // text.
    fl_rectf(tx - TEXT_X_OFFSET, ty + lh, 
	tw+TEXT_X_OFFSET, ob->h - 3*abs(ob->bw) - lh , ob->col2);
	
    /* restore the current linenr */
    tb_set_current_line(tb, spec->r);

    tl = tb->firstline;
    while (tl)	{
	tl->flags &= ~TLINE_MODIFIED;
	tl = tl->next;
    }
    fl_winset(wn);
} /* fl_textedit_draw_screen() */

void
fl_textedit_draw_line(FL_OBJECT *ob, int n)
/*
 * Desc: redraw the n-th line
 * Pre:  n<tb_get_lines(ob->spec->tb)
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    TextLine *tl;
    int	nlines, tx, ty, tw, th;		/* bounding box of line */
    int x, y, w, h, sels, sele;

    if (ob->form->frozen)
	return;

    //cerr << "Draw line\n";

    /* check if n is valid */
    if (n<0 || n >= tb_get_nlines(tb)) {
	return;
    }

    nlines = fl_textedit_get_wsize(ob, &th);

    if ((n < spec->topline) || (n >= spec->topline+nlines)) {
	/* line is out of visible range */
	return;
    }

    /* get the line */
    if ((tl = tb_get_lineptr_by_num(tb, n)) == NULL)
	return;

    /* get bounding box of line */
    fl_textedit_get_textbb(ob, &x, &y, &w, &h);
    tw = w;
    tx = x;
    ty = y + (n - spec->topline) * th;
    tw -= abs(ob->bw);  // mung

    /* enable text clipping */
    fl_set_text_clipping(tx, ty, tw, th);

    /* erase the current line */
    fl_rectf(tx, ty, tw, th, ob->col2);

    sels = -1;
    sele = -1;
    /* calculate selection */
    if ((n >= spec->sselr) &&
	(n <= spec->eselr))	{
	sels = 0;
	if (n == spec->sselr)
	    sels = spec->sselc;

	if (n == spec->eselr) {
	    sele = spec->eselc;
	    if (sels == sele) {
		sels = -1;
		sele = -1;	  }
	}
    }

    if (spec->r == n) {
	/* draw a cursor */
	fl_textedit_draw_textline(FL_ALIGN_LEFT, tx, ty, tw, th, 
	    tl->fgcolor,
	    tl->bgcolor,
	    spec->text_size, 
	    spec->text_style,tl->buf, spec->ccol, spec->c,
	    sels, sele, spec->leftcol);
    } else {
	/* just draw the text */
	fl_textedit_draw_textline(FL_ALIGN_LEFT, tx, ty, tw, th,
	    tl->fgcolor,
	    tl->bgcolor,
	    spec->text_size, 
	    spec->text_style,tl->buf, spec->ccol, -1,
	    sels, sele, spec->leftcol);
    }
    
    fl_unset_text_clipping();
    tl->flags &= ~TLINE_MODIFIED;
} /* fl_textedit_draw_line() */

void
fl_textedit_refresh_screen(FL_OBJECT *ob, int bclear)
/*
 * Desc: update screen. Draw only lines that require update
 *       bclear != 0 means that if screen is not filled with text
 *       then we should "erase" the empty bottom part
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    TextLine *tl;
    int line, nlines, th;
    int x, y, w, h;

    if (ob->form->frozen)
	return;

    tl = tb->firstline;
    line = 0;
    while (tl) 	{
	if (!(tl->flags & TLINE_MODIFIED))	{
	    tl = tl->next;
	    line++;
	    continue;			}

	fl_textedit_draw_line(ob, line);
	tl->flags &= ~TLINE_MODIFIED;

	tl = tl->next;
	line++;
    }

    if (bclear)	{
	nlines = fl_textedit_get_wsize(ob, &th);
	if (tb->n >= (spec->topline + nlines))
	    return;
	line = tb->n - spec->topline;
	fl_textedit_get_textbb(ob, &x, &y, &w, &h);
	w -= abs(ob->bw); // mung
	y += (th * line);
	h -= (th * line);
	h -= abs(ob->bw);
	fl_rectf(x - TEXT_X_OFFSET, y, w + TEXT_X_OFFSET, h, ob->col2);
	if ((line == 0) && (spec->leftcol == 0)) 	{
	    w = fl_get_char_width(spec->text_style, spec->text_size);
	    fl_drw_text_cursor(FL_ALIGN_LEFT, x, y, w, th,ob->col1, spec->text_style, spec->text_size, "", spec->ccol, 0);
	}
    }
}

void
fl_textedit_draw_selection(FL_OBJECT *ob)
/*
 * Desc: redraw the selected text
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    int i;

    if ((spec->sselr < 0) ||
	(spec->eselr < 0))
	return;

    if (spec->sselr == spec->eselr)
	fl_textedit_draw_line(ob, spec->sselr);
    else {
	for (i = spec->sselr; i <= spec->eselr; i++)
	    fl_textedit_draw_line(ob, i);
    }
}

/* 0:  not found.
 * >0: found.
 * <0: wrapped found. 
 */
int
fl_textedit_search(FL_OBJECT *ob, int again)
/*
 * Desc: search for text
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf	 *tb = &spec->tb;
    char *p;
    int line, pos, n;

    if (again)	{
	if (!spec->exp)	{
	    fl_error_msg("No previous search pattern");
	    fl_winset(ob->form->window);
	    return 0;	
	} else {
	    // hack, if the function is called externally with again=1.
	    fl_winset(ob->form->window);
	}
    } else {
	p = (char *)fl_show_input("Search for", spec->exp ? spec->exp : "");
	fl_winset(ob->form->window);
	if (spec->exp)
	    free(spec->exp);
	if (!p || !strlen(p))	{
	    spec->exp = NULL;
	    return 0;		
	}
	spec->exp = strdup(p);
    }

    if ((line = tb_find_text(tb, spec->exp, spec->r, spec->c + 1, &pos)) == -1){
	if (tb_find_text(tb, spec->exp, 0, 0, &pos) >= 0) {
	    // wrap around possible.
	    return -1;
	} else {
	    return 0; 
	}
    }

    fl_textedit_remove_selection(ob);
    spec->sselr = spec->eselr = line;
    spec->sselc = pos;
    spec->eselc = pos + strlen(spec->exp);
    n = spec->r;
    spec->r = line;
    spec->c = pos;

    if (fl_textedit_line_visible(ob, line)) {
	if (!update_hsb(ob, TRUE)) {
	    fl_textedit_draw_line(ob, n);
	    fl_textedit_draw_selection(ob);
	}
    } else {
	// if possible, center the line.
	Pad5KeyFunc(ob);
    }
    return 1;
}

void
fl_textedit_heditscrollbar_sync_geometry(FL_OBJECT *ob)
{
    SPEC *spec = (SPEC *)ob->spec;

    spec->hsb->x = ob->x;
    spec->hsb->y = ob->y + ob->h;
    if (spec->sb->visible) {
	spec->hsb->w = ob->w; /* + TEXT_SB_WIDTH; */
    } else {
	spec->hsb->w = ob->w;
    }
    spec->hsb->h = TEXT_SB_WIDTH;
    spec->hsb->resize = ob->resize;
    spec->hsb->resize &= ~FL_RESIZE_Y;
    /*
    spec->hsb->segravity = ob->segravity;
    spec->hsb->nwgravity = ob->nwgravity;
    */
    spec->hsb->segravity = SouthEastGravity;
    spec->hsb->nwgravity = SouthWestGravity;
}

void
fl_textedit_veditscrollbar_sync_geometry(FL_OBJECT *ob)
{
    SPEC *spec = (SPEC *)ob->spec;

    spec->sb->x = ob->x + ob->w;
    spec->sb->y = ob->y;
    spec->sb->w = TEXT_SB_WIDTH;
    spec->sb->h = ob->h;
    spec->sb->resize = ob->resize;
    spec->sb->resize &= ~FL_RESIZE_X;
    /*
    spec->sb->segravity = ob->segravity;
    spec->sb->nwgravity = ob->nwgravity;
    */
    spec->sb->segravity = SouthEastGravity;
    spec->sb->nwgravity = NorthEastGravity;
}

int
fl_textedit_get_llength(FL_OBJECT *ob)
/*
 * Desc: returns length of longest displayed line
 *  screen size in characters is returned if every line fits on the screen
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf	 *tb = &spec->tb;
    TextLine *tl;
    int csize, wsize, llength;

    csize = fl_textedit_get_csize(ob);
    wsize = fl_textedit_get_wsize(ob);

    llength = csize;
    tl = tb->firstline;
    while (tl) {
	if (tl->strlen > llength)
	    llength = tl->strlen;
	tl = tl->next;
    }

    return llength;
}

void
fl_textedit_set_heditscrollbar_wsize(FL_OBJECT *ob)
/*
 * Desc: set heditscrollbar wsize value
 */
{
    update_hsb(ob);
}

/***********************************************************************
 *
 * Event handling
 *
 ***********************************************************************/

void
fl_textedit_scroll_with_mouse(FL_OBJECT *ob, int key)
/*
 * Desc: scroll the text while the mouse button <key> is being
 *	 held down
 */
{
    int		 mx, my, oldmx, oldmy, lastmy;
    unsigned int keymask;
    int 	 button, nlines, maxlines;
    SPEC	 *spec = (SPEC *)ob->spec;
    TextBuf	 *tb = &spec->tb;
    int 	 oldtopline = spec->topline;
    int		 lineskip = 4;

    button = Button1Mask;
    switch (key) {
    case 1: button = Button1Mask; break;
    case 2: button = Button2Mask; break;
    case 3: button = Button3Mask; break;
    }

    /* Query the pointer and move the text around */
    fl_get_mouse(&oldmx, &oldmy, &keymask);
    lastmy = oldmy;
    while (keymask & button) {
	fl_get_mouse(&mx, &my, &keymask);
	if (abs(lastmy-my) >= lineskip) {
	    nlines = (oldmy-my) / lineskip;
	    spec->topline = oldtopline + nlines;
	    maxlines = tb_get_nlines(&spec->tb);
	    if (spec->topline < 0) spec->topline = 0;
	    if (spec->topline >= maxlines) spec->topline = maxlines-1;
	    spec->r = spec->topline;
	    tb_set_current_line(tb, spec->r);
	    if (spec->c > tb_get_linelen(tb)) {
		spec->c = tb_get_linelen(tb);
	    }
	    update_vsb(ob);
	    fl_textedit_set_cursor(ob, mx, my, 1);
	    fl_textedit_draw_screen(ob);
	    lastmy = my;
	}
	else
	    fl_textedit_set_cursor(ob, mx, my, 1);
    }

    return;
} /* fl_textedit_scroll_with_mouse() */

void
fl_textedit_cleft(FL_OBJECT *ob)
/*
 * Desc: move the whole page one character left
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    if (spec->leftcol == 0)
	return;

    spec->leftcol--;
    spec->c = spec->leftcol;
    fl_textedit_draw_screen(ob);
    update_hsb(ob);
}

void
fl_textedit_cright(FL_OBJECT *ob, int bMungCursor = FALSE)
/*
 * Desc: move the whole page one character right
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    if (spec->leftcol >= fl_textedit_get_llength(ob) - fl_textedit_get_csize(ob)) {
	return;
    }

    spec->leftcol++;
    if (bMungCursor) {
	spec->c = spec->leftcol + fl_textedit_get_csize(ob);
    }
    fl_textedit_draw_screen(ob);
    update_hsb(ob);
}

void
fl_textedit_pagedown(FL_OBJECT *ob, int bExtendSelection){
/*
 * Desc: move the cursor one page down
 */
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int th, nlines;

    int old_r = spec->r, old_c = spec->c;

    nlines = fl_textedit_get_wsize(ob, &th);

    int new_max = tb->n - nlines;
    if (new_max < 0) new_max = 0;
    
    spec->c = spec->cpos;

    if (spec->r == tb_get_nlines(tb) - 1){
	spec->c = tb_get_linelen(tb); // go to last
    } 

    if (spec->r < tb_get_nlines(tb) - nlines) {
	spec->r = spec->r + nlines;
    } else {
	spec->r = tb_get_nlines(tb) - 1;
	if (spec->r < 0) spec->r = 0;
    }
    if (spec->topline < tb_get_nlines(tb) - nlines) {
	spec->topline = spec->topline + nlines;
    }
    if (spec->topline > new_max) {
	spec->topline = new_max;
    }
    
    tb_set_current_line(tb, spec->r);
    if (spec->c > tb_get_linelen(tb)) {
	spec->c = tb_get_linelen(tb);
    }

    if (bExtendSelection) {
	// no redraw, as we redraw the full screen anyway.
	set_or_extend_selection(ob, old_r, old_c, TRUE);
    }

    update_vsb(ob);
    update_hsb(ob, TRUE, FALSE);
    fl_textedit_draw_screen(ob);
} /* fl_textedit_pagedown() */

static void fl_textedit_pageup(FL_OBJECT *ob, int bExtendSelection){
/*
 * Desc; move the cursor one page up
 */
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int th, nlines;

    int old_r = spec->r, old_c = spec->c;

    nlines = fl_textedit_get_wsize(ob, &th);

    spec->c = spec->cpos;
    if (spec->r == 0) spec->c = 0; // go to first. 

    if (spec->r > nlines-1) {
	spec->r = spec->r - nlines;
    } else {
	spec->r = 0;
    }
    if (spec->topline > spec->r) {
	spec->topline = spec->r;
    }
    tb_set_current_line(tb, spec->r);

    if (spec->c > tb_get_linelen(tb)) {
	spec->c = tb_get_linelen(tb);
    }
    if (bExtendSelection) {
	// no redraw, as we redraw the full screen anyway.
	set_or_extend_selection(ob, old_r, old_c, TRUE);
    }
    update_vsb(ob);
    update_hsb(ob, TRUE, FALSE);
    fl_textedit_draw_screen(ob);
} /* fl_textedit_pageup() */

void
fl_textedit_home_file(FL_OBJECT *ob, int bExtendSelection)
/*
 * Desc; move the cursor to 0,0.
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    int old_r = spec->r, old_c = spec->c;

    if (fl_textedit_movecursor(ob, 0, 0) < 0) {
	spec->topline = 0;
	spec->leftcol = 0;
	spec->r = 0;
	spec->c = 0;
	if (bExtendSelection) {
	    // no redraw, as we redraw the full screen anyway.
	    set_or_extend_selection(ob, old_r, old_c, TRUE);
	}
	fl_textedit_draw_screen(ob);		    
    } else {
	if (bExtendSelection) {
	    // do redraw, as we just jumped with the cursor.
	    set_or_extend_selection(ob, old_r, old_c);
	}
    }

    update_vsb(ob);
    update_hsb(ob);
    spec->cpos = 0;
}

static void
fl_textedit_home_line(FL_OBJECT *ob, int bExtendSelection)
/*
 * Desc; move the cursor to line,0.
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    int old_r = spec->r, old_c = spec->c;
    /* Just Home */
    if (spec->leftcol) {
	spec->leftcol = 0;
	spec->c = 0;
	spec->cpos = 0;
	if (bExtendSelection) {
	    // no redraw, as we redraw the full screen anyway.
	    set_or_extend_selection(ob, old_r, old_c, TRUE);
	}
	fl_textedit_draw_screen(ob);
	update_hsb(ob);
    } else {
	fl_textedit_movecursor(ob, spec->r, 0);
	spec->cpos = 0;
	if (bExtendSelection) {
	    // do redraw, as we just jumped with the cursor.
	    set_or_extend_selection(ob, old_r, old_c);
	}
    }
}

static void
fl_textedit_end_file(FL_OBJECT *ob, int bExtendSelection)
/*
 * Desc; move the cursor to last line, last char.
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int wsize = fl_textedit_get_wsize(ob);
    int new_max = tb->n - wsize;
    if (new_max < 0) new_max = 0;
    int old_r = spec->r, old_c = spec->c;

    spec->r = tb_get_nlines(tb) - 1;
    if (spec->r < 0)
	spec->r = 0;
    tb_set_current_line(tb, spec->r);
    // was: spec->c = 0;
    spec->c = tb_get_linelen(tb);
    // was: spec->cpos = 0;
    spec->cpos = spec->c;
    spec->topline = new_max;
    spec->leftcol = 0;

    if (bExtendSelection) {
	// no redraw, as we redraw the full screen anyway.
	set_or_extend_selection(ob, old_r, old_c, TRUE);
    }
    update_vsb(ob);
    update_hsb(ob);

    fl_textedit_draw_screen(ob);
}

static void 
fl_textedit_linedown(FL_OBJECT *ob, int bExtendSelection = FALSE)
/*
 * Desc: move the cursor one line down
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int th, nlines;
    int old_r = spec->r, old_c = spec->c;

    nlines = fl_textedit_get_wsize(ob, &th);

    if ((spec->r < spec->topline+nlines-1) && (spec->r < tb_get_nlines(tb)-1)) {
	tb_set_next_line(tb);
	fl_textedit_movecursor(ob, 
	    spec->r + 1,
	    spec->cpos >= tb_get_linelen(tb) ? 
	    tb_get_linelen(tb) : spec->cpos
	);
	update_hsb(ob, TRUE);
	if (bExtendSelection) {
	    set_or_extend_selection(ob, old_r, old_c);
	}
    } else {
	/* scroll the text down by one line */
	spec->c = spec->cpos;
	if (spec->r < tb_get_nlines(tb)-1) {
	    spec->topline++;
	    spec->r++;
	    tb_set_next_line(tb);
	    if (spec->c >= tb_get_linelen(tb)){
		spec->c = tb_get_linelen(tb);
	    }
	    if (bExtendSelection) {
		set_or_extend_selection(ob, old_r, old_c, TRUE);
	    }
	    update_vsb(ob);
	    update_hsb(ob, TRUE, FALSE);
	    fl_textedit_draw_screen(ob);
	} else {
	    textedit_eol(ob, bExtendSelection);
	    update_hsb(ob, TRUE);
/*
	    spec->c = tb_get_linelen(tb);
	    if (bExtendSelection) {
		set_or_extend_selection(ob, old_r, old_c);
		}
		*/
	}
    }
} /* fl_textedit_linedown() */

static void
fl_textedit_lineup(FL_OBJECT *ob, int bExtendSelection = FALSE)
/*
 * Desc: move the cursor one line up
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int old_r = spec->r, old_c = spec->c;

    if (spec->r > spec->topline) {
	tb_set_prev_line(tb);
	fl_textedit_movecursor(ob, 
	    spec->r - 1,
	    spec->cpos >= tb_get_linelen(tb) ? 
	    tb_get_linelen(tb) : spec->cpos
	);
	update_hsb(ob, TRUE);
	if (bExtendSelection) {
	    set_or_extend_selection(ob, old_r, old_c);
	}
    } else {
	/* scroll the text up by one line */
	if (spec->topline > 0) {
	    spec->topline--;
	    spec->r--;
	    tb_set_prev_line(tb);
	    spec->c = spec->cpos;
	    if (spec->c >= tb_get_linelen(tb)) {
		spec->c = tb_get_linelen(tb);
	    }
	    if (bExtendSelection) {
		set_or_extend_selection(ob, old_r, old_c, TRUE);
	    }
	    update_vsb(ob);
	    update_hsb(ob, TRUE, FALSE);
	    fl_textedit_draw_screen(ob);
	} else {
	    fl_textedit_home_line(ob, bExtendSelection);
	}
    }
} /* fl_textedit_lineup() */

static void textedit_eol(FL_OBJECT *ob, int bExtendSelection){
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int old_r = spec->r, old_c = spec->c;

    if (fl_textedit_movecursor(ob, spec->r, tb_get_linelen(tb)) == -1) {
	spec->leftcol = tb_get_linelen(tb) - fl_textedit_get_csize(ob) + 1;
	update_hsb(ob);
	if (bExtendSelection) {
	    // no redraw, as we redraw the full screen anyway.
	    set_or_extend_selection(ob, old_r, old_c, TRUE);
	}
	fl_textedit_draw_screen(ob);	
    } else {
	if (bExtendSelection) {
	    // do redraw, as we just jumped with the cursor.
	    set_or_extend_selection(ob, old_r, old_c);
	}
    }
    spec->cpos = spec->c;
}

void fl_textedit_goto_line(FL_OBJECT *ob){
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    char buffer[64];

    sprintf(buffer, "%d", spec->r+1);

    const char *pLine = OKCancel("Go to line:", buffer);
    if (!pLine) return;
    int nLine = atoi(pLine);
    nLine--;
    int nLines = tb_get_nlines(tb);
    if (nLine >= nLines) nLine = nLines-1;
    if (nLine<0) nLine = 0;
    //cerr << nLine << endl;

    spec->r = nLine;
    Pad5KeyFunc(ob);
}

static void edit_alt_keys(FL_OBJECT *ob, int key){
    if (key != 'V' && key != 'B' && fl_textedit_selected(ob)) {
	fl_textedit_remove_selection(ob);
    }
    switch (key) {
    case 'b': fl_textedit_wordup(ob); 	return;
    case 'B': fl_textedit_wordup(ob, TRUE); 	return;
    case 'd': fl_textedit_delete_word(ob); 	return;
    case 'g': // fall-through.
    case 'G': fl_textedit_goto_line(ob); 	return;
    case 'v': fl_textedit_pageup(ob); 		return;
    case 'V': fl_textedit_pageup(ob, TRUE); 	return;
    case '<': fl_textedit_home_file  (ob); 		return;
    case '>': fl_textedit_end_file   (ob); 		return;
    }
}

static void edit_backspace_key(FL_OBJECT *ob){
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int n;
    if (spec->c > 0) {
	spec->c--;
	tb_del_char(tb, spec->c);
	fl_textedit_refresh_screen(ob, 1);
    } else {
	if (spec->r > 0) {
	    spec->r--;
	    tb_set_prev_line(tb);
	    spec->c = tb_get_linelen(tb);
	    tb_del_char(tb, spec->c);
	    if (spec->c > tb_get_linelen(tb))
		spec->c = tb_get_linelen(tb);
	    if (spec->r < spec->topline) {
		spec->topline--;
		n = tb_get_nlines(tb);
		update_vsb(ob);
	    }
	    fl_textedit_refresh_screen(ob, 1);
	    update_hsb(ob);
	}
    }
    spec->cpos = spec->c;
}

static void edit_charinsert_key(FL_OBJECT *ob, int key){
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int n = tb_insert_char(tb, spec->c, key);
    int nlines;
    /* line was wrapped */
    if (n == -1)   {
	/* cursor is at the end of line */
	nlines = tb_get_linelen(tb);
	if (spec->c >= nlines) {
	    n = spec->c;
	    fl_textedit_linedown(ob);
	    spec->c = n - nlines ? n - nlines : 1; 
	} else {
	    spec->c++;
	}
	
	update_vsb(ob);
	spec->cpos = spec->c;
	if (spec->r >= spec->topline + nlines) {
	    spec->topline++;
	    update_vsb(ob);
	}
	fl_textedit_refresh_screen(ob, 0);
    } else {
	/* advance the cursor */
	spec->c += n;
	spec->cpos = spec->c;

	int csize = fl_textedit_get_csize(ob);
	if (spec->c >= spec->leftcol + csize) {
	    update_hsb(ob, TRUE);
	} else {
	    fl_textedit_draw_line(ob, spec->r);
	    update_hsb(ob, FALSE);
	}
    }
}

static void edit_cr_key(FL_OBJECT *ob, int nlines){
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int n;
    /* insert a newline at position spec->c */
    tb_insert_cr(tb, spec->c);	
    n = tb_get_nlines(tb);
    spec->r++;
    spec->c = 0;
    spec->cpos = 0;
    tb_set_next_line(tb);
    if (spec->r >= (spec->topline + nlines)) {
	spec->leftcol = 0;
	spec->topline++;
	update_vsb(ob);
	fl_textedit_draw_screen(ob);
    } else {
	// adjust button only
	update_vsb(ob);
    }
    if (spec->leftcol) {
	spec->leftcol = 0;
	fl_textedit_draw_screen(ob); 
    } else {
	fl_textedit_refresh_screen(ob, 0);
    }
    fl_textedit_set_heditscrollbar_wsize(ob);
}

static void edit_delete_key(FL_OBJECT *ob, int bNoRedraw = FALSE){
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int n = tb_get_nlines(tb);
    if (tb_del_char(tb, spec->c)) {
	if (spec->c > tb_get_linelen(tb))
	    spec->c = tb_get_linelen(tb);
	if (!bNoRedraw) fl_textedit_refresh_screen(ob, 1);
	/* check if a line got deleted */
	if (n != tb_get_nlines(tb)) {
	    update_vsb(ob);
	}
    } else {
	if (!bNoRedraw) fl_textedit_draw_line(ob, spec->r);
    }
}

static void fl_textedit_delete_word(FL_OBJECT *ob){
    // eat all trailing whitespace after the word up to 
    // the next word or newline. If leading whitespace
    // is a newline, delete only that one.
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    char *line;
    tb_get_line(tb, &line);
    if (!line) return;
    int len=strlen(line);
    if (spec->c >= len) {
	// only a newline.
        edit_delete_key(ob);
	return;
    }
    int count=0, index=spec->c;
    while(index < len && (
	line[index] != ' '  && 
	line[index] != '\n' &&
	line[index] != '\r' &&
	line[index] != '\t'
    )){
        index++;
	count++;
    }
    // jump over trailing whitespace.
    while(index < len && (
	line[index] == ' ' ||
	line[index] == '\t'
    )){
        index++;
	count++;
    }
    for (int i=0; i<count-1; i++){
	edit_delete_key(ob, TRUE);
    }
    edit_delete_key(ob);
}

/* for ctrl-k :/) like emacs ! */

static char *pDelBuf = NULL;

static void addto_incremental_delbuf(FL_OBJECT *ob, char *pStr){
    if (!pStr) return;
    //cerr << "adding `" << pStr << "'\n";
    if (pDelBuf) {
	// append
	pDelBuf = (char *) realloc(pDelBuf, strlen(pDelBuf)+strlen(pStr)+1);
	if (!pDelBuf) return;
	strcat(pDelBuf, pStr);
    } else {
	// new
	pDelBuf = strdup(pStr);
	if (!pDelBuf) return;
    }
    fl_textedit_copy(ob, pDelBuf);
}

static void clean_incremental_delbuf(void){
    //cerr << "cleaning.\n";
    free (pDelBuf);
    pDelBuf = NULL;
}

//static void KeyFunc(FL_OBJECT *ob, SPEC *spec, XKeyEvent *xkev) {
//    if (!ob || !spec || !xkev) return;
//}

static void EndKeyFunc(FL_OBJECT *ob, XKeyEvent *xkev, int shift_flag) {
    if (!ob || !xkev) return;
    if (xkev->state & ControlMask) {
	/* Control-End was pressed, go to the end of the buffer */
	fl_textedit_end_file(ob, shift_flag); 
    } else {
	textedit_eol(ob, shift_flag);
    } 
}

static void HomeKeyFunc(FL_OBJECT *ob, XKeyEvent *xkev) {
    if (!ob || !xkev) return;
    int shift_flag = xkev->state & ShiftMask;
    if (xkev->state & ControlMask) {	    
	/* Control-Home was pressed, go to the beginning of the buffer */
	fl_textedit_home_file(ob, shift_flag);
    } else {
	fl_textedit_home_line(ob, shift_flag);
    }
}

static void Pad5KeyFunc(FL_OBJECT *ob) {
    if (!ob) return;
    SPEC *spec = (SPEC *)ob->spec;
    /* would want to center the current line and redraw the screen */
    int wsize = fl_textedit_get_wsize(ob);
    spec->topline = spec->r - wsize/2;
    if (spec->topline < 0) { 
	spec->topline = 0; 
    }
    update_hsb(ob, TRUE, TRUE);
    good_redraw(ob);
}

static void LeftArrowKeyFunc(FL_OBJECT *ob, int bExtendSelection) {
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int old_r = spec->r, old_c = spec->c;

    if (spec->c > 0) {
	fl_textedit_movecursor(ob, spec->r, spec->c - 1);
	update_hsb(ob, TRUE);
    } else {
	if (spec->r > 0) {
	    fl_textedit_lineup(ob);
	    fl_textedit_movecursor(ob, spec->r, tb_get_linelen(tb));
	    update_hsb(ob, TRUE);
	}
    }
    spec->cpos = spec->c;
    if (bExtendSelection) {
	set_or_extend_selection(ob, old_r, old_c);
    }
}

static void fl_textedit_wordup(FL_OBJECT *ob, int bExtendSelection) {
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int old_r = spec->r, old_c = spec->c;

    if (spec->c > 0) {
	char *line;
	tb_get_line(tb, &line);
	if (line == NULL) return;
	// eat whitespace up to the word
	int index = spec->c;
	index--;
	while ((index > 0) && (line[index] == ' ' || line[index] == '\t')) {
	    index--;
	}
	// jump over the word
	while ((index > 0) && (line[index-1] != ' ' && line[index-1] != '\t')) {
	    index--;
	}
	fl_textedit_movecursor(ob, spec->r, index);
	update_hsb(ob, TRUE);
	spec->cpos = spec->c;
	if (bExtendSelection) {
	    set_or_extend_selection(ob, old_r, old_c);
	}
    } else {
	LeftArrowKeyFunc(ob, bExtendSelection);
    }
}

static void RightArrowKeyFunc(FL_OBJECT *ob, int bExtendSelection) {
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int old_r = spec->r, old_c = spec->c;

    if (spec->c < tb_get_linelen(tb)) {
	fl_textedit_movecursor(ob, spec->r, spec->c + 1);
	update_hsb(ob, TRUE);
    } else {
	if (spec->r < tb_get_nlines(tb)-1) {
	    fl_textedit_linedown(ob);
	    fl_textedit_movecursor(ob, spec->r, 0);
	    update_hsb(ob, TRUE);
	}
    }
    spec->cpos = spec->c;
    if (bExtendSelection) {
	set_or_extend_selection(ob, old_r, old_c);
    }
}

static void fl_textedit_worddown(FL_OBJECT *ob, int bExtendSelection) {
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int old_r = spec->r, 
	old_c = spec->c,
        len   = tb_get_linelen(tb)
    ;

    if (spec->c < len) {
	char *line;
	tb_get_line(tb, &line);
	if (line == NULL) return;
	int index = spec->c;
	// eat whitespace up to the next word
	index++;
	while ((index < len) && (line[index] == ' ' || line[index] == '\t')) {
	    index++;
	}
	// jump over the word
	while ((index < len) && (line[index-1] != ' ' && line[index-1] != '\t')) {
	    index++;
	}
	// eat whitespace up to the beginning of the target word
	while ((index < len) && (line[index] == ' ' || line[index] == '\t')) {
	    index++;
	}
	fl_textedit_movecursor(ob, spec->r, index);
	update_hsb(ob, TRUE);
	spec->cpos = spec->c;
	if (bExtendSelection) {
	    set_or_extend_selection(ob, old_r, old_c);
	}
    } else {
	RightArrowKeyFunc(ob, bExtendSelection);
    }
}

static int edit_compute_extend(XKeyEvent *xkev, int key, int bAltFlag){
    /* if shift is pressed and a cursor positioning command is given,
     * do not remove the selection, but extend it instead.
     * commands are up, down, left, right, pageup/down, home, end,
     * ctrl-home, ctrl-end. 
     * This is tricky, as e.g. '<' often is the result of a shift
     * operation. In that case, shift is a bad indicator.  
     *
     * The same is true for the keypad with numlock on. 
     *
     * It's ok that there are no checks for alt in here, they are
     * done in edit_alt_keys.
     */

    if(is_down(key, xkev))	return TRUE;
    if(is_up(key, xkev))	return TRUE;
    if(is_left(key, xkev))	return TRUE;
    if(is_right(key, xkev))	return TRUE;
    if(is_home(key, xkev))	return TRUE;
    if(is_end(key, xkev))	return TRUE;
    if(is_page_down(key, xkev))	return TRUE;
    if(is_page_up(key, xkev))	return TRUE;

    if (bAltFlag){
	switch(key){
	case 'V': return TRUE;
	default:  return FALSE;
	}
    }
    switch(key){
    case ctrl('A'): return TRUE;
    case ctrl('B'): return TRUE;
    case ctrl('E'): return TRUE;
    case ctrl('F'): return TRUE;
    case ctrl('N'): return TRUE;
    case ctrl('P'): return TRUE;
    case ctrl('V'): return TRUE;
    default:	    return FALSE;
    }
}

#ifdef UNUSED_CODE
static void fl_textedit_clear(FL_OBJECT *ob){
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    if (spec->flags & FL_TEXTEDIT_READONLY) return;
    tb_clear(tb);
    update_hsb(ob);
    update_vsb(ob);
    fl_textedit_draw_screen(ob);
    spec->c = 0;
    spec->r = 0;
    spec->cpos = 0;
}
#endif

void
fl_textedit_handle_keyboard(FL_OBJECT *ob, int key, void *xev)
/*
 * Desc: handle the keyboard input 
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int nlines, th, n, nlen, nOldTopline;
    XKeyEvent *xkev = (XKeyEvent *)xev;
    int alt_flag    = Mod1Mask & xkev->state || tool_get_esc(),
        ctrl_flag   = ControlMask & xkev->state,
        shift_flag  = ShiftMask & xkev->state
    ;

    int bIsExtendSelection = shift_flag ? 
	edit_compute_extend(xkev, key, alt_flag) : FALSE
    ;

    //cerr << "handle keyboard.\n"; 

    /* clean-up state. May be time consuming, may even be wrong,
     * but I don't think so. The bugs I'm getting otherwise, like
     * non-local backspace, are horrid. 
     */
    tb_set_current_line(tb, spec->r);

    if (key != ctrl('K')) {
	clean_incremental_delbuf();
    }

#ifdef DEBUG
    printf("key character = %c (%d)\n", key, key);
    printf("alt  is pressed = %d\n", alt_flag);
    printf("ctrl is pressed = %d\n", ControlMask & xkev->state);
#endif
    if (alt_flag) {
	edit_alt_keys(ob, key);
	return;
    }

    /* check if read-only */
    if (spec->flags & FL_TEXTEDIT_READONLY)	{
	if ((key > 26) && (key < 256))
	    return;

	switch (key)	{
	case '\b':
	case '\n':
	case '\r':
	case '\t':
	case  127:
	    return;
	    break;
	}
    }

    /* check selection */
    if (!bIsExtendSelection && fl_textedit_selected(ob)) {
	//cerr << "remove sel for key" << (int) key << endl;
	
	switch (key)	{
	case '\b':	/* delete selected block */
	    delete_selected_block(ob);
	    return;

	case ctrl('L'): /* leave the selection alone */
	    break;

	/* deselect everything unless we try to type something */
	default:
	    if ((key == '\t') || 
		(key == '\n') || 
		(key == '\r') ||
		((key > 26) && (key < 256))) {
		delete_selected_block(ob);
	    } else {
		fl_textedit_remove_selection(ob);
	    }
	    break;
	}
    }

    /* first calculate the # of lines on the screen */
    nlines = fl_textedit_get_wsize(ob, &th);
    
    /* handle ctrl keys */
    if (key == '\n' || key == '\r') {
	edit_cr_key(ob, nlines);
	return;
    }

    if (key == '\b') {
	/* backspace */
	edit_backspace_key(ob);
	return;
    }
    if (key == 127) {
	/* del key */
	edit_delete_key(ob);
	return;
    }

    /* handle character inserts */
    if ((key == '\t') || ((key > 26) && (key < 256))) {
	edit_charinsert_key(ob, key);
	return;
    }

    if(is_delete(key, xkev)){
	edit_delete_key(ob);
	return;
    }
    if(is_up(key, xkev)){
        fl_textedit_lineup(ob, shift_flag);
	return;
    }
    if(is_down(key, xkev)){
	fl_textedit_linedown(ob, shift_flag);
	return;
    }
    if(is_home(key, xkev)){
	HomeKeyFunc(ob, xkev);
	return;
    }
    if(is_end(key, xkev)){
	EndKeyFunc(ob, xkev, shift_flag);
	return;
    }
    if(is_left(key, xkev)){
	if (ctrl_flag) {
	    fl_textedit_wordup(ob, shift_flag);
	} else {
	    LeftArrowKeyFunc(ob, shift_flag);
	}
	return;
    }
    if(is_right(key, xkev)){
	if (ctrl_flag) {
	    fl_textedit_worddown(ob, shift_flag);
	} else {
	    RightArrowKeyFunc(ob, shift_flag);
	}
	return;
    }
    if(is_page_down(key, xkev)){
	fl_textedit_pagedown(ob, shift_flag);
	return;
    }
    if(is_page_up(key, xkev)){
	fl_textedit_pageup(ob, shift_flag);
	return;
    }
    if(is_begin(key, xkev)){
	Pad5KeyFunc(ob);
	return;
    }

    /* handle function keys */
    switch(key) {
    case ctrl('A'):
	fl_textedit_home_line(ob, shift_flag);
    return;
    break;
    case ctrl('B'):
	LeftArrowKeyFunc(ob, shift_flag);
	return;
	break;
    case ctrl('D'):
	edit_delete_key(ob);
	return;
	break;
    case ctrl('E'):
	textedit_eol(ob, shift_flag);
    return;
    break;
    case ctrl('F'):
	RightArrowKeyFunc(ob, shift_flag);
	return;
	break;
    case ctrl('P'):
	/* move line up */
        fl_textedit_lineup(ob, shift_flag);
	return;
	break;
    case ctrl('N'):
	/* move line down */
	fl_textedit_linedown(ob, shift_flag);
	return;
	break;
    case ctrl('V'):
	/* pagedown */
	fl_textedit_pagedown(ob, shift_flag);
    return;
    break;
	/* pageup is alt-v, so no entry here */
    case ctrl('K'):
	if (spec->flags & FL_TEXTEDIT_READONLY) return;
        /* delete to end of line 
         * if the line is empty delete it
	 * if eol but not empty, concatenate lines
	 */
        nlen = tb_get_linelen(tb);
	nOldTopline = spec->topline;
        if (nlen > 0) {
	    if (nlen > spec->c) {
		char *line;
		tb_get_line(tb, &line);
		addto_incremental_delbuf(ob, line+spec->c);
		tb_set_linelen(tb, spec->c);
		fl_textedit_draw_line(ob, spec->r);
	    } else {
		addto_incremental_delbuf(ob, "\n");
		edit_delete_key(ob);
	    }
	    return;
	} else {
	    tb_del_line(tb);
	    addto_incremental_delbuf(ob, "\n");
	    n = tb_get_nlines(tb);
	    if (n == 0) {
		spec->r = 0;
	    } else {
		if (spec->r >= n)  spec->r = n - 1;
	    }

	    if (spec->c > tb_get_linelen(tb)) {
		spec->c = tb_get_linelen(tb);
	    }
	    update_vsb(ob);
	    spec->cpos = spec->c;
	    fl_textedit_refresh_screen(ob, 1);
	}
	if(nOldTopline != spec->topline) {
	    //cerr << "ctrl-k force-redraw!\n";
	    fl_textedit_draw_screen(ob);
	}
	return;
	break;
    case ctrl('L'):
	/* refresh screen */
	Pad5KeyFunc(ob);
    return;
    break;
    case ctrl('U'):
	/* U == undo, IMHO! */
	return;
	if (spec->flags & FL_TEXTEDIT_READONLY)
	    return;
    /* clear line */
    tb_set_linelen(tb, 0);
    spec->c = 0;
    spec->cpos = 0;
    fl_textedit_draw_line(ob, spec->r);
    return;
    break;
    case ctrl('T'):
	if (spec->flags & FL_TEXTEDIT_READONLY) return;
    fl_textedit_paste(ob);
    return;
    break;
    case ctrl('Y'):
	/* y == yank IMHO !*/
	return;
    break;
    case ctrl('Z'):
	/* super dangerous, IMHO */
	return;
    break;
    }
} /* fl_textedit_handle_keyboard() */

int
fl_textedit_paste(FL_OBJECT *ob)
/*
 * Paste selection into the textarea
 * returns TRUE if change occurred.
 */
{
    Window win;
    Atom prop;
    char *buf;
    int buflen, retval = TRUE;
    Window wn = fl_winget();

    fl_winset(ob->form->window);
    clean_incremental_delbuf();

    buf = NULL;
    buflen = 0;

    if (XGetSelectionOwner(fl_display,XA_PRIMARY) == None) {
	buf = XFetchBytes(fl_display, &buflen);
	if (!buf || !buflen) return FALSE;
	fl_textedit_inssel(ob, buf);		
    } else {
	win = XGetSelectionOwner(fl_display,XA_PRIMARY);
	if (win == ob->form->window)	{
	    if ((buf = fl_get_textedit_seltext(ob)) == NULL) {
		buf = XFetchBytes(fl_display, &buflen);
		if (buf && buflen) {
		    fl_textedit_inssel(ob, buf);
		}
	    } else {
		fl_textedit_inssel(ob, buf);
	    }
	} else {
	    FL_FORM *form = fl_win_to_form(win);
	    if (form) {
		//cerr << "internal selection mechanism necessary.\n";
		buf = XFetchBytes(fl_display, &buflen);
		if (buf && buflen) {
		    fl_textedit_inssel(ob, buf);
		}
	    } else {
		prop = XInternAtom(fl_display, "XF_SELECTION", FALSE );
		XConvertSelection(fl_display, XA_PRIMARY, XA_STRING, prop, ob->form->window, CurrentTime);
		XFlush(fl_display);
		retval = FALSE;
	    }
	}
    }
    if (buf) XFree((char *)buf);
    fl_winset(wn);
    return retval;
}

void
fl_textedit_handle_selnotify(FL_OBJECT *ob, XEvent *xevent)
/*
 * Desc: process SelectionNotify event, extract text buffer and paste it
 */
{
    char *buf;
    int buflen, actual_format;
    Window win;
    Atom prop, actual_type;
    u_long nitems, bytes_after, nread = 0;
    unsigned char *back;

    prop = xevent->xselection.property;
    win  = xevent->xselection.requestor;

    buflen = 0;
    buf = NULL;

    do {
        int r = XGetWindowProperty(fl_display, win, prop,
	    nread/4, 1024, TRUE,
	    AnyPropertyType, &actual_type,
	    &actual_format, &nitems,
	    &bytes_after, &back );

        if ( r != Success || actual_type != XA_STRING )
            break;

        while (((u_long) nread + nitems) >= (u_long) buflen) {
	    if (buflen) {
		buflen *= 2;
		buf = (char *)realloc(buf, buflen);
	    }
	    else	{
		buflen = 256;
		buf = (char *)malloc(buflen);
	    }
	}
        memcpy(buf + nread, back, nitems );
        nread += nitems;
        XFree((char *)back);
    } while ( bytes_after > 0 );

    if (buflen == 0)
	return;

    buf[nread] = 0;

    fl_textedit_inssel(ob, buf);
    free(buf);
}

static void
fl_textedit_inssel(FL_OBJECT *ob, char *buf)
/*
 * Desc: insert buf into text area
 *       buf is freed after insertion is done
 */
{
    TextBuf *tb;
    SPEC    *spec;
    int lines;
    u_long offt;

    spec = (SPEC *)ob->spec;
    tb = &spec->tb;

    fl_get_textedit_cursorpos(ob, NULL, NULL, &offt);

    lines = tb_get_nlines(tb);
    tb_insert_block(tb, spec->r, spec->c, buf);
    offt += strlen(buf);
    if (strchr(buf, '\n') || (lines != tb_get_nlines(tb)))	{
	if (spec->c > tb_get_linelen(tb)) {
	    spec->c = tb_get_linelen(tb);
	}
	spec->sselr = -1;
	spec->eselr = -1;
	
	// fuckup
	//fl_calc_cursorpos(ob, offt, &spec->c, &spec->r);

	fl_textedit_refresh_screen(ob, 0);
	update_vsb(ob);
	update_hsb(ob);
    } else {
	fl_calc_cursorpos(ob, offt, &spec->c, &spec->r);

	fl_textedit_remove_selection(ob);
	fl_textedit_refresh_screen(ob, 0);
	update_hsb(ob);
    }
}

char *
fl_get_textedit_seltext(FL_OBJECT *ob)
/*
 * Desc: returns pointer to selected text
 *       The text should be freed after using
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    char *buf;

    /* check if we have any text selected */
    if ((spec->sselr < 0) || (spec->eselr < 0) ||
	((spec->sselr == spec->eselr) && (spec->sselc == spec->eselc)))
	return NULL;

    tb_get_block(tb, spec->sselr, spec->sselc, spec->eselr, spec->eselc, &buf);

    return buf;
}

long fl_get_textedit_selsize(FL_OBJECT *ob){
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    if (!fl_textedit_selected(ob)) return -1;
    return (long) tb_get_block_size(
	tb, spec->sselr, spec->sselc, spec->eselr, spec->eselc
    );
}

int
fl_textedit_copy(FL_OBJECT *pObj, char *pBuf)
/*
 * Copy selected text from textarea to X-servers buffer
 */
{
    Window win;
    char *buf;
    int bExternal = (pBuf != NULL);
    if (!pObj || !pObj->form) return FALSE;
    //SPEC *spec = (SPEC *)pObj->spec;
    remove_other_selections(pObj);

    if (fl_textedit_selected(pObj) || bExternal) {
	XSetSelectionOwner(fl_display, XA_PRIMARY, pObj->form->window, CurrentTime);
	win = XGetSelectionOwner(fl_display, XA_PRIMARY);
	if (win != pObj->form->window) {
	    cerr << "failed to set selection owner.\n";
	    return FALSE;
	}

	buf = bExternal ? pBuf : fl_get_textedit_seltext(pObj);
	if (buf) {
	    /*
	    cerr << "storing bytes (" << spec->sselr << "," << spec->sselc;
	    cerr << " -> " << spec->eselr << "," << spec->eselc;
	    cerr << "): '" << buf << "', external: " << bExternal << "\n";
	    */
	    XStoreBytes(fl_display, buf, strlen(buf));
	    if (!bExternal) free(buf);	
	    return TRUE;
	} else {
	    cerr << "selection status true, selection text empty!\n";
	}
    }
    return FALSE;
}

int
fl_textedit_set_cursor(FL_OBJECT *ob, int mx, int my, int)
/* int draw not used.
 * Desc: set the textcursor to the textcoordinates 
 *	  corresponding with (mx,my)
 * Returns: -1 if cursor is below the text area
 *          -2 at the left
 *           0 in the text area
 *           2 at the right
 *           1 above the text area
 */
{
    TextBuf *tb;
    SPEC    *spec;
    char    *line, ch[2];
    int	    i, lh, x, oldline, nlines, r, c;

    /* correct the mx and my offset to be relative to the 
       start of the text */
    mx = mx - ob->x - TEXT_X_OFFSET;
    my = my - ob->y - TEXT_Y_OFFSET;

    //cerr << "set cursor mx, my: " << mx << ", " << my << endl;

    if (my < 0)
	return -1;

    if (mx < 0)
	return -2;

    if (mx > (ob->w - 2*abs(ob->bw) - 2* TEXT_X_OFFSET))
	return 2;

    spec = (SPEC *)ob->spec;
    tb = &spec->tb;

    /* get the height of the font */
    nlines = fl_textedit_get_wsize(ob, &lh);

    /* save the old line with the cursor on it */
    oldline = spec->r;


    /* set the current line to the right one, but check if it's not
       out of range */
    int nMaxLine = tb_get_nlines(tb) - 1;

    if ((my/lh) >= nlines && spec->r < nMaxLine) {
	// still space for scrolling
	return 1;
    }
    int bEndFlag = FALSE;

    if (nMaxLine < 0) nMaxLine = 0;
    if (spec->topline + (my/lh) > nMaxLine) {
	// limit the range, do not return here!
	// return 0;
	r = nMaxLine;
	bEndFlag = TRUE;
    } else {
	r = spec->topline + (my/lh);
    }

    tb_set_current_line(tb, r);

    /* now calculate which character the x position points to */
    /* first get the current line */
    tb_get_line(tb, &line);
    if (!line) {
	/* no line available */
	return 0;
    }
    int nDummy = 0;
    char *pSearchLine = expand_tabs(line, nDummy, 0);

    lh = strlen(pSearchLine);
    x = 0;
    i = spec->leftcol;
    if (i > lh)
	i = lh;
    ch[1] = '\0';
    while ((x < mx) && (i < lh) && pSearchLine[i]) {
	/* get the width of the i-th character */
	ch[0] = pSearchLine[i];
	x = x + fl_get_string_width(spec->text_style, spec->text_size, ch, 1);
	i++;
    } /* x>=mx || pSearchLine[i]==NULL */
    if (x>=mx && i>0) {
	i--;
    }
    if (bEndFlag) {
	i = lh;
    }
    /* if ((spec->c == i) && (spec->r == oldline)) */
    if ((spec->c == i) && (r == oldline)) return 0;

    //cerr << "offset position i:   " << i << endl;
    c = calculate_real_cpos (line, i);
    //cerr << "calculated position: " << c << endl;
    fl_textedit_movecursor(ob, r, c);
    spec->cpos = spec->c;

    return 0;
} /* fl_textedit_set_cursor() */

int
fl_textedit_selected(FL_OBJECT *ob)
{
    SPEC *spec = (SPEC *)ob->spec;

    /* check if there is any text selected */
    if ((spec->sselr < 0) ||
	(spec->eselr < 0))
	return 0;

    /* check if length of the selected part is bigger then zero */
    if ((spec->sselr == spec->eselr) &&
	(spec->sselc == spec->eselc))
	return 0;

    return 1;
}

void
fl_textedit_remove_selection_backend(FL_OBJECT *ob)
/*
 * Desc: remove text selection. work on this object only.
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    int i, ssel, esel;
    //cerr << "remove_selection() called.\n";
    if (!fl_textedit_selected(ob)) {
	return;
    }
    ssel = spec->sselr;
    esel = spec->eselr;

    spec->sselr = -1;
    spec->eselr = -1;

    Window wn = fl_winget();
    fl_winset(ob->form->window);
    
    for (i = ssel; i <= esel; i++) {
	fl_textedit_draw_line(ob, i);
    }
    fl_winset(wn);

    SelectionDirection = NoneE;
}

FL_CALLBACKPTR fl_textedit_remove_other_selections_cb = NULL;
static void remove_other_selections(FL_OBJECT *ob) {
    if (fl_textedit_remove_other_selections_cb) {
	fl_textedit_remove_other_selections_cb(ob, 0);
    }
}

void fl_textedit_remove_selection(FL_OBJECT *ob){
    /* call user callback to remove selection in other open editors */
    remove_other_selections(ob);
    fl_textedit_remove_selection_backend(ob);
}


// returns wether Pt is before Ref.

static int PtBefore(int nPtR, int nPtC, int nRefR, int nRefC) {
    if (nPtR < nRefR) return TRUE;
    if (nPtR > nRefR) return FALSE;
    return (nPtC < nRefC);
}

static int PtEqual(int nPtR, int nPtC, int nRefR, int nRefC) {
    return (nPtR == nRefR && nPtC == nRefC); 
}

/* start a new selection or extend an existing one */
static void set_or_extend_selection(
    FL_OBJECT *ob, int old_r, int old_c, int bNoRedraw
){
    SPEC *spec = (SPEC *)ob->spec;
    if (!fl_textedit_selected(ob)) {
/*
	cerr << "start new selection at old position.\n";
	cerr << "sselr, sselc" << spec->sselr << ", " << spec->sselc << endl;
	cerr << "eselr, eselc" << spec->eselr << ", " << spec->eselc << endl;
*/
	spec->sselr = spec->eselr = old_r;
	spec->sselc = spec->eselc = old_c;
    }
    fl_textedit_extend_selection(ob, old_r, old_c, bNoRedraw);
}

static void fl_textedit_extend_selection(
    FL_OBJECT *ob, int old_r, int old_c, int bNoRedraw
){
/*
 * Desc: extend selection up to cursor location
 */
    SPEC *spec = (SPEC *)ob->spec;
    int sselr, eselr, sselc, eselc, i, nTmpStart = -1, nTmpEnd = -1;

    /* no selection */
    if ((spec->sselr < 0) ||
	(spec->sselc < 0))
	return;

    if (old_r == spec->r && old_c == spec->c) return;
/*
    cerr << "\nextend_selection from old_r, old_c " << old_r << ", " << old_c 
	 << " to r,c " << spec->r << ", " << spec->c << endl;
    cerr << "    sselr, sselc" << spec->sselr << ", " << spec->sselc << endl;
    cerr << "    eselr, eselc" << spec->eselr << ", " << spec->eselc << endl;
    switch(SelectionDirection){
    case DownE:	cerr << "Down\n"; break;
    case UpE: 	cerr << "Up\n"; break;
    default: 	cerr << "None\n"; break;
    }
 */
    // special case if reversing direction. Normally, this code assumes
    // that the starting point of the selection is before the end point,
    // which is wrong. In the reverse case, set the end point to the 
    // start point, and adjust the start point. Don't forget redraw.

    if (SelectionDirection == DownE) {
	if (PtBefore(spec->r, spec->c, spec->sselr, spec->sselc)) {
	    // cursor is before start of selection. Make cursor the new
	    // start, and start the new end.
	    //cerr << "Down Reverse warning\n";
	    nTmpStart = spec->eselr;
	    spec->eselr = spec->sselr;
	    spec->eselc = spec->sselc;
	    SelectionDirection = UpE;
	} else if (PtEqual(spec->r, spec->c, spec->sselr, spec->sselc)) {
	    // we went down, but back to the starting point. Set the
	    // selection length to zero and the direction to unknown.
	    //cerr << "Down Equal warning\n";
	    nTmpStart = spec->eselr;
	    spec->eselr = spec->sselr;
	    spec->eselc = spec->sselc;
	    SelectionDirection = NoneE;
	}   // else o.k.

	if (nTmpStart >=0 && !bNoRedraw) {
	    // these redraws get lost otherwise.
	    for (i = nTmpStart; i >= spec->eselr; i--) {
		fl_textedit_draw_line(ob, i);
	    }
	}
    } else if (SelectionDirection == UpE) {
	nTmpEnd = spec->eselr;
	if (PtBefore(spec->eselr, spec->eselc, spec->r, spec->c)) {
	    // cursor is after the end of selection. make cursor the
	    // new end, and set end the new start.
	    //cerr << "Up Reverse warning\n";
	    nTmpStart = spec->sselr;
	    spec->sselr = spec->eselr;
	    spec->sselc = spec->eselc;
	    SelectionDirection = DownE;
	} else if (PtEqual(spec->r, spec->c, spec->eselr, spec->eselc)) {
	    // we went up, but back to the ending point. Reset the 
	    // start point and the direction to unknown. 
	    //cerr << "Up Equal warning\n";
	    nTmpStart = spec->sselr;
	    spec->sselr = spec->eselr;
	    spec->sselc = spec->eselc;
	    SelectionDirection = NoneE;
	} else if (PtBefore(old_r, old_c, spec->r, spec->c)) {
	    // cursor is before the end of the selection, but possibly 
	    // after the start, so adjust the start.
	    //cerr << "Up decrease warning\n";
	    nTmpStart = spec->sselr;
	    nTmpEnd = spec->r;
	    spec->sselr = spec->r;
	    spec->sselc = spec->c;
	}
	if (nTmpStart >=0 && !bNoRedraw) {
	    // these redraws gets lost otherwise.
	    for (i = nTmpStart; i <= nTmpEnd; i++) {
		fl_textedit_draw_line(ob, i);
	    }
	}
    } else {
	// we have no direction yet, so establish one. 
	if (PtBefore(spec->eselr, spec->eselc, spec->r, spec->c)) {
	    SelectionDirection = DownE;
	} else if (PtBefore(spec->r, spec->c, spec->sselr, spec->sselc)) {
	    SelectionDirection = UpE;
	}
    }

    eselr = spec->eselr;
    sselr = spec->sselr;
    eselc = spec->eselc;
    sselc = spec->sselc;

    /* the selection now starts/ends at cursor position */
    if (spec->sselr > spec->r)	{
	spec->sselr = spec->r;
	spec->sselc = spec->c;
    } else {
	if (spec->sselr == spec->r)	{
	    if (spec->c < spec->sselc) {
		spec->sselr = spec->r;
		spec->sselc = spec->c;
	    }
	    else
		if (spec->c > spec->sselc) {
		    spec->eselr = spec->r;
		    spec->eselc = spec->c;
		}	
		else	{
		    if (old_r != spec->r && !bNoRedraw) {
			fl_textedit_draw_line(ob, old_r);
			fl_textedit_draw_line(ob, spec->r);
		    }
		    return;
		}
	} else {
	    spec->eselr = spec->r;
	    spec->eselc = spec->c;
	}
    }

    if (bNoRedraw) return;

    if (eselr < spec->eselr) {
	for (i = eselr; i <= spec->eselr; i++) {
	    fl_textedit_draw_line(ob, i);
	}
    } else {
	if (eselr > spec->eselr) {
	    for (i = spec->eselr; i <= eselr; i++) {
		fl_textedit_draw_line(ob, i);
	    }
	} else {
	    if (sselr < spec->sselr) {
		for (i = sselr; i <= spec->sselr; i++) {
		    fl_textedit_draw_line(ob, i);
		}
	    } else {
		if (sselr > spec->sselr) {
		    for (i = spec->sselr; i <= sselr; i++) {
			fl_textedit_draw_line(ob, i);
		    }
		} else {
		    if ((eselc != spec->eselc) || 
			(sselc != spec->sselc) || 
			(old_r != spec->r)
		    ) {
			fl_textedit_draw_line(ob, spec->r);
		    }
		}
	    }
	}
    }

    if (old_r != spec->r) {
	fl_textedit_draw_line(ob, old_r);
    }
}

static void edit_extend_from_mouse_pos(
    FL_OBJECT *ob, 
    FL_Coord mx, 
    FL_Coord my
){
    SPEC *spec = (SPEC *)ob->spec;
    /* store the old cursor position */
    int old_c = spec->c, old_r = spec->r;

    /* set the cursor at the correct position */
    switch (fl_textedit_set_cursor(ob, mx, my, 0))	{
    case -1:
	fl_textedit_lineup(ob);
	// special case: hit the top. 
	if (spec->r == 0) {
	    fl_textedit_home_file(ob);
	}
	break;
    case 1:
	//fl_textedit_linesdown(ob, nMouseCount+1);
	fl_textedit_linedown(ob);
	break;
    case -2:
	fl_textedit_cleft(ob);
	break;
    case 2:
	fl_textedit_cright(ob, TRUE);
	break;
    default:
	//cerr << "default return from set cursor.\n";
	//cerr << " end (set cursor)";
	break;
    }

    /* select everything up to the cursor position */
    fl_textedit_extend_selection(ob, old_r, old_c);
    //XFlush(fl_display);
}

/***********************************************************************
 *
 * Main event handling 
 *
 ***********************************************************************/

static int
handle_textedit(FL_OBJECT *ob, int event, FL_Coord mx, FL_Coord my, 
    int key, void *xev)
/*
 * Desc: the event handler for the textedit object 
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int hsbh;
    char *line;
    XEvent *txev;

    // for acceleration. Increase the count only on successfull
    // scrolling, otherwise, it may be an event at the same x,y-pos.
    static int nMouseCount = 0;
    if (event != FL_MOUSE) nMouseCount = 0;
     
    txev = (XEvent *)xev;
/*
    static int count=0;
    cerr << "handle text edit (" << count << ") start..." << event << endl;
    count++;
    */
#ifdef DEBUG
    if (xev && ((XEvent *)xev)->type == KeyPress) {
	XKeyEvent *xkev = (XKeyEvent *)xev;
	printf("keycode = %d, state = %d\n", xkev->keycode, xkev->state);
	printf("Control = %d, Shift = %d, Alt = %d\n",
	    xkev->state & ControlMask, xkev->state & ShiftMask, 
	    xkev->state & Mod1Mask);
    }
#endif

    switch(event) {
    case FL_DRAW:
	if (spec->flags & FL_TEXTEDIT_NOTDRAWN)	{
	    spec->flags &= ~FL_TEXTEDIT_NOTDRAWN;
	    //spec->sb->x = ob->x;
	    //spec->hsb->x = ob->y;
	}

	if (spec->sb->visible) {
	    if ((spec->sb->x != ob->x + ob->w) ||
		(spec->sb->y != ob->y) ||
		(spec->sb->w != TEXT_SB_WIDTH) ||
		(spec->sb->h != ob->h) ||
		((spec->sb->resize & FL_RESIZE_X) != (ob->resize & FL_RESIZE_X)) ||
		(spec->sb->segravity != ob->segravity) ||
		(spec->sb->nwgravity != ob->nwgravity)) {
		fl_textedit_veditscrollbar_sync_geometry(ob);
		if (spec->oldh != ob->h){
		    update_vsb(ob);
		}
		fl_editscrollbar_draw(spec->sb);
	    }
	} else {
	    if (spec->oldh != ob->h){
		update_vsb(ob);
	    }
	}

	if (spec->hsb->visible) 	{
	    hsbh = spec->sb->visible ? ob->w + TEXT_SB_WIDTH : ob->w;
	    if ((spec->hsb->y != ob->y + ob->h) ||
		(spec->hsb->x != ob->x) ||
		(spec->hsb->h != TEXT_SB_WIDTH) ||
		(spec->hsb->w != hsbh) ||
		((spec->hsb->resize & FL_RESIZE_Y) != (ob->resize & FL_RESIZE_Y)) ||
		(spec->hsb->segravity != ob->segravity) ||
		(spec->hsb->nwgravity != ob->nwgravity)) {
		fl_textedit_heditscrollbar_sync_geometry(ob);
		if (spec->oldw != ob->w)
		    fl_textedit_set_heditscrollbar_wsize(ob);
		fl_editscrollbar_draw(spec->hsb);
	    }
	}
	else	{
	    if (spec->oldw != ob->w)
		fl_textedit_set_heditscrollbar_wsize(ob);
	}

	fl_textedit_draw_screen(ob);
	spec->oldw = ob->w;
	spec->oldh = ob->h;
	/* fall through */
    case FL_DRAWLABEL:
	fl_draw_object_label(ob);
	break;
    case FL_PUSH:
	clean_incremental_delbuf();
	SelectionDirection = NoneE;
	switch (key)  {
	case 1:
	    if (txev->xbutton.state & ShiftMask) {
		edit_extend_from_mouse_pos(ob, mx, my);
	    } else {
		/* set the cursor at the correct position */
		fl_textedit_set_cursor(ob, mx, my, 1);

		/* start new selection */
		fl_textedit_remove_selection(ob);
		spec->sselr = spec->eselr = spec->r;
		spec->sselc = spec->eselc = spec->c;
	    }
	    break;

	case 2:
	    /* paste selection into the text area */
	    if (spec->flags & FL_TEXTEDIT_READONLY)
		break;
	    fl_textedit_paste(ob);
	    break;

	case 3:
	    /* call callback function */
	    //cerr << " end (push)";
	    break;
	}
	return(1); // was 1
	break;
    case FL_MOUSE:
	if (key == 1) {
	    edit_extend_from_mouse_pos(ob, mx, my);
	}
	break;
    case FL_RELEASE:
	/* copy selected text */
	clean_incremental_delbuf();
	SelectionDirection = NoneE;
	if (key == 1) {
	    fl_textedit_copy(ob);
	    return(1);	// was 1
	}
	return(0);
    case FL_KEYBOARD:
	fl_textedit_handle_keyboard(ob, key, xev);
	break;
    case FL_UNFOCUS:
	/* deselect everything */
	fl_textedit_remove_selection(ob);
	/* hide cursor */
	if (ob->form->visible)
	    fl_textedit_movecursor(ob, -1, -1);
	break;
    case FL_FOCUS:
	/* show cursor */
	if (ob->form->visible)
	    fl_textedit_movecursor(ob, spec->r, spec->c);
	break;
    case FL_FREEMEM:
	/* free the memory used by the spec member of ob */
    {
	tb_clear(tb);
	if (spec->exp)
	    free(spec->exp);
	free(spec);
    }
    return(0);
    case FL_DBLCLICK:
	/* a double click event was received, the timeout must be 
	   set in ob->click_timeout */
	/* mark the current word */
	clean_incremental_delbuf();
	tb_get_line(tb, &line);
	if (line == NULL)
	    break;

	fl_textedit_remove_selection(ob);
	spec->sselr = spec->r;
	spec->eselr = spec->r;

	spec->sselc = spec->c;
	spec->eselc = spec->c;

	while ((spec->sselc > 0) && (line[spec->sselc - 1] != ' '))
	    spec->sselc--;

	while ((line[spec->eselc] != '\0') && (line[spec->eselc] != ' '))
	    spec->eselc++;

	fl_textedit_draw_selection(ob);
	fl_textedit_copy(ob);
	break;
    case FL_TRPLCLICK:
	/* a triple click event was received, the timeout must be 
	   set in ob->click_timeout */
	clean_incremental_delbuf();
	/* mark the current line */
	fl_textedit_remove_selection(ob);
	spec->sselr = spec->r;
	spec->eselr = spec->r;

	spec->sselc = 0;
	spec->eselc = -1;
	fl_textedit_draw_selection(ob);
	fl_textedit_copy(ob);
	break;
    case FL_OTHER:
	switch (txev->type) {
	case SelectionClear:
	    fl_textedit_remove_selection(ob);
	    //cerr << "remove selection.\n";
	    break;

	case SelectionNotify:
	    //cerr << "SelectionNotify\n";
	    if (!ob->focus) {
		cerr << "got selection notify, but don't have focus.\n";
		return(0);
	    }
	    fl_textedit_remove_selection(ob);
	    fl_textedit_handle_selnotify(ob, txev);
	    break;

	case SelectionRequest:	{
            XSelectionRequestEvent *req = &txev->xselectionrequest;
            XEvent evt;
	    char *buf;
            evt.xselection.type = SelectionNotify;
            evt.xselection.display      = req->display;
            evt.xselection.requestor    = req->requestor;
            evt.xselection.selection    = req->selection;
            evt.xselection.target       = req->target;
            evt.xselection.property     = None;
            evt.xselection.time = req->time;

	    if (!fl_textedit_selected(ob))
		break;

            if ( req->target == XA_STRING ) {
		buf = fl_get_textedit_seltext(ob);
		if (buf) 	{
		    XChangeProperty ( fl_display, req->requestor, req->property,
			XA_STRING, 8,
			PropModeReplace,
			(unsigned char *)buf, strlen(buf) );
		    evt.xselection.property = req->property;
		    free(buf);	}
            }
            XSendEvent( fl_display, req->requestor, False, 0, &evt );
	}
	break;
	}	
	break;
    default:
	return 0;
    }
    // cerr << " end (normal)\n"; 
    return(1);
} /* handle_textedit() */

void
fl_textedit_sb_cb(FL_OBJECT *pEditscrollbarObj, long data)
/*
 * Desc: the callback for the editscrollbar object to call to redraw
 *	  the screen
 */
{
    FL_OBJECT *pEditObj = (FL_OBJECT *)data;
    SPEC *spec = (SPEC *)pEditObj->spec;

    spec->r = (int) fl_get_editscrollbar_val(pEditscrollbarObj);
    spec->c = spec->cpos = 0;
    spec->topline = (int) fl_get_editscrollbar_val(pEditscrollbarObj);
    fl_textedit_draw_screen(pEditObj);
    fl_trigger_object(pEditObj);
} /* fl_textedit_sb_cb() */

void
fl_textedit_hsb_cb(FL_OBJECT *ob, long data)
/*
 * Desc: the callback for the horizontal editscrollbar object to call to redraw
 *	  the screen
 */
{
    FL_OBJECT *obj = (FL_OBJECT *)data;
    SPEC *spec = (SPEC *)obj->spec;

    spec->leftcol = (int) fl_get_editscrollbar_val(ob);
    fl_textedit_draw_screen((FL_OBJECT *) data);
}


FL_OBJECT *
fl_create_textedit(int type, FL_Coord x, FL_Coord y, 
    FL_Coord w, FL_Coord h, const char *label)
/*
 * Desc: create the textedit object
 */
{
    FL_OBJECT *ob;
    SPEC *spec;

    ob = fl_make_object(FL_TEXTEDIT, type, x, y, w, h,
	label, handle_textedit);

    ob->col1 = FL_BLACK;
    ob->col2 = FL_MCOL;
    ob->boxtype = FL_DOWN_BOX;
    ob->wantkey = FL_KEY_TAB|FL_KEY_SPECIAL;
    ob->click_timeout = 300;
    ob->input = TRUE;

    /* initialize the object specific user data */
    ob->spec = (SPEC *) fl_calloc(1, sizeof(SPEC));
    spec = (SPEC *)ob->spec;
    if (!ob->spec) {
	fl_error_msg("Could not malloc space for SPEC structure in object:%s", 
	    label);
	exit(1);
    }
    tb_init(&spec->tb);
    spec->tb.fgcolor_def = ob->col1;
    spec->tb.bgcolor_def = ob->col2;
    spec->r = 0;
    spec->c = 0;
    spec->cpos = 0;
    spec->topline = 0;
    spec->leftcol = 0;
    spec->name[0] = '\0';
    spec->text_style = FL_FIXED_STYLE;
    spec->text_size = FL_NORMAL_SIZE;
    spec->ccol = FL_RED;
    spec->sselr = -1;
    spec->sselc = 0;
    spec->eselr = -1;
    spec->eselc = 0;
    spec->flags = FL_TEXTEDIT_NOTDRAWN;
    spec->exp = NULL;
    spec->oldw = ob->w;
    spec->oldh = ob->h;

    /* editscrollbar section */
    spec->sb = fl_add_editscrollbar(FL_VERTICAL_EDITSCROLLBAR, x+w-TEXT_SB_WIDTH,
	y, TEXT_SB_WIDTH, h, "");
    fl_set_object_callback(spec->sb, fl_textedit_sb_cb, (long) ob);
    spec->sb->visible = TRUE;
    fl_textedit_veditscrollbar_sync_geometry(ob);

    update_vsb(ob);		// enable
    ob->w -= TEXT_SB_WIDTH;	// mung editor
    fl_textedit_veditscrollbar_sync_geometry(ob);

    spec->hsb = fl_add_editscrollbar(FL_HORIZONTAL_EDITSCROLLBAR, x, y+h-TEXT_SB_WIDTH,
	w-TEXT_SB_WIDTH, TEXT_SB_WIDTH, "");
    fl_set_object_callback(spec->hsb, fl_textedit_hsb_cb, (long) ob);
    spec->hsb->visible = TRUE;
    fl_textedit_heditscrollbar_sync_geometry(ob);

    update_hsb(ob);
    ob->h -= TEXT_SB_WIDTH;
    fl_textedit_heditscrollbar_sync_geometry(ob);

    clean_incremental_delbuf();

    return(ob);
} /* fl_create_textedit() */

/***********************************************************************
 *
 * Exported API
 *
 ***********************************************************************/

FL_OBJECT *
fl_add_textedit(int type, FL_Coord x, FL_Coord y, 
    FL_Coord w, FL_Coord h, const char *label)
{
    FL_OBJECT *ob = fl_create_textedit(type, x, y, w, h, label);

    fl_add_object(fl_current_form, ob);

    return(ob);
} /* fl_add_textedit() */

void
fl_set_textedit(FL_OBJECT *ob, char *buf, long len)
/*
 * Desc: insert text from <buf> with length <len> into the textedit object <ob>
 * Pre:  The textedit object has been created
 *
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int n;

    tb_set_text(tb, buf, len);
    strcpy(spec->name, "noname");
    spec->r = 0;
    spec->c = 0;
    spec->cpos = 0;
    spec->topline = 0;
    spec->leftcol = 0;
    spec->sselr = spec->eselr = -1;
    n = tb_get_nlines(tb);

    update_vsb(ob);
    fl_textedit_set_heditscrollbar_wsize(ob);

    fl_redraw_object(ob);
} /* fl_set_textedit() */

int
fl_load_textedit(FL_OBJECT *ob, char *fname)
/*
 * Desc: load a file into the textedit object <ob>
 * Pre:  The textedit object has been created
 *
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int retval = FALSE;

    spec->r = 0;
    spec->c = 0;
    spec->cpos = 0;
    spec->topline = 0;
    spec->leftcol = 0;
    spec->sselr = spec->eselr = -1;
    if (tb_load_file(tb, fname)) {
        strcpy(spec->name, fname);
	retval = TRUE;
    }
    update_vsb(ob);
    fl_textedit_set_heditscrollbar_wsize(ob);

    fl_redraw_object(ob);
    clean_incremental_delbuf();

    return retval;
} /* fl_load_textedit() */

int
fl_save_textedit(FL_OBJECT *ob, char *fname)
/*
 * Desc: Save the textedit file under the name <fname>
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;

    if (tb_save_file(tb, fname)) {
        strcpy(spec->name, fname);
	return TRUE;
    }
    return FALSE;  
} /* fl_save_textedit() */

char *
fl_get_textedit(FL_OBJECT *ob, u_long *len)
/*
 * Desc: return pointer to buffer with text of object <ob>
 * length of the buffer will be returned in <len>
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;

    return tb_get_text(tb, len);

} /* fl_get_textedit() */

u_long fl_get_textedit_buflen(FL_OBJECT *ob){
    if (!ob) return 0;
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    return tb_get_buflen(tb);
}

void 
fl_set_textedit_fontsize(FL_OBJECT *ob, int size)
/*
 * Desc: set the font size of the object 
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;

    if ((spec->text_size != size) && (tb->n > 0))	{
	spec->text_size = size;
	fl_redraw_object(ob);
	update_vsb(ob);
	fl_textedit_set_heditscrollbar_wsize(ob);		}

    spec->text_size = size;
} /* fl_set_textedit_fontsize() */

int
fl_get_textedit_fontsize(FL_OBJECT *ob)
/*
 * Desc: get the font size of the object
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    
    return(spec->text_size);
} /* fl_get_textedit_fontsize() */

void
fl_set_textedit_fontstyle(FL_OBJECT *ob, int style)
/*
 * Desc: set the font style of the object 
 */
{
    SPEC *spec = (SPEC *) ob->spec;
    TextBuf *tb = &spec->tb;
    
    if ((style != spec->text_style) && (tb->n > 0)) {
	spec->text_style = style;
	fl_redraw_object(ob);
	update_vsb(ob);
	fl_textedit_set_heditscrollbar_wsize(ob);	    }

    spec->text_style = style;
} /* fl_set_textedit_fontstyle() */

int
fl_get_textedit_fontstyle(FL_OBJECT *ob)
/*
 * Desc: get the font style of the object
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    return(spec->text_style);
} /* fl_get_textedit_fontstyle() */

void
fl_set_textedit_color(FL_OBJECT *ob, int bgcol, int fgcol, int ccol, int all)
/*
 * Desc: set the color of the object 
 * <bgcol>, <fgcol> and <ccol> are background , foreground and cursor colors
 *	respectively
 * if <all> is not 0 then color of every line is set to given colors
 */
{
    int redraw;
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;

    redraw = all;
    if (spec->ccol != ccol)
	redraw = 1;
    spec->ccol = ccol;
    ob->col1 = fgcol;
    ob->col2 = bgcol;
    tb->fgcolor_def = ob->col1;
    tb->bgcolor_def = ob->col2;

    if (all && (tb_get_nlines(tb) > 0)) {
        tb_set_current_line(tb, 0);
        tb_set_linefgcolor(tb, fgcol);
        tb_set_linebgcolor(tb, bgcol);
        while (tb_set_next_line(tb)) {
            tb_set_linefgcolor(tb, fgcol);
            tb_set_linebgcolor(tb, bgcol);
        }
    }

    if (redraw)
	fl_redraw_object(ob);
}

void
fl_get_textedit_color(FL_OBJECT *ob, int *bgcol, int *fgcol, int *ccol)
/*
 * Desc: get the color of the object 
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    if (ccol)
	*ccol = spec->ccol;
    if (bgcol)
	*bgcol = ob->col2;
    if (fgcol)
	*fgcol = ob->col1;
}

char *
fl_get_textedit_bufname(FL_OBJECT *ob)
/* 
 * Desc: return the name of the buffer 
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    
    return(spec->name);
} /* fl_get_textedit_bufname() */

void
fl_set_textedit_bufname(FL_OBJECT *ob, char *fname)
/*
 * Desc: set the name of the buffer
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    strcpy(spec->name, fname);
} /* fl_set_textedit_bufname() */

int
fl_get_textedit_bufchanged(FL_OBJECT *ob)
/*
 * Desc: return the change status of the textbuffer 
 *	  TRUE if it needs saving, FALSE if not
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;

    return(tb_bufchanged(tb));
} /* fl_textedit_bufchanged() */

void
fl_set_textedit_bufchanged(FL_OBJECT *ob, int status)
/*
 * Desc: set the bufferchanged status 
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    
    tb_set_bufchanged(tb, status);
} /* fl_set_textedit_bufchanged() */

FL_OBJECT *
fl_get_textedit_heditscrollbar(FL_OBJECT *ob)
/*
 * return pointer to horizontal editscrollbar associated with textedit 
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    return spec->hsb;
}

void
fl_clear_textedit(FL_OBJECT *ob)
/*
 * clear text area
 */
{
    SPEC *spec = (SPEC *)ob->spec;
 
    tb_clear(&spec->tb);
    spec->r = 0;
    spec->c = 0;
    spec->sselr = -1;
    spec->eselr = -1;
    spec->cpos = 0;
    spec->topline = 0;
    spec->leftcol = 0;
    if (spec->exp) {
	free(spec->exp);
	spec->exp = NULL; 
    }
    update_vsb(ob);
    update_hsb(ob);

    fl_redraw_object(ob);
}

int
fl_textedit_movecursor(FL_OBJECT *ob, int r, int c)
/*
 * move cursor to specified position, adjust for tabs. 
 * expand the line before processing it, and adjust the 
 * cursor position accordingly (only locally to this function);
 * Otherwise, a ccol of 8 would be possible in a line with len == 1 tab :-)
 *
 * The function deals with the redrawing of both the old and new lines, if
 * they are different. It tries to redraw only the affected characters. 
 *
 * We need to adjust c, and mung the text. Problematic are the length/
 * visibility calculations, expecially where they feedback into spec. 
 */
{
    SPEC    *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int	nlines, tx, ty, tw, th;		/* bounding box of line */
    int x, y, w, h, swidth, rc;
    int nHackLine = -1;
    char onechar[2];
    char *pDrawLine;
    int  csize;
    TextLine *tl;

    //cerr << "move cursor (r,c): " << r << ", " << c << endl;
    if (r == spec->r && c == spec->c) return 0;	// HAK

    fl_textedit_get_textbb(ob, &x, &y, &w, &h);
    nlines = fl_textedit_get_wsize(ob, &th);
    csize = fl_textedit_get_csize(ob);

    // handle old line first:
    tl = tb_get_lineptr_by_num(tb, spec->r);
    if (!tl) {
	if (!spec->r && !spec->c) {
	    swidth = fl_get_string_width(spec->text_style, spec->text_size, " ", 1);
	    fl_rectf(x, y, swidth, th, ob->col2); 
	}
	goto showcur;   
    }

    /* line is out of visible range */
    if ((spec->r < spec->topline) || (spec->r >= spec->topline+nlines)) {
	goto showcur;
    }

    if (spec->c > tl->strlen) {
	spec->c = tl->strlen;
    }
    rc = spec->c - spec->leftcol;
    pDrawLine = expand_tabs(tl->buf, rc, spec->leftcol);
    if (!pDrawLine) return 0;

    /* cursor is beyond the left margin */
    if (rc < 0) {
	goto showcur;
    }

    /* cursor is beyond the right margin */
    if (rc >= csize) {
	if (rc == csize) {
	    // special case
	    // cerr << "special case.\n";
	    nHackLine = spec->r;
	}
	goto showcur;
    }

    onechar[0] = (spec->c == tl->strlen) ? ' ' : tl->buf[spec->c];
    if (onechar[0] == '\t') onechar[0] = ' ';
    onechar[1] = '\0';

    /* get string width up to cursor position */
    swidth = fl_get_string_width(spec->text_style, spec->text_size, pDrawLine + spec->leftcol, rc);
    //swidth = fl_get_string_width(spec->text_style, spec->text_size, tl->buf + spec->leftcol, rc);

    tw = fl_get_string_width(spec->text_style, spec->text_size,onechar, 1);
    tx = x + swidth;
    ty = y + (spec->r - spec->topline) * th;

    /* erase the current line */
    fl_rectf(tx, ty, tw, th, tl->bgcolor);

    fl_drw_text_cursor(FL_ALIGN_LEFT, tx, ty, tw, th, tl->fgcolor, spec->text_style, spec->text_size, onechar, spec->ccol, -1);

showcur:

    tl = tb_get_lineptr_by_num(tb, r);
    if (!tl) {
	if (!r && !c) {
	    swidth = fl_get_string_width(spec->text_style,spec->text_size, " ", 1);
	    fl_drw_text_cursor(FL_ALIGN_LEFT, x, y, swidth, th, ob->col1, spec->text_style, spec->text_size, " ", spec->ccol, 0);
	    return 0;
	}
	return -1;
    }
    
    /* line is out of visible range */
    if ((r < spec->topline) || (r >= spec->topline+nlines)) {
	return -1;
    }

    if (c > tl->strlen) {
	c = tl->strlen;
    }
    rc = c - spec->leftcol;

    // we have to check wether the onechar is selected or not.
    // this is a lot trickier than it looks.
    int bSelected = FALSE;
    int sels = -1;
    int sele = -1;
    /* calculate selection */
    if ((r >= spec->sselr) &&
	(r <= spec->eselr)
    ){
	sels = 0;
	if (r == spec->sselr)
	    sels = spec->sselc;

	if (r == spec->eselr) {
	    sele = spec->eselc;
	    if (sels == sele) {
		sels = -1;
		sele = -1;	  
	    }
	}
    }
    //cerr << "sels, sele: " << sels << ", " << sele << endl;
    if ((sels == 0) && ((sele == -1) || (sele > c))) {
	bSelected = TRUE;    /* the whole line up to c is selected */
    } else {
	if ((sels == -1) || ((sele >= 0) && (sele <= sels))) {
	    /* line is not selected */
	} else {
	    if (sels <= c && (c < sele || sele < 0)) {
		bSelected = TRUE; /* line is partially selected */
	    }
	}
    }

    //cerr << "bSelected: " << bSelected << ", ";
    //cerr << spec->sselr << ", ";
    //cerr << spec->sselc << ", ";
    //cerr << spec->eselr << ", ";
    //cerr << spec->eselc << endl;

    pDrawLine = expand_tabs(tl->buf, rc, spec->leftcol);
    if (!pDrawLine) return 0;
    /*
    cerr << "leftcol: " << spec->leftcol << endl;
    cerr << "rc:      " << rc << endl;
    cerr << "pDraw+l:`" << pDrawLine + spec->leftcol << "'" << endl;
    */
    /* cursor is beyond the left margin */
    if (rc < 0) {
	// bug! Need to scroll here. 
	spec->r = r;
	spec->c = c;
	return -1; 
    }

    /* cursor is beyond the right margin */
    if (rc >= csize) {	// was: >=
	spec->r = r;
	spec->c = c;
	if (rc == csize) {
	    // special case
	    fl_textedit_draw_line(ob, spec->r);	// HACK
	}
	    
	//cerr << "move cursor says beyond right margin!\n";
	return -1;  
    }

    if (nHackLine >= 0) {
	int rtmp = spec->r, ctmp = spec->c;

	spec->r = r;
	spec->c = c;
	fl_textedit_draw_line(ob, nHackLine);	// HACK
	if (nHackLine == r) {
	    return 0;
	} else {
	    spec->r = rtmp;
	    spec->c = ctmp;
	}
    }


    onechar[0] = (c >= tl->strlen) ? ' ' : tl->buf[c];
    if (onechar[0] == '\t') onechar[0] = ' ';
    onechar[1] = '\0';

    /* get string width up to cursor position */
    swidth = fl_get_string_width(spec->text_style, spec->text_size, pDrawLine + spec->leftcol, rc);
    //swidth = fl_get_string_width(spec->text_style, spec->text_size, tl->buf + spec->leftcol, rc);

    tw = fl_get_string_width(spec->text_style, spec->text_size,onechar, 1);
    tx = x + swidth;
    ty = y + (r - spec->topline) * th;

    /* erase the current line */
    fl_rectf(tx, ty, tw, th, bSelected ? tl->fgcolor : tl->bgcolor);

    fl_drw_text_cursor(FL_ALIGN_LEFT, tx, ty, tw, th, 
	bSelected ? tl->bgcolor : tl->fgcolor, 
	spec->text_style, spec->text_size, onechar, spec->ccol, 0
    );

    spec->r = r;
    spec->c = c;

    return 0;
}

void
fl_calc_cursorpos(FL_OBJECT *ob, u_long cofft, int *cx, int *cy)
/*
 * Calculate cursor position from offset
 */
{
    SPEC    *spec = (SPEC *)ob->spec;
    TextBuf *tb;
    int i, len;
    char *line;

    tb = &spec->tb;
    *cx = 0;
    *cy = 0;
    /* scan all lines */
    for (i = 0; i < tb_get_nlines(tb);i++) {
	tb_get_line_by_num(tb, &line, i);
	if (!line) return;
	
	len = strlen(line) + 1;
	if ((u_long) len < cofft) {
	    cofft -= len;
	    (*cy)++; 
	} else {
	    *cx = cofft;
	    return;
	}
    }
}

void
fl_set_textedit_cursorpos(FL_OBJECT *ob, int cx, int cy, long cofft)
/*
 * set position of the cursor
 */
{
    SPEC    *spec = (SPEC *)ob->spec;
    TextBuf *tb;
    int r, c;
    char *line;
    Window wn;

    tb = &spec->tb;

    /* if offset specified then calculate the cursor position from it */
    if (cofft >= 0)
	fl_calc_cursorpos(ob, cofft, &cx, &cy);

    /* set the current line to the right one, but check if it's not
       out of range */
    if (spec->topline + cy >= tb_get_nlines(tb)) {
	return;
    }
    r = spec->topline + cy;
    tb_set_current_line(tb, spec->r);

    /* now calculate which character the cx position points to */
    /* first get the current line */
    tb_get_line(tb, &line);
    if (!line) {
	/* no line available */
	return;
    }

    c = (cx >= (int)strlen(line)) ? (int)strlen(line) : cx;
    spec->cpos = c;

    /* redraw the lines */
    wn = fl_winget();
    fl_winset(ob->form->window);
    fl_textedit_movecursor(ob, r, c);
    fl_winset(wn);
}

void
fl_get_textedit_cursorpos(FL_OBJECT *ob, int *cx, int *cy, u_long *cofft)
/*
 * get position of the cursor
 */
{
    SPEC    *spec = (SPEC *)ob->spec;
    char *line;
    int i;

    if (cx) {
	*cx = spec->c;
    }

    if (cy){
	*cy = spec->r;
    }

    if (cofft) 	{
	*cofft = 0;
	if (spec->r == 0) {
	    *cofft = spec->c;
	    return;     
	}

	for (i = 0; i < spec->r; i++) {
	    tb_get_line_by_num(&spec->tb, &line, i);
	    if (!line)
		break;

	    *cofft += strlen(line) + 1; 
	}

	*cofft += spec->c;
    }
}

int
fl_set_textedit_wrap(FL_OBJECT *ob, int wrap, int all)
/*
 * set word/line wrap
 */
{
    SPEC    *spec = (SPEC *)ob->spec;
    TextBuf *tb;
    int owrap;

    tb = &spec->tb;
    owrap = tb_get_linewrap(tb);
    tb_set_linewrap(tb, wrap);

    if (all && (tb_get_nlines(tb) > 0)) {
        tb_set_current_line(tb, 0);
        tb_wrap_line(tb);
        while (tb_set_next_line(tb))
	    tb_wrap_line(tb);
	fl_redraw_object(ob);
    }

    return owrap;
}

void
fl_insert_textedit(FL_OBJECT *ob, char *buf)
/*
 * insert <buf> at cursor position
 */
{
    SPEC    *spec = (SPEC *)ob->spec;
    TextBuf *tb;
    u_long offt;

    tb = &spec->tb;

    if (!buf || !*buf)
	return;

    fl_get_textedit_cursorpos(ob, NULL, NULL, &offt);

    tb_insert_block(tb, spec->r, spec->c, buf);
    if (strchr(buf, '\n'))	{
	if (spec->c > tb_get_linelen(tb))
	    spec->c = tb_get_linelen(tb);
    }
    else	{
	offt += strlen(buf);
	fl_calc_cursorpos(ob, offt, &spec->c, &spec->r);
    }
    fl_redraw_object(ob);

    update_vsb(ob);
    fl_textedit_set_heditscrollbar_wsize(ob);
}

void
fl_insert_textedit_file(FL_OBJECT *ob, char *fname)
/*
 * insert contents of file <fname> at cursor position
 */
{
    SPEC    *spec = (SPEC *)ob->spec;
    TextBuf *tb;

    tb = &spec->tb;
    if (!fname)
	return;

    tb_insert_file(tb, spec->r, spec->c, fname);
    if (spec->c > tb_get_linelen(tb))
	spec->c = tb_get_linelen(tb);
    fl_redraw_object(ob);

    update_vsb(ob);
    update_hsb(ob);
}

int
fl_textedit_readonly(FL_OBJECT *ob, int readonly)
/*
 * set/remove read-only flag
 */
{
    SPEC    *spec = (SPEC *)ob->spec;
    int oronly;

    oronly = spec->flags & FL_TEXTEDIT_READONLY;

    if (readonly)
	spec->flags |= FL_TEXTEDIT_READONLY;
    else
	spec->flags &= ~FL_TEXTEDIT_READONLY;

    return oronly;
}

int
fl_get_textedit_topline(FL_OBJECT *ob)
/*
 * Desc: return number of the first line displayed
 */
{
    SPEC    *spec = (SPEC *)ob->spec;

    return spec->topline;
}

int
fl_get_textedit_maxline(FL_OBJECT *ob)
/*
 * Desc: return number of lines in textbuffer
 */
{
    SPEC    *spec = (SPEC *)ob->spec;
    TextBuf *tb;

    tb = &spec->tb;

    return tb->n;
}

int
fl_get_textedit_screenlines(FL_OBJECT *ob)
/*
 * Desc: return number of visible lines 
 */
{
    return fl_textedit_get_wsize(ob);
}

char *
fl_get_textedit_curline(FL_OBJECT *ob)
/*
 * Desc: returns text of the line with cursor
 */
{
    SPEC    *spec = (SPEC *)ob->spec;
    TextBuf *tb;

    tb = &spec->tb;

    if (tb->currentline)
	return tb->currentline->buf;
    else
	return NULL;
}

char *
fl_get_textedit_line(FL_OBJECT *ob, int n)
/*
 * Desc: returns text of the line number <n>
 */
{
    SPEC    *spec = (SPEC *)ob->spec;
    TextBuf *tb;
    char *line;

    tb = &spec->tb;

    tb_get_line_by_num(tb, &line, n);

    return line;
}

void
fl_add_textedit_line(FL_OBJECT *ob, char *line)
/*
 * Desc: append line to text buffer
 */
{
    SPEC    *spec = (SPEC *)ob->spec;
    TextBuf *tb;
    Window wn;

    tb = &spec->tb;
    tb_append_line(tb, line);

    wn = fl_winget();
    fl_winset(ob->form->window);
    fl_textedit_refresh_screen(ob, 0);
    update_vsb(ob);
    update_hsb(ob);
    fl_winset(wn);
}

void
fl_set_textedit_topline(FL_OBJECT *ob, int n)
/*
 * Desc: set topline to <n>
 */
{
    SPEC    *spec = (SPEC *)ob->spec;
    Window wn;

    if ((n < 0) || (n > fl_get_textedit_maxline(ob)))
	return;

    spec->topline = n;
    spec->r = n;
    spec->c = 0;
    spec->cpos = 0;

    wn = fl_winget();
    fl_winset(ob->form->window);
    update_vsb(ob);
    fl_redraw_object(ob);
    fl_winset(wn);
}

void
fl_addto_textedit(FL_OBJECT *ob, char *line)
/*
 * Desc: append line to text buffer and make it visible
 */
{
    SPEC    *spec = (SPEC *)ob->spec;
    TextBuf *tb;
    int slines;
    Window wn;

    tb = &spec->tb;
    tb_append_line(tb, line);

    wn = fl_winget();
    fl_winset(ob->form->window);
    slines = fl_get_textedit_screenlines(ob);
    if (tb->n > (spec->topline + slines))
	fl_set_textedit_topline(ob, tb->n - slines);
    else
	fl_redraw_object(ob);

    update_vsb(ob);
    update_hsb(ob, TRUE);
#if 0
    if ((int)strlen(line) > (int)fl_get_editscrollbar_max(spec->hsb))
	fl_set_editscrollbar_max(spec->hsb, strlen(line));
#endif
    fl_winset(wn);

}

int
fl_textedit_line_visible(FL_OBJECT *ob, int line)
/*
 * Desc: check if <line> is visible
 */
{
    SPEC    *spec = (SPEC *)ob->spec;

    if ((line >= spec->topline) &&
	(line < (spec->topline + fl_get_textedit_screenlines(ob))))
	return 1;

    return 0;
}

void
fl_insert_textedit_line(FL_OBJECT *ob, int where, char *line)
/*
 * Desc: insert new line after line number <where>
 */
{
    SPEC    *spec = (SPEC *)ob->spec;
    TextBuf *tb;
    TextLine *tl;
    Window wn;

    tb = &spec->tb;
    if ((where < 0) || (where > tb->n))
	return;

    tl = tb->currentline;

    if (!tb_set_current_line(tb, where))
	return;

    tb_insert_line(tb, line);
    tb->currentline = tl;

    wn = fl_winget();
    fl_winset(ob->form->window);
    if (fl_textedit_line_visible(ob, where) || (where < spec->topline))
	fl_redraw_object(ob);

    update_vsb(ob);
    update_hsb(ob, TRUE);
#if 0
    if ((int)strlen(line) > (int)fl_get_editscrollbar_max(spec->hsb))
	fl_set_editscrollbar_max(spec->hsb, strlen(line));
#endif
    fl_winset(wn);
}

void
fl_delete_textedit_line(FL_OBJECT *ob, int line)
/*
 * Desc: delete line with number <line>
 */
{
    SPEC    *spec = (SPEC *)ob->spec;
    TextBuf *tb;
    TextLine *tl;
    Window wn;

    tb = &spec->tb;
    if ((line < 0) || (line > tb->n))
	return;

    tl = tb->currentline;

    if (!tb_set_current_line(tb, line))
	return;

    tb_del_line(tb);
    tb->currentline = tl;

    if (spec->r > tb->n)
	spec->r = tb->n;

    if (spec->topline > tb->n)
	spec->topline = tb->n;

    wn = fl_winget();
    fl_winset(ob->form->window);
    if (fl_textedit_line_visible(ob, line) || (line < spec->topline))
	fl_redraw_object(ob);

    update_vsb(ob);
    fl_winset(wn);
}

void
fl_replace_textedit_line(FL_OBJECT *ob, int line, char *nline)
/*
 * Desc: replace contents of line <line> with <nline>
 */
{
    SPEC    *spec = (SPEC *)ob->spec;
    Window wn;
    TextBuf *tb;
    TextLine *tl;
    int n;

    tb = &spec->tb;
    if ((line < 0) || (line > tb->n))
	return;

    tl = tb->currentline;

    if (!tb_set_current_line(tb, line))
	return;

    n = tb->n;
    tb_del_line(tb);
    tb_insert_line(tb, nline);

    tb->currentline = tl;

    wn = fl_winget();
    fl_winset(ob->form->window);
    if (n != tb->n) {
	if (fl_textedit_line_visible(ob, line) || (line < spec->topline))
	    fl_redraw_object(ob);

	update_vsb(ob);
    } else {
	fl_textedit_draw_line(ob, line);
    }

    update_hsb(ob, TRUE);
#if 0
    if ((int) strlen(nline) > (int)fl_get_editscrollbar_max(spec->hsb))
	fl_set_editscrollbar_max(spec->hsb, strlen(nline));
#endif
    fl_winset(wn);
}

void
fl_deselect_textedit(FL_OBJECT *ob)
/*
 * Desc: deselect any selected text
 */
{
    Window wn;

    wn = fl_winget();
    fl_winset(ob->form->window);
    fl_textedit_remove_selection(ob);
    fl_winset(wn);
}

int
fl_isselected_textedit_line(FL_OBJECT *ob, int line)
/*
 * Desc: check if <line> is selected (at least part of it)
 */
{
    SPEC    *spec = (SPEC *)ob->spec;

    if ((line >= spec->sselr) && (line <= spec->eselr)) 	{
	if ((line == spec->sselr) && (line == spec->eselr))	{
	    if (spec->sselc != spec->eselc)
		return 1;
	    else
		return 0;				}

	return 1;					}

    return 0;
}

void
fl_select_textedit_line(FL_OBJECT *ob, int line)
/*
 * Select <line>
 */
{
    SPEC    *spec = (SPEC *)ob->spec;
    Window wn;

    wn = fl_winget();
    fl_winset(ob->form->window);
    fl_textedit_remove_selection(ob);

    spec->sselr = spec->eselr = line;
    spec->sselc = 0;
    spec->eselc = -1;

    fl_textedit_draw_selection(ob);
    fl_winset(wn);
}

void
fl_select_textedit_all(FL_OBJECT *ob) {
    SPEC    *spec = (SPEC *)ob->spec;
    Window wn = fl_winget();
    fl_winset(ob->form->window);
    
    spec->sselr = 0;
    spec->eselr = fl_get_textedit_maxline(ob);
    spec->sselc = 0;
    spec->eselc = -1;

    fl_textedit_draw_selection(ob);
    fl_winset(wn);
}

void
fl_deselect_textedit_line(FL_OBJECT *ob, int)
/* int line not used yet.
 * Deselect <line>
 */
{
    fl_deselect_textedit(ob);
}

void
fl_set_textedit_line_color(FL_OBJECT *ob, int line, int bgcol, int fgcol)
/*
 * Desc: set back and foreground colors of <line>
 */
{
    SPEC    *spec = (SPEC *)ob->spec;
    TextBuf *tb;
    Window wn;

    tb = &spec->tb;

    if (!tb_set_current_line(tb, line))
	return;

    if (bgcol > 0)
	tb_set_linebgcolor(tb, bgcol);

    if (fgcol > 0)
	tb_set_linefgcolor(tb, fgcol);

    wn = fl_winget();
    fl_winset(ob->form->window);
    fl_textedit_draw_line(ob, line);
    fl_winset(wn);
}

void
fl_get_textedit_line_color(FL_OBJECT *ob, int line, int *bgcol, int *fgcol)
/*
 * Desc: get back and foreground colors of <line>
 */
{
    SPEC    *spec = (SPEC *)ob->spec;
    TextBuf *tb;
    TextLine *tl;

    tb = &spec->tb;

    if ((tl = tb_get_lineptr_by_num(tb, line)) == NULL)
	return;

    if (bgcol)
	*bgcol = tl->bgcolor;

    if (fgcol)
	*fgcol = tl->fgcolor;
}

int 
fl_textedit_cut(FL_OBJECT *ob)
/*
 * Cut selected text from textarea to X-servers buffer
 * copy it first, then delete it from text.
 */ 
{
    if (fl_textedit_copy(ob)) {
	delete_selected_block(ob);
	return TRUE;
    }
    return FALSE;
}

static void
delete_selected_block(FL_OBJECT *ob){
    SPEC    *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    Window wn = fl_winget();
    int n;

    fl_winset(ob->form->window);
    if (tb_del_block(tb, spec->sselr, spec->sselc,
	spec->eselr, spec->eselc)) {
	//cerr << "del: delblock.\n";
	while (spec->sselr &&
	    !tb_set_current_line(tb, spec->sselr))
	    spec->sselr--;

	spec->r = spec->sselr;
	spec->c = spec->sselc;
	if (spec->c > tb_get_linelen(tb))
	    spec->c = tb_get_linelen(tb);
	spec->sselr = -1;
	spec->eselr = -1;
	n = tb_get_nlines(tb);
	if (spec->topline > spec->r)
	    spec->topline = spec->r;

	good_redraw(ob);
    } else {
	//cerr << "del: remove sel.\n";
	spec->c = spec->sselc;
	if (spec->c > tb_get_linelen(tb))
	    spec->c = tb_get_linelen(tb);
	fl_textedit_remove_selection(ob);
    }
    fl_winset(wn);
}
    
void fl_get_textedit_selectionpos(
    FL_OBJECT *ob, 
    int *start_col, int *start_row, 
    int *end_col, int *end_row
){
    SPEC *sp = (SPEC *)ob->spec;
    *start_row = sp->sselr;
    *start_col = sp->sselc;
    *end_row   = sp->eselr;
    *end_col   = sp->eselc;	
}

int fl_get_textedit_n_sellines(FL_OBJECT *ob){
    SPEC *sp = (SPEC *)ob->spec;
    if (!fl_textedit_selected(ob)) return 0;
    return abs(sp->eselr - sp->sselr) + 1;
}

static void good_redraw(FL_OBJECT *ob){
    update_vsb(ob);
    update_hsb(ob);

    // draw the bugger:
    fl_textedit_draw_screen(ob);
}	

/* EDITSCROLLBAR FUNCTIONS */

static int update_hsb(
    FL_OBJECT *ob, 
    int bForceCursorVisible,
    int bNoRedraw
){
    SPEC    *spec   = (SPEC *)ob->spec;
    int     csize   = fl_textedit_get_csize(ob);
    int     llength = fl_textedit_get_llength(ob);
    int     nOldLeftCol = spec->leftcol;

    // normalize leftcol
    int    nColMax = llength - csize;
    if (nColMax < 0) nColMax = 0;

    if (bForceCursorVisible){
	if (spec->c >= spec->leftcol + csize) {
	    spec->leftcol = spec->c - csize;
	} else if (spec->c < spec->leftcol) {
	    spec->leftcol = spec->c;
	}
    }

    if (spec->leftcol > nColMax) spec->leftcol = nColMax;
    if (spec->leftcol < 0) spec->leftcol = 0;

    // update the new ColMaximum, leftcol and cwindow size:
    //cerr << "HParams: nColMax, val, csize: " << nColMax << ", "
    //     << spec->leftcol << ", " << csize << endl;
    double min, max, val, wsize;
    fl_get_editscrollbar_parameters(spec->hsb, min, max, val, wsize);
    if (min == (double) 0 && max == (double) nColMax && 
        val == (double) spec->leftcol && wsize == (double) csize
    ){
      //cerr << "Skipping hsb redraw.\n";
    } else {
	fl_set_editscrollbar_parameters(spec->hsb, 
	    0, nColMax, spec->leftcol, csize
        );
	fl_redraw_object(spec->hsb);
    }
    if (nOldLeftCol != spec->leftcol && !bNoRedraw) {
	fl_textedit_draw_screen(ob);
	return TRUE;
    } else {
	return FALSE;
    }
}

static void update_vsb(FL_OBJECT *ob){
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf    *tb = &spec->tb;
    int      wsize = fl_textedit_get_wsize(ob);

    // normalize topline 
    int    nMax = tb->n - wsize;
    if (nMax < 0) nMax = 0;
    if (spec->topline > nMax) spec->topline = nMax;

    // update the new maximum, topline and window size:
    // cerr << "VParams: nMax, val, wsize: " << nMax << ", "
    //     << spec->topline << ", " << wsize << endl;
    fl_set_editscrollbar_parameters(spec->sb, 
	0, nMax, spec->topline, wsize
    );
    fl_redraw_object(spec->sb);
}
