/***********************************************************************
 *
 * Module: fl_editscrollbar.c
 * Author: Marc van Kempen (marc@bowtie.nl)
 * Desc:   General editscrollbar-implementation. I don't like the current
 *	   slider for editscrollbar behaviour.
 *
 * Copyright (c) 1996, Marc van Kempen
 *
 * All rights reserved.
 *
 * This software may be used, modified, copied, distributed, and
 * sold, in both source and binary form provided that the above
 * copyright and these terms are retained, verbatim, as the first
 * lines of this file.  Under no circumstances is the author
 * responsible for the proper functioning of this software, nor does
 * the author assume any responsibility for damages incurred with
 * its use.
 *
 ***********************************************************************/

/*       $Id: fl_editscrollbar.C,v 1.8 1997/04/21 17:33:53 henrik Exp henrik $
 */

/*
 * What I need: 
 * 
 * - The range for the editscrollbar must be settable. The total # of lines
 *   (or whatever) and the stepsize, as well as the window size.
 *   When the user clicks above or under the editscrollbar the editscrollbarbutton
 *   will be moved by a corresponding amount. The editscrollbarobject will 
 *   invoke its callback function when this happens and the user
 *   can request the current lineposition.
 *
 * - the location of the editscrollbar must be settable
 *
 */

#include <sys/param.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <iostream.h>
#include <forms.h>
#include "fl_editscrollbar.h"
#include "error_notifier_forms_aux.h"
#include "utility.h"
#include "global.h"

#define MIN_SCRL_HEIGHT 30.0	/* the minimum height of the veditscrollbar */
#define MIN_SCRL_WIDTH	30.0	/* the minimum width of the heditscrollbar */
void fl_draw_object_label();

typedef enum {DRAW_NONE, DRAW_NORMAL, DRAW_PUSHED} draw_et;
typedef enum {PUSHED_NONE, PUSHED_UPPER, PUSHED_LOWER} state_et;

typedef struct {
    double	min;		/* minimum value */
    double	max;		/* maximum value */
    double	step;		/* stepsize when clicking the arrows */
    double	wsize;		/* windowsize */
    double	val;		/* current editscrollbar value */
    int		coord;		/* the x/y-coordinate of the editscrollbar */
    state_et pushed;    	/* none, upper, lower */
    int      bBarPushed;	/* no button, but in area */
    int      bBarUp;	        /* are we moving up? */
    int      dragging;  /* true if dragging with button */
} SPEC;

/* change state and draw */
static void
change_draw_buttons(FL_OBJECT *ob, draw_et upper_state, draw_et lower_state, int bDrawCorner);

/* draw only */
static void
draw_buttons(FL_OBJECT *ob);

static int
fix_boxtype(FL_OBJECT *ob){
    return (ob->boxtype == FL_DOWN_BOX) ? FL_UP_BOX : FL_DOWN_BOX;
}

/***********************************************************************
 *
 * The dimensions of the editscrollbar.
 *
 * get the editscrollbarbutton dimensions with 
 *
 *    fl_editscrollbar_get_bbox(ob, &bx, &by, &bw, &bh)
 *
 *  +------------------------------------------------------------+
 *  |	
 *  |     ob->x
 *  |      |
 *  |      |+--- bx = ob->x + ob->bw
 *  |      v|
 *  |       V
 *  |      +------+  <-- ob->y
 *  |      |      |  <-- ob->y + ob->bw
 *  |      |      |                                                
 *  |by -> |+----+|  ^
 *  |      ||    ||  |                                              
 *  |      ||    ||  |                                              
 *  |      ||    ||  | bh			
 *  |      ||    ||  |
 *  |      |+----+|  v
 *  |      |      |
 *  |      |      |
 *  |      |      |  <-- ob->y + ob->bw + ob->h
 *  |      +------+  <-- ob->y + 2*ob->bw + ob->h
 *  |      <------> ob->w
 *  |       <----> bw
 *  |
 *  +------------------------------------------------------------+
 *
 ***********************************************************************/

/* the coordinates are changed for the arrow buttons. The Arrow buttons
 * are always square, with the side length of the square equal to ob->w
 * for vertical editscrollbars, and equal to ob->h for horizontal ones.
 * The bbox calculations and drawing routines are adapted accordingly.
 */

/***********************************************************************
 *
 * Misc functions
 *
 ***********************************************************************/

#ifndef MAX
#define MAX(a, b)      ((a) < (b) ? (b) : (a))
#endif
#ifndef MIN
#define MIN(a, b)      ((a) < (b) ? (a) : (b))
#endif

static void
fl_editscrollbar_get_bbox(FL_OBJECT *ob, int *bx, int *by, int *bw, int *bh)
/*
 * Desc: calculate bounding box of veditscrollbar button 
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    /* pretend ob has different y,h */
    int obh, oby;
    oby = ob->y + ob->w;
    obh = ob->h - 2*ob->w - 2*abs(ob->bw);


    *bw = ob->w - 2*abs(ob->bw);
    if (fabs(spec->max - spec->min) < DBL_MIN) {
	//cerr << "direct obh\n";
	*bh = obh;
    } else {
	//cerr << "calculated wsize, max, min: " << spec->wsize;
	//cerr << ", " << spec->max << ", " << spec->min << "\n";
	*bh = (int) MAX(MIN_SCRL_HEIGHT, 
	    (double) spec->wsize/
	    (spec->wsize + spec->max - spec->min) * obh
	);
	*bh = MIN(*bh, obh);
    }
    *bx = ob->x + abs(ob->bw);
    *by = spec->coord + oby + abs(ob->bw);

    if (*bh <= 0) *bh = 1;
} /* fl_editscrollbar_get_bbox() */

static void
fl_heditscrollbar_get_bbox(FL_OBJECT *ob, int *bx, int *by, int *bw, int *bh)
/*
 * Desc: calculate bounding box of heditscrollbar button 
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    /* pretend ob has different x,w */
    int obx, obw;
    obx = ob->x + ob->h;
    obw = ob->w - 2*ob->h;

    *bh = ob->h - 2*abs(ob->bw);
    if (fabs(spec->max - spec->min) < DBL_MIN) {
	*bw = obw - 2*abs(ob->bw);
    } else {
	*bw = (int) MAX(MIN_SCRL_HEIGHT, 
		  (double) spec->wsize/(spec->wsize + spec->max - spec->min) * 
		  (obw - 2*abs(ob->bw)));
	*bw = MIN(*bw, obw - 2*abs(ob->bw));
    }
    *by = ob->y + abs(ob->bw);
    *bx = spec->coord + obx + abs(ob->bw);

    if (*bw <= 0) *bw = 1;

    return;
} /* fl_editscrollbar__get_bbox() */

static void
fl_editscrollbar_set_y(FL_OBJECT *ob, int y)
/*
 * Desc: set the y value of the editscrollbar and calculate the 
 *	corresponding editscrollbarvalue
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    int bx, by, bw, bh, h;

    /* get the bounding box of the editscrollbar button */
    fl_editscrollbar_get_bbox(ob, &bx, &by, &bw, &bh);

    /* the inner height of the editscrollbar */
    h = ob->h - 2*abs(ob->bw) - bh - 2*ob->w;

    /* check if the y-value is sane and set the new coordinate */
    if (y > h) y = h;
    if (y < 0) y = 0;

    /* 0 <= y <= ob->h - bh */
    spec->coord = y;

    /* calculate the new value based upon the y-coordinate */
    if ((fabs(spec->max - spec->min) < DBL_MIN) || (h == 0)) {
	spec->val = spec->min;
    } else {
	spec->val = (double) spec->coord/h * (spec->max - spec->min)
	    + spec->min;
    }

    /* now modulate by spec->step */
    spec->val = ((int) spec->val / spec->step) * spec->step;

    return;
} /* fl_editscrollbar_set_y() */

static void
fl_editscrollbar_set_x(FL_OBJECT *ob, int x)
/*
 * Desc: set the x value of the heditscrollbar and calculate the 
 *	corresponding editscrollbarvalue
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    int bx, by, bw, bh, w;

    /* get the bounding box of the editscrollbar button */
    fl_heditscrollbar_get_bbox(ob, &bx, &by, &bw, &bh);

    /* the inner width of the editscrollbar */
    w = ob->w - 2*abs(ob->bw) - bw - 2*ob->h;

    /* check if the x-value is sane and set the new coordinate */
    if (x > w) x = w;
    if (x < 0) x = 0;

    /* 0 <= x <= ob->w - bw */
    spec->coord = x;

    /* calculate the new value based upon the y-coordinate */
    if ((fabs(spec->max - spec->min) < DBL_MIN) || (w == 0)) {
	spec->val = spec->min;
    } else {
	spec->val = (double) spec->coord/w * (spec->max - spec->min)
	    + spec->min;
    }

    /* now modulate by spec->step */
    spec->val = ((int) spec->val / spec->step) * spec->step;

    return;
} /* fl_editscrollbar_set_x() */

static int
fl_editscrollbar_set_val(FL_OBJECT *ob, double val)
/*
 * Desc: set the value of the editscrollbar and calculate the 
 *	corresponding x/y coordinate
 * return TRUE if value changed. 
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    int bx, by, bw, bh;

    /* check if the value is sane */
    if (val < spec->min) val = spec->min;
    if (val > spec->max) val = spec->max;

    //cerr << "set_val " << val << "\n";
    if (val == spec->val) {
	//cerr << "equal warning!\n";
	return 0;
    }
    spec->val = val;

    /* get the bounding box of the editscrollbar button */
    if (ob->type == FL_VERTICAL_EDITSCROLLBAR){
	fl_editscrollbar_get_bbox(ob, &bx, &by, &bw, &bh);
    } else {
	fl_heditscrollbar_get_bbox(ob, &bx, &by, &bw, &bh);
    }

    /* calculate the new x/y-coordinate based upon the value */
    if ((fabs(spec->max - spec->min) < DBL_MIN)) {
	spec->coord = 0;
	//cerr << "coord zero." << endl;
    } else {
	switch (ob->type) {
	case FL_VERTICAL_EDITSCROLLBAR:
	    spec->coord = 
		(int) ((spec->val - spec->min)/(spec->max - spec->min) *
		    (ob->h - bh - 2*abs(ob->bw) - 2*ob->w)
		);
	    //cerr << "vert coord: " << spec->coord << endl;
	    break;

	case FL_HORIZONTAL_EDITSCROLLBAR:
	    spec->coord = 
		(int) ((spec->val - spec->min)/(spec->max - spec->min) *
		    (ob->w - bw - 2*abs(ob->bw) -2*ob->h)
		);
	    //cerr << "horz coord: " << spec->coord << endl;
	    break;
	default:
	    cerr << "unknown editscrollbar type." << endl;
	    break;
	}
    }

    return 1;
} /* fl_editscrollbar_set_val() */

/***********************************************************************
 *
 * Object drawing routines
 *
 ***********************************************************************/

void
fl_editscrollbar_draw(FL_OBJECT *ob)
/*
 * Desc: draw the editscrollbar in its current form
 */
{
    int bx, by, bw, bh;
    int x, y, w, h;

    if (ob->form->frozen) {
	return;
    }

    if (ob->type == FL_VERTICAL_EDITSCROLLBAR) {
      fl_drw_box(FL_DOWN_BOX, ob->x, ob->y+ob->w, ob->w, ob->h-2*ob->w, ob->col2, abs(ob->bw));
      fl_editscrollbar_get_bbox(ob, &bx, &by, &bw, &bh);
      fl_drw_box(FL_UP_BOX, bx, by, bw, bh, ob->col2, abs(ob->bw));
      
      x = bx + abs(ob->bw) + 1;
      y = by + bh/2;
      w = bw - 3*abs(ob->bw) - 1;
    
      fl_line(x, y-5, x+w,  y-5, FL_WHITE);
      fl_line(x, y-4, x+w,  y-4, FL_BLACK);
      
      fl_line(x, y,   x+w,  y,   FL_WHITE);
      fl_line(x, y+1, x+w,  y+1, FL_BLACK);
      
      fl_line(x, y+4, x+w,  y+4, FL_WHITE);
      fl_line(x, y+5, x+w,  y+5, FL_BLACK); 
    } else {
      fl_drw_box(FL_DOWN_BOX, ob->x+ob->h, ob->y, ob->w-2*ob->h, ob->h, ob->col2, abs(ob->bw));
      fl_heditscrollbar_get_bbox(ob, &bx, &by, &bw, &bh);
      fl_drw_box(FL_UP_BOX, bx, by, bw, bh, ob->col2, abs(ob->bw));

      y = by + abs(ob->bw) + 1;
      x = bx + bw/2;
      h = bh - 3*abs(ob->bw) - 1;
      
      fl_line(x-5, y, x-5, y+h, FL_WHITE);
      fl_line(x-4, y, x-4, y+h, FL_BLACK);
      
      fl_line(x,   y, x,   y+h, FL_WHITE);
      fl_line(x+1, y, x+1, y+h, FL_BLACK);
      
      fl_line(x+4, y, x+4, y+h, FL_WHITE);
      fl_line(x+5, y, x+5, y+h, FL_BLACK); 
    }
    
    draw_buttons(ob);
    return;
} /* fl_editscrollbar_draw() */

static void
fl_editscrollbar_fastdraw(FL_OBJECT *ob, int oldcoord)
/*
 * Desc: only redraw the editscrollbarbutton
 * Pre:  0 <= oldcoord <= "editscrollbarlength"
 */
{
    int bx, by, bw, bh;
    int x, y, w, h, limit;

    if (ob->form->frozen) {
	return;
    }

    if (ob->type == FL_VERTICAL_EDITSCROLLBAR) {
      fl_editscrollbar_get_bbox(ob, &bx, &by, &bw, &bh);
      if (oldcoord < 0) oldcoord = 0;
      limit = ob->h - 2*abs(ob->bw) - bh - ob->w;
      if (oldcoord > limit) oldcoord = limit;
      oldcoord += ob->y + abs(ob->bw) + ob->w;
      
      fl_drw_box(FL_UP_BOX, bx, by, bw, bh, ob->col2, abs(ob->bw));
      if (oldcoord < by) {
	fl_rectf(bx, oldcoord, bw, by-oldcoord, ob->col2);
      } else {
	/* oldcoord >= by */
	fl_rectf(bx, by+bh, bw, oldcoord-by, ob->col2);
      }
      
      x = bx + abs(ob->bw) + 1;
      y = by + bh/2;
      w = bw - 3*abs(ob->bw);
      
      fl_line(x, y-5, x+w,  y-5, FL_WHITE);
      fl_line(x, y-4, x+w,  y-4, FL_BLACK);
      
      fl_line(x, y,   x+w,  y,   FL_WHITE);
      fl_line(x, y+1, x+w,  y+1, FL_BLACK);
      
      fl_line(x, y+4, x+w,  y+4, FL_WHITE);
      fl_line(x, y+5, x+w,  y+5, FL_BLACK);
    } else {
      fl_heditscrollbar_get_bbox(ob, &bx, &by, &bw, &bh);
      if (oldcoord < 0) oldcoord = 0;
      limit = ob->w - 2*abs(ob->bw) - bw - ob->h;
      if (oldcoord > limit) oldcoord = limit;
      oldcoord += ob->x + abs(ob->bw) + ob->h;
      
      fl_drw_box(FL_UP_BOX, bx, by, bw, bh, ob->col2, abs(ob->bw));
      if (oldcoord < bx) {
	fl_rectf(oldcoord, by, bx-oldcoord, bh, ob->col2);
      } else {
	/* oldcoord >= bx */
	fl_rectf(bx+bw, by, oldcoord-bx, bh, ob->col2);
      }
      
      y = by + abs(ob->bw) + 1;
      x = bx + bw/2;
      h = bh - 3*abs(ob->bw) - 1;
      
      fl_line(x-5, y, x-5, y+h, FL_WHITE);
      fl_line(x-4, y, x-4, y+h, FL_BLACK);
      
      fl_line(x,   y, x,   y+h, FL_WHITE);
      fl_line(x+1, y, x+1, y+h, FL_BLACK);
      
      fl_line(x+4, y, x+4, y+h, FL_WHITE);
      fl_line(x+5, y, x+5, y+h, FL_BLACK);
    }
    draw_buttons(ob);
    return;
} /* fl_editscrollbar_fastdraw(ob) */    
    

/***********************************************************************
 *
 * Object definition/handling routines
 *
 ***********************************************************************/

static void
fl_heditscrollbar_drag(FL_OBJECT *ob, int mx, int)
/* int my not needed.
 * Desc: drag the editscrollbar
 */
{
    int bx, by, bw, bh, dx, oldx;
    double oldval;
    SPEC *spec = (SPEC *)ob->spec;

    /* get the bounding box of the editscrollbar button */
    fl_heditscrollbar_get_bbox(ob, &bx, &by, &bw, &bh);
    if ((mx > bx+bw || mx < bx) && !spec->dragging) return;
	spec->dragging = TRUE;

    oldx = spec->coord;
    dx = mx - (bx + bw/2);  /* initial offset of y position */
    if (dx == 0)
	return;
    oldval = spec->val;
    /* set the new editscrollbar x-coordinate and calculate the
       corresponding value */
    fl_editscrollbar_set_x(ob, spec->coord + dx);
    /* check to see if spec->val was changed => callback */
    if (oldval != spec->val) {
	if (ob->object_callback)
	    ob->object_callback(ob, ob->argument);
	    }
    fl_editscrollbar_fastdraw(ob, oldx);

    return;
} /* fl_hcrollbar_drag() */

static int
fl_heditscrollbar_click(FL_OBJECT *ob, int mx, int my)
/*
 * Desc: react on a click on the editscrollbar somewhere
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    int bx, by, bw, bh;

    /* first find out if we click at the left, on or at the right of the
     * editscrollbar button. For this we need to calculate
     * the size of the button
     */
    fl_heditscrollbar_get_bbox(ob, &bx, &by, &bw, &bh);

    if (my <= by && my >= by+bh) {
      return (0);
    }

    /* leftarrow button ? */
    if (mx>ob->x && mx<ob->x + ob->h) {
	FL_Coord oldx = spec->coord;
	if (fl_editscrollbar_set_val(ob, spec->val - spec->step)) {
	    change_draw_buttons(ob, DRAW_PUSHED, DRAW_NONE, FALSE);
	    fl_editscrollbar_fastdraw(ob, oldx);
	    return 1;
	}
	return 0; 
    }
    /* rightarrow button ? */
    if (mx>ob->x + ob->w - ob->h && mx<ob->x + ob->w) {
	FL_Coord oldx = spec->coord;
	if (fl_editscrollbar_set_val(ob, spec->val + spec->step)) {
	    change_draw_buttons(ob, DRAW_NONE, DRAW_PUSHED, FALSE); 
	    fl_editscrollbar_fastdraw(ob, oldx);
	    return 1;
	}
	return 0; 
    }	

    /* left of the editscrollbar button ? */
    if ((my > by && my < by+bh) && (mx<bx)) {
	/* page left */
	fl_editscrollbar_set_val(ob, spec->val - spec->wsize);
	fl_editscrollbar_draw(ob);
	return(1);
    }
    
    /* right of the editscrollbar button ? */
    if ((my > by && my < by+bh) && (mx>bx+bw)) {
	/* page right */
	fl_editscrollbar_set_val(ob, spec->val + spec->wsize);
	fl_editscrollbar_draw(ob);
	return(1);
    }
    
    /* we should never reach this point, but the compiler doesn't know */
    return(0);
} /* fl_heditscrollbar_click() */

static void
fl_editscrollbar_drag(FL_OBJECT *ob, int, int my)
/* int mx not needed.
 * Desc: drag the editscrollbar
 */
{
    int bx, by, bw, bh, dy, oldy;
    double oldval;
    SPEC *spec = (SPEC *)ob->spec;

    /* get the bounding box of the editscrollbar button */
    fl_editscrollbar_get_bbox(ob, &bx, &by, &bw, &bh);
    if ((my > by+bh || my < by) && !spec->dragging) return;
	spec->dragging = TRUE;

    oldy = spec->coord;
    dy = my - (by + bh/2);  /* initial offset of y position */
    if (dy == 0) return;
    oldval = spec->val;
    /* set the new editscrollbar y-coordinate and calculate the
       corresponding value */
    fl_editscrollbar_set_y(ob, spec->coord + dy);
    /* check to see if spec->val was changed => callback */
    if (oldval != spec->val) {
	if (ob->object_callback)
	    ob->object_callback(ob, ob->argument);
	    }
    fl_editscrollbar_fastdraw(ob, oldy);

    return;
} /* fl_editscrollbar_drag() */

static int
fl_editscrollbar_click(FL_OBJECT *ob, int mx, int my)
/*
 * Desc: react on a click on the editscrollbar somewhere
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    int bx, by, bw, bh;

    /* first find out if we click above, on or below the
     * editscrollbar button. For this we need to calculate
     * the size of the button
     */
    fl_editscrollbar_get_bbox(ob, &bx, &by, &bw, &bh);

    if (mx <= bx || mx>=bx+bw) {
	return 0;
    }

    /* uparrow button ? */
    if (my>ob->y && my<ob->y + ob->w) {
	FL_Coord oldy = spec->coord;
	if (fl_editscrollbar_set_val(ob, spec->val - spec->step)) {
	    change_draw_buttons(ob, DRAW_PUSHED, DRAW_NONE, FALSE);
	    fl_editscrollbar_fastdraw(ob, oldy);
	    return 1;
	}
	return 0; 
    }
    /* downarrow button ? */
    if (my>ob->y + ob->h - ob->w && my<ob->y + ob->h) {
	FL_Coord oldy = spec->coord;
	if (fl_editscrollbar_set_val(ob, spec->val + spec->step)) {
	    change_draw_buttons(ob, DRAW_NONE, DRAW_PUSHED, FALSE); 
	    fl_editscrollbar_fastdraw(ob, oldy);
	    return 1;
	}
	return 0; 
    }	

    /* above editscrollbar button ? */
    if (my<by) {
	/* page up */
	fl_editscrollbar_set_val(ob, spec->val - spec->wsize);
	fl_editscrollbar_draw(ob);
	spec->bBarPushed = TRUE;
	spec->bBarUp = TRUE;
	return(1);
    }
    
    /* on editscrollbar button ? */
/*
    if (my>=by && my<=by+bh) {
	fl_editscrollbar_drag(ob, mx, my);
	return(0);
    }
*/

    /* below the editscrollbar button ? */
    if (my>by+bh) {
	/* page down */
	fl_editscrollbar_set_val(ob, spec->val + spec->wsize);
	fl_editscrollbar_draw(ob);
	spec->bBarPushed = TRUE;
	spec->bBarUp = FALSE;
	return(1);
    }
    
    /* we should never reach this point, but the compiler doesn't know */
    return(0);
} /* fl_editscrollbar_click() */

static int
handle_editscrollbar(FL_OBJECT *ob, int event, FL_Coord mx, FL_Coord my, 
		 int, void *)
/* int key and void *xev not needed.
 * Desc: the event handler for the editscrollbar object 
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    static int mouse_count = 0;
    if (event != FL_MOUSE) mouse_count = 0;
    if (event != FL_MOUSE) spec->dragging = FALSE;

    switch(event) {
    case FL_DRAW:
	fl_editscrollbar_draw(ob);
	/* fall through */
    case FL_DRAWLABEL:
	fl_draw_object_label(ob);
	return(0);
	break;
    case FL_PUSH:  /* a mouse button was pushed */
	if (ob->type == FL_VERTICAL_EDITSCROLLBAR)
	 return(fl_editscrollbar_click(ob, mx, my));
	else
	 return(fl_heditscrollbar_click(ob, mx, my));
	break;
    case FL_RELEASE:
	spec->bBarPushed = FALSE;	// allways reset. 
	if (spec->pushed == PUSHED_UPPER) {
	    change_draw_buttons(ob, DRAW_NORMAL, DRAW_NONE, FALSE);
	    return 0;
	} else if (spec->pushed == PUSHED_LOWER) {
	    change_draw_buttons(ob, DRAW_NONE, DRAW_NORMAL, FALSE);
	    return 0;
	}
	break;
    case FL_MOUSE:
	/* fprintf(stderr, "fl_mouse\n");  */
	if (spec->pushed != PUSHED_NONE) {
	    mouse_count++;
	    if (mouse_count > 10) { 
		int jump = MIN(mouse_count - 10, 20);
		double step = spec->step * jump;
		if (spec->pushed == PUSHED_UPPER) {
		    if (fl_editscrollbar_set_val(ob, spec->val - step)) {
			fl_editscrollbar_draw(ob);
			sginap(5);
			return 1;
		    }	
		    return 0;
		} else if (spec->pushed == PUSHED_LOWER) {
		    if (fl_editscrollbar_set_val(ob, spec->val + step)) {
			fl_editscrollbar_draw(ob);
			sginap(5);
			return 1;
		    }
		    return 0;
		}
	    }
	    return 0;
	}
	if (spec->bBarPushed && ob->type == FL_VERTICAL_EDITSCROLLBAR) {
	    mouse_count++;
	    if (mouse_count > 15 && (mouse_count % 2 == 0)) { 
		if (spec->bBarUp) {
		    fl_editscrollbar_set_val(ob, spec->val - spec->wsize);
		} else {
		    fl_editscrollbar_set_val(ob, spec->val + spec->wsize);
		}
		fl_editscrollbar_draw(ob);
	    }
	    return 1;
	}
	if (ob->type == FL_VERTICAL_EDITSCROLLBAR) {
	    fl_editscrollbar_drag(ob, mx, my);
	} else {
	    fl_heditscrollbar_drag(ob, mx, my);
	}
	break;
    case FL_KEYBOARD:
	return(0);
	break;
    case FL_FREEMEM:
	/* free the memory used by the spec member of ob */
	{
	    SPEC *spec = (SPEC *)ob->spec;
	    free(spec);
	}	
	return(0);
	break;
    }
    
    return(0);
} /* handle_editscrollbar() */

static FL_OBJECT *
fl_create_editscrollbar(int type, FL_Coord x, FL_Coord y, FL_Coord w, FL_Coord h,
        	        	    const char *label)
/*
 * Desc: create the editscrollbar object 
 */
{
    FL_OBJECT *ob;
    SPEC *spec;

    ob = fl_make_object(FL_EDITSCROLLBAR, type, x, y, w, h, 
			label, handle_editscrollbar);

    ob->col1 = FL_BLACK;
    ob->boxtype = FL_UP_BOX;
    ob->resize = (type == FL_VERTICAL_EDITSCROLLBAR) ? FL_RESIZE_Y : FL_RESIZE_X;

    /* initialize the object specific user data */
    ob->spec = (SPEC *) fl_calloc(1, sizeof(SPEC));
    spec = (SPEC *)ob->spec;
    if (!ob->spec) {
	fl_error_msg("Could not malloc space for SPEC structure in object:%s", 
		 label);
	exit(1);
    }

    spec->coord = 0;
    spec->val = 0.0;
    spec->min = 0.0;
    spec->max = 1.0;
    spec->step = 1.0;
    spec->wsize = 1.0;
    spec->pushed = PUSHED_NONE;
    spec->dragging = FALSE;
    spec->bBarPushed = FALSE;

    return(ob);
} /* fl_create_editscrollbar() */

/***********************************************************************
 *
 * Object API
 *
 ***********************************************************************/

FL_OBJECT *
fl_add_editscrollbar(int type, FL_Coord x, FL_Coord y, 
		 FL_Coord w, FL_Coord h, const char *label)
{
    FL_OBJECT *ob = fl_create_editscrollbar(type, x, y, w, h, label);

    fl_add_object(fl_current_form, ob);

    return(ob);
} /* fl_add_editscrollbar() */

void
fl_set_editscrollbar_min(FL_OBJECT *ob, double min)
/*
 * Desc: set the minimum size
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    spec->min = min;
    if (spec->val < min)
	fl_editscrollbar_set_val(ob, min);
    fl_redraw_object(ob);

    return;
} /* fl_set_editscrollbar_min() */

double
fl_get_editscrollbar_max(FL_OBJECT *ob)
/*
 * Desc: return maximum value of the editscrollbar
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    return spec->max;
}

double
fl_get_editscrollbar_wsize(FL_OBJECT *ob)
/*
 * Desc: return wsize value of the editscrollbar
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    return spec->wsize;
}

void
fl_set_editscrollbar_max(FL_OBJECT *ob, double max)
/*
 * Desc: set the maximum size
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    spec->max = max;
    if (spec->val > max)
	fl_editscrollbar_set_val(ob, max);
    fl_redraw_object(ob);

    return;
} /* fl_set_editscrollbar_max() */

void
fl_set_editscrollbar_step(FL_OBJECT *ob, double step)
/*
 * Desc: set the stepsize
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    
    spec->step= step;

    return;
} /* fl_set_editscrollbar_step() */

void
fl_set_editscrollbar_wsize(FL_OBJECT *ob, double wsize)
/*
 * Desc: set the window size
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    spec->wsize = wsize;
    fl_redraw_object(ob);

    return;
} /* fl_set_editscrollbar_wsize() */

void
fl_set_editscrollbar_wsizeval(FL_OBJECT *ob, double wsize)
/*
 * Desc: set the window size value
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    spec->wsize = wsize;

    return;
} /* fl_set_editscrollbar_wsizeval() */

void
fl_set_editscrollbar_val(FL_OBJECT *ob, double val)
/*
 * Desc: set the editscrollbar value'
 */
{
    int oldcoord;
    SPEC *spec = (SPEC *)ob->spec;

    oldcoord = spec->coord;
    fl_editscrollbar_set_val(ob, val);
    if (spec->coord != oldcoord)
	fl_editscrollbar_fastdraw(ob, oldcoord);

    return;
} /* fl_set_editscrollbar_val() */

double
fl_get_editscrollbar_val(FL_OBJECT *ob)
/*
 * Desc: return the current editscrollbar value
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    return(spec->val);
} /* fl_get_editscrollbar_value() */

void
fl_set_editscrollbar_parameters(FL_OBJECT *ob, double min, double max, double val, double wsize)
/*
 * Desc: set editscrollbar parameters
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    if (max < 0) max = 0;
    spec->max = max;
    if (min < 0) min = 0;
    spec->min = min;
    if (wsize < 0) wsize = 0;
    spec->wsize = wsize;
    if (val < min) val = min;
    fl_editscrollbar_set_val(ob, val);

    //fl_redraw_object(ob);
}

void
fl_get_editscrollbar_parameters(FL_OBJECT *ob,double& min, double& max, double& val, double& wsize){
/*
 * Desc: get editscrollbar parameters
 */
    SPEC *spec = (SPEC *)ob->spec;

    max		= spec->max;
    min		= spec->min;
    wsize	= spec->wsize;
    val		= spec->val;
}


/***********************************************************************
 *
 * garbage
 *
 ***********************************************************************/

#ifdef notused

static double
fl_editscrollbar_calcval(FL_OBJECT *ob, int h)
/*
 * Desc: Calculate the editscrollbar value, corresponding to the
 *	 y-coordinate of the editscrollbarbutton and the height
 *	 of the editscrollbar 
 * Pre:  if (y==0) val = spec->min
 *	 [] (y==h) val = spec->max
 *	 fi
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    double val;

    val = (double) spec->coord/h * (spec->max - spec->min) + spec->min;

    return(val);
} /* fl_editscrollbar_calcval() */
#endif

static void
draw_buttons(FL_OBJECT * ob){
    SPEC *spec = ((SPEC *) (ob->spec));
    change_draw_buttons(ob, 
	spec->pushed == PUSHED_UPPER ? DRAW_PUSHED : DRAW_NORMAL,
	spec->pushed == PUSHED_LOWER ? DRAW_PUSHED : DRAW_NORMAL,
	TRUE
    );
}

static void
change_draw_buttons(FL_OBJECT * ob, 
    draw_et upper_state, 
    draw_et lower_state, 
    int bDrawCorner
){
    SPEC *sp = ((SPEC *) (ob->spec));
    FL_Coord button_x = ob->x, boxb = 0;
    int btype, button_w;

    button_w = (ob->type == FL_VERTICAL_EDITSCROLLBAR) ? ob->w : ob->h;

    if(upper_state == DRAW_PUSHED){
        sp->pushed = PUSHED_UPPER;
    } else if(lower_state == DRAW_PUSHED){
        sp->pushed = PUSHED_LOWER;
    } else {
        sp->pushed = PUSHED_NONE;
    }

    if(upper_state != DRAW_NONE){
	btype = ob->boxtype;
	if (upper_state == DRAW_PUSHED) {
	    btype = fix_boxtype(ob);
	}
        fl_drw_box(btype, button_x, ob->y, button_w, button_w, 
            FL_BROWSER_SLCOL, ob->bw
        );
        fl_drw_text_beside(
            FL_ALIGN_CENTER, 
            button_x+boxb, ob->y+boxb, 
            button_w-2*boxb, button_w-2*boxb,
            FL_BLACK, FL_FIXEDBOLD_STYLE, FL_MEDIUM_FONT, 
            (ob->type == FL_VERTICAL_EDITSCROLLBAR) ? (char *) "@8" : (char *) "@4"
        );
    }
    if(lower_state != DRAW_NONE){
	btype = ob->boxtype;
	if (lower_state == DRAW_PUSHED) {
	    btype = fix_boxtype(ob);
	}
	button_x = (ob->type == FL_VERTICAL_EDITSCROLLBAR) ? ob->x : ob->x+ob->w-button_w;
        fl_drw_box(btype, button_x, ob->y + ob->h - button_w, button_w, button_w, 
            FL_BROWSER_SLCOL, ob->bw
        );
        fl_drw_text_beside(
            FL_ALIGN_CENTER, 
            button_x+boxb, ob->y + ob->h - button_w -boxb, 
            button_w-2*boxb, button_w-2*boxb,
            FL_BLACK, FL_FIXEDBOLD_STYLE, FL_MEDIUM_FONT, 
            (ob->type == FL_VERTICAL_EDITSCROLLBAR) ? (char *) "@2" : (char *) "@6"
        );
    }

    if (bDrawCorner) {
        fl_drw_box(FL_UP_BOX, 
	    ob->x+ob->w-button_w,
	    ob->y + ob->h,
	    button_w, button_w, 
            FL_BROWSER_SLCOL, ob->bw
        );
    }
}
